Package: tcltk2
Type: Package
Version: 1.6.1
Date: 2025-06-01
Title: Tcl/Tk Additions
Description: A series of additional Tcl commands and Tk widgets to supplement the tcltk package.
Authors@R: c(person("Philippe", "Grosjean", role = c("aut", "cre"),
  email = "phgrosjean@sciviews.org",
  comment = c(ORCID = "0000-0002-2694-9471")))
Maintainer: Philippe Grosjean <phgrosjean@sciviews.org>
Depends: R (>= 4.2.0), tcltk
Suggests: utils, covr, knitr, rmarkdown
SystemRequirements: Tcl/Tk (>= 8.5), Tktable (>= 2.9, optional)
License: LGPL-3
URL: https://github.com/SciViews/tcltk2/
BugReports: https://github.com/SciViews/tcltk2/issues/
RoxygenNote: 7.3.2
VignetteBuilder: knitr
Encoding: UTF-8
Language: en-US
NeedsCompilation: no
Packaged: 2025-06-01 16:52:06 UTC; phgrosjean
Author: Philippe Grosjean [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-2694-9471>)
Repository: CRAN
Date/Publication: 2025-06-01 17:30:01 UTC
Built: R 4.6.0; ; 2026-04-26 09:20:29 UTC; unix

Package: tcltk2
Type: Package
Version: 1.6.1
Date: 2025-06-01
Title: Tcl/Tk Additions
Description: A series of additional Tcl commands and Tk widgets to supplement the tcltk package.
Authors@R: c(person("Philippe", "Grosjean", role = c("aut", "cre"),
  email = "phgrosjean@sciviews.org",
  comment = c(ORCID = "0000-0002-2694-9471")))
Maintainer: Philippe Grosjean <phgrosjean@sciviews.org>
Depends: R (>= 4.2.0), tcltk
Suggests: utils, covr, knitr, rmarkdown
SystemRequirements: Tcl/Tk (>= 8.5), Tktable (>= 2.9, optional)
License: LGPL-3
URL: https://github.com/SciViews/tcltk2/
BugReports: https://github.com/SciViews/tcltk2/issues/
RoxygenNote: 7.3.2
VignetteBuilder: knitr
Encoding: UTF-8
Language: en-US
NeedsCompilation: no
Packaged: 2025-06-01 16:52:06 UTC; phgrosjean
Author: Philippe Grosjean [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-2694-9471>)
Repository: CRAN
Date/Publication: 2025-06-01 17:30:01 UTC

# Generated by roxygen2: do not edit by hand

S3method("config<-",tk2label)
S3method("config<-",tk2widget)
S3method("disabled<-",tk2widget)
S3method("label<-",tk2widget)
S3method("selection<-",tk2listbox)
S3method("selection<-",tk2widget)
S3method("tag<-",tk2widget)
S3method("tip<-",tk2widget)
S3method("value<-",tk2listbox)
S3method("value<-",tk2widget)
S3method("values<-",tk2listbox)
S3method("values<-",tk2widget)
S3method(config,tk2label)
S3method(config,tk2widget)
S3method(disabled,tk2widget)
S3method(label,tk2widget)
S3method(print,tclTask)
S3method(print,tk2cfglist)
S3method(print,tk2widget)
S3method(selection,tk2listbox)
S3method(selection,tk2widget)
S3method(size,tk2listbox)
S3method(size,tk2widget)
S3method(state,tk2widget)
S3method(tag,tk2widget)
S3method(tip,tk2widget)
S3method(value,tk2listbox)
S3method(value,tk2widget)
S3method(values,tk2listbox)
S3method(values,tk2widget)
S3method(visibleItem,tk2listbox)
S3method(visibleItem,tk2widget)
export("config<-")
export("disabled<-")
export("label<-")
export("selection<-")
export("tag<-")
export("tip<-")
export("value<-")
export("values<-")
export(config)
export(disabled)
export(getLanguage)
export(is.tk)
export(is.tk2widget)
export(is.ttk)
export(label)
export(makeTclNames)
export(selection)
export(setLanguage)
export(size)
export(state)
export(tag)
export(tclAfter)
export(tclAfterCancel)
export(tclAfterInfo)
export(tclFun)
export(tclGetValue)
export(tclSetValue)
export(tclTaskChange)
export(tclTaskDelete)
export(tclTaskGet)
export(tclTaskRun)
export(tclTaskSchedule)
export(tclVarExists)
export(tclVarFind)
export(tclVarName)
export(tclmc)
export(tclmclocale)
export(tclmcset)
export(tip)
export(tk2button)
export(tk2canvas)
export(tk2cfglist)
export(tk2checkbutton)
export(tk2chooseFont)
export(tk2column)
export(tk2combobox)
export(tk2configList)
export(tk2ctext)
export(tk2dataList)
export(tk2dde)
export(tk2dde.exec)
export(tk2dde.poke)
export(tk2dde.request)
export(tk2dde.services)
export(tk2edit)
export(tk2entry)
export(tk2font.get)
export(tk2font.set)
export(tk2font.setstyle)
export(tk2frame)
export(tk2ico.create)
export(tk2ico.destroy)
export(tk2ico.list)
export(tk2ico.load)
export(tk2ico.set)
export(tk2ico.setFromFile)
export(tk2ico.sizes)
export(tk2insert.multi)
export(tk2killtip)
export(tk2label)
export(tk2labelframe)
export(tk2list.delete)
export(tk2list.get)
export(tk2list.insert)
export(tk2list.set)
export(tk2list.size)
export(tk2listbox)
export(tk2mclistbox)
export(tk2menu)
export(tk2menubutton)
export(tk2message)
export(tk2notebook)
export(tk2notetab)
export(tk2notetab.select)
export(tk2notetab.text)
export(tk2notetraverse)
export(tk2panedwindow)
export(tk2progress)
export(tk2radiobutton)
export(tk2reg.broadcast)
export(tk2reg.delete)
export(tk2reg.deletekey)
export(tk2reg.get)
export(tk2reg.keys)
export(tk2reg.set)
export(tk2reg.setkey)
export(tk2reg.type)
export(tk2reg.values)
export(tk2scale)
export(tk2scrollbar)
export(tk2separator)
export(tk2spinbox)
export(tk2state.set)
export(tk2style)
export(tk2swaplist)
export(tk2table)
export(tk2tablelist)
export(tk2text)
export(tk2theme)
export(tk2theme.elements)
export(tk2theme.list)
export(tk2tip)
export(tk2tree)
export(tk2unicode_bind)
export(tk2unicode_config)
export(tk2unicode_select)
export(value)
export(values)
export(visibleItem)
importFrom(tcltk,"tclvalue<-")
importFrom(tcltk,.Tcl)
importFrom(tcltk,.Tcl.callback)
importFrom(tcltk,as.tclObj)
importFrom(tcltk,tcl)
importFrom(tcltk,tclArray)
importFrom(tcltk,tclObj)
importFrom(tcltk,tclRequire)
importFrom(tcltk,tclServiceMode)
importFrom(tcltk,tclVar)
importFrom(tcltk,tclvalue)
importFrom(tcltk,tkactivate)
importFrom(tcltk,tkadd)
importFrom(tcltk,tkbind)
importFrom(tcltk,tkbindtags)
importFrom(tcltk,tkbutton)
importFrom(tcltk,tkcget)
importFrom(tcltk,tkconfigure)
importFrom(tcltk,tkcurselection)
importFrom(tcltk,tkdelete)
importFrom(tcltk,tkdestroy)
importFrom(tcltk,tkfocus)
importFrom(tcltk,tkfont.actual)
importFrom(tcltk,tkfont.configure)
importFrom(tcltk,tkfont.families)
importFrom(tcltk,tkfont.measure)
importFrom(tcltk,tkfont.names)
importFrom(tcltk,tkget)
importFrom(tcltk,tkgrab.set)
importFrom(tcltk,tkgrid)
importFrom(tcltk,tkgrid.columnconfigure)
importFrom(tcltk,tkgrid.configure)
importFrom(tcltk,tkgrid.rowconfigure)
importFrom(tcltk,tkindex)
importFrom(tcltk,tkinsert)
importFrom(tcltk,tklabel)
importFrom(tcltk,tkmessageBox)
importFrom(tcltk,tkscrollbar)
importFrom(tcltk,tksee)
importFrom(tcltk,tkselect)
importFrom(tcltk,tkselection.clear)
importFrom(tcltk,tkselection.set)
importFrom(tcltk,tkset)
importFrom(tcltk,tksize)
importFrom(tcltk,tktag.configure)
importFrom(tcltk,tktoplevel)
importFrom(tcltk,tkwait.window)
importFrom(tcltk,tkwidget)
importFrom(tcltk,tkwm.iconbitmap)
importFrom(tcltk,tkwm.title)
importFrom(tcltk,tkxview)
importFrom(tcltk,tkyview)

Mclistbox version 1.02
Copyright (c) 1999, Bryan Douglas Oakley
All Rights Reserved.

http://purl.oclc.org/net/oakley/tcl/mclistbox/index.html
mailto:oakley@channelpoint.com

This software is provided AS-IS with no waranty expressed or
implied. This software may be used free of charge, though I would
appreciate it if you give credit where credit is due and mention my
name when you use it.

To use, place mclistbox.tcl where your programs can get at it either
via autoloading or from a "source" command. Then, use
"mclistbox::mclistbox" to create a mclistbox, or do "import
::mclistbox::*" to use the mclistbox command without the namespace qualifier.

To see a quick example, start up a wish process and cd to the directory
where mclistbox.tcl is located. Then do "source example.tcl".

To run some simple tests, source "test.tcl" or "mclistbox.test"

Snit's Not Incr Tcl README.txt
-----------------------------------------------------------------

Snit is pure-Tcl object and megawidget framework.  See snit.html
for full details.

Snit is part of "tcllib", the standard Tcl Library.

Snit lives in "tcllib" now, but it is available separately at
http://www.wjduquette.com/snit.  If you have any questions, bug
reports, suggestions, or comments, feel free to contact me, Will
Duquette, at will@wjduquette.com; or, join the Snit mailing list (see
http://www.wjduquette.com/snit for details).

Differences Between Snit 2.1 and Snit 1.x
--------------------------------------------------------------------

V2.0 and V1.x are being developed in parallel.

  Version 2.1 takes advantage of some new Tcl/Tk 8.5 commands
  ([dict], [namespace ensemble], and [namespace upvar]) to improve 
  Snit's run-time efficiency.  Otherwise, it's intended to be 
  feature-equivalent with V1.x.  When running with Tcl/Tk 8.5, both 
  V2.0 and V1.x are available; when running with Tcl/Tk 8.3 or Tcl/Tk 
  8.4, only V1.x is available.

  Snit 1.x is implemented in snit.tcl; Snit 2.1 in snit2.tcl.

V2.1 includes the following enhancements over V1.x:

* A type's code (methods, type methods, etc.) can now call commands
  from the type's parent namespace without qualifying or importing
  them, i.e., type ::parentns::mytype's code can call
  ::parentns::someproc as just "someproc".

  This is extremely useful when a type is defined as part of a larger
  package, and shares a parent namespace with the rest of the package;
  it means that the type can call other commands defined by the
  package without any extra work.

  This feature depends on the new Tcl 8.5 [namespace path] command,
  which is why it hasn't been implemented for V1.x.  V1.x code can
  achieve something similar by placing

    namespace import [namespace parent]::*

  in a type constructor.  This is less useful, however, as it picks up
  only those commands which have already been exported by the parent
  namespace at the time the type is defined.

There are four incompatibilities between V2.1 and V1.x:

* Implicit naming of objects now only works if you set 
    
    pragma -hastypemethods 0

  in the type definition.  Otherwise, 

    set obj [mytype]

  will fail; you must use 

    set obj [mytype %AUTO%]

* In Snit 1.x and earlier, hierarchical methods and type methods
  could be called in two ways:

    snit::type mytype {
        method {foo bar} {} { puts "Foobar!"}
    }  

    set obj [mytype %AUTO%]
    $obj foo bar     ;# This is the first way
    $obj {foo bar}   ;# This is the second way

  In Snit 2.1, the second way no longer works.

* In Snit 1.x and earlier, [$obj info methods] and 
  [$obj info typemethods] returned a complete list of all known
  hierarchical methods.  In the example just above, for example,
  the list returned by [$obj info methods] would include 
  "foo bar".  In Snit 2.1, only the first word of a hierarchical
  method name is returned, [$obj info methods] would include 
  "foo" but not "foo bar".

* Because a type's code (methods, type methods, etc.) can now 
  call commands from the type's parent namespace without qualifying 
  or importing them, this means that all commands defined in the
  parent namespace are visible--and can shadow commands defined
  in the global namespace, including the standard Tcl commands.
  There was a case in Tcllib where the Snit type ::tie::std::file
  contained a bug with Snit 2.1 because the type's own name
  shadowed the standard [file] command in the type's own code.


Changes in V1.2
--------------------------------------------------------------------

* Defined a family of validation types.  Validation types are used
  to validate data values; for example, snit::integer and its
  subtypes can validate a variety of classes of integer value, e.g.,
  integers between 3 and 9 or integers greater than 0.

Changes in V1.1
--------------------------------------------------------------------

* It's now explicitly an error to call an object's "destroy" method
  in the object's constructor.  (If you need to do it, just throw
  an error; construction will fail and the object will be cleaned
  up.

* The Tile "ttk::frame" widget is now a valid hulltype for 
  snit::widgets.  Any widget with a -class option can be used
  as a hulltype; lappend the widget name to
  snit::hulltypes to enable its use as a hulltype.

* The TK labelframe widget and the Tile ttk::labelframe widget are
  now valid hulltypes for snit::widgets.

Changes in V1.0
--------------------------------------------------------------------

Functionally, V1.0 is identical to version V0.97.

* Added a number of speed optimizations provided by Jeff Hobbs.
  (Thanks, Jeff!)

* Returned to the name "Snit's Not Incr Tcl".

* Fixed SourceForge Tcllib Bug 1161779; it's no longer an error
  if the destructor is defined before the constructor.

* Fixed SourceForge Tcllib Bug 1106375; the hull widget is now
  destroyed properly if there's an error in the constructor of 
  a widget or widgetadaptor.

Changes in V0.97
--------------------------------------------------------------------

The changes listed here were actually made over time in Snit V0.96;
now that they are complete, the result has been renumbered Snit V0.97.

* Bug fix: methods called via [mymethod] can now return exotic
  return codes (e.g., "return -code break").

* Added the -hasinfo pragma, which controls whether there's an
  "info" instance method or not.  By default, there is.

* POSSIBLE INCOMPATIBILITY: If no options are defined for a type, neither
  locally nor delegated, then Snit will not define the "configure", 
  "configurelist", and "cget" instance methods or the "options" 
  instance variable.

* If a snit::type's command is called without arguments, AND the type 
  can have instances, then an instance is created using %AUTO% to 
  create its name.  E.g., the following commands are all equivalent:

    snit::type dog { ... }

    set mydog [dog create %AUTO%]
    set mydog [dog %AUTO%]
    set mydog [dog]

  This doesn't work for widgets, for obvious reasons.

* Added pragma -hastypemethods.  If its value is "yes" (the
  default), then the type has traditional Snit behavior with
  respect to typemethods.  If its value is "no", then the type
  has no typemethods (even if typemethods were included 
  explicitly in the type definition).  Instead, the first argument
  of the type proc is the name of the object to create.  As above,
  the first argument defaults to "%AUTO%" for snit::types but not
  for snit::widgets.

* Added pragma -simpledispatch.  This pragma is intended to make
  simple, heavily used types (e.g. stacks or queues) more efficient.
  If its value is "no" (the default), then the type has traditional
  Snit behavior with respect to method dispatch.  If its value is
  "yes", then a simpler, faster scheme is used; however, there are
  corresponding limitations. See the man page for details.

* Bug fix: the "pragma" statement now throws an error if the specified 
  pragma isn't defined, e.g., "pragma -boguspragma yes" is now an
  error.

* Bug fix: -readonly options weren't.  Now they are.

* Added support for hierarchical methods, like the Tk text widget's
  tag, mark, and image methods.  You define the methods like so:

    method {tag add}       {args} {...}
    method {tag configure} {args} {...}
    method {tag cget}      {args} {...}

  and call them like so:

    $widget tag add ....

  The "delegate method" statement also supports hierarchical methods.
  However, hierarchical methods cannot be used with -simpledispatch.

* Similarly, added support for hierarchical typemethods.

Changes in V0.96
--------------------------------------------------------------------

V0.96 was the development version in which most of the V0.97 changes
were implemented.  The name was changed to V0.97 when the changes
were complete, so that the contents of V0.97 will be stable.

Changes in V0.95
--------------------------------------------------------------------

The changes listed here were actually made over time in Snit V0.94;
now that they are complete, the result has been renumbered Snit V0.95.

* Snit method invocation (both local and delegated) has been 
  optimized by the addition of a "method cache".  The primary
  remaining cost in method invocation is the cost of declaring
  instance variables.

* Snit typemethod invocation now also uses a cache.

* Added the "myproc" command, which parallels "mymethod".  "codename"
  is now deprecated.

* Added the "mytypemethod" command, which parallels "mymethod".

* Added the "myvar" and "mytypevar" commands.  "varname" is now
  deprecated.

* Added ::snit::macro.

* Added the "component" type definition statement.  This replaces
  "variable" for declaring components explicitly, and has two nifty 
  options, "-public" and "-inherit".

* Reimplemented the "delegate method" and "delegate option"
  statements; among other things, they now have more descriptive error
  messages.

* Added the "using" clause to the "delegate method" statement.  The
  "using" clause allows the programmer to specify an arbitrary command
  prefix into which the component and method names (among other
  things) can be automatically substituted.  It's now possible to
  delegate a method just about any way you'd like.

* Added ::snit::compile.

* Added the "delegate typemethod" statement.  It's similar to 
  "delegate method" and has the same syntax, but delegates typemethods
  to commands whose names are stored in typevariables.

* Added the "typecomponent" type definition statement.  Parallel to
  "component", "typecomponent" is used to declare targets for the new 
  "delegate typemethod" statement.

* "delegate method" can now delegate methods to components or
  typecomponents.

* The option definition syntax has been extended; see snit.man.  You
  can now define methods to handle cget or configure of any option; as
  a result, The "oncget" and "onconfigure" statements are now deprecated.
  Existing "oncget" and "onconfigure" handlers continue to function as
  expected, with one difference: they get a new implicit argument,
  "_option", which is the name of the option being set.  If your
  existing handlers use "_option" as a variable name, they will need
  to be changed.

* In addition, the "option" statement also allows you to define a
  validation method.  If defined, it will be called before the value
  is saved; its job is to validate the option value and call "error"
  if there's a problem.

* In addition, options can be defined to be "-readonly".  A readonly
  option's value can be set at creation time (i.e., in the type's
  constructor) but not afterwards.

* There's a new type definition statement called "pragma" that 
  allows you to control how Snit generates the type from the
  definition.  For example, you can disable all standard typemethods
  (including "create"); this allows you to use snit::type to define
  an ensemble command (like "string" or "file") using typevariables
  and typemethods.

* In the past, you could create an instance of a snit::type with the 
  same name as an existing command; for example, you could create an
  instance called "::info" or "::set".  This is no longer allowed, as
  it can lead to errors that are hard to debug.  You can recover the
  old behavior using the "-canreplace" pragma.

* In type and widget definitions, the "variable" and "typevariable"
  statements can now initialize arrays as well as scalars.

* Added new introspection commands "$type info typemethods",
  "$self info methods", and "$self info typemethods".

* Sundry other internal changes.

Changes in V0.94
--------------------------------------------------------------------

V0.94 was the development version in which most of the V0.95 changes
were implemented.  The name was changed to V0.95 when the changes
were complete, so that the contents of V0.95 will be stable.

Changes in V0.93
--------------------------------------------------------------------

* Enhancement: Added the snit::typemethod and snit::method commands; 
  these allow typemethods and methods to be defined (and redefined)
  after the class already exists.  See the Snit man page for 
  details.

* Documentation fixes: a number of minor corrections were made to the
  Snit man page and FAQ.  Thanks to everyone who pointed them out,
  especially David S. Cargo.

* Bug fix: when using %AUTO% to create object names, the counter 
  will wrap around to 0 after it reaches (2^32 - 1), to prevent 
  integer overflow errors. (Credit Marty Backe)

* Bug fix: in a normal Tcl proc, the command

    variable ::my::namespace::var

  makes variable "::my::namespace::var" available to the proc under the 
  local name "var".  Snit redefines the "variable" command for use in
  instance methods, and had lost this behavior.  (Credit Jeff
  Hobbs)

* Bug fix: in some cases, the "info vars" instance method didn't
  include the "options" instance variable in its output.

* Fixed bug: in some cases the type command was created even if there 
  was an error defining the type.  The type command is now cleaned 
  up in these cases.  (Credit Andy Goth)


Changes in V0.92
--------------------------------------------------------------------

* Bug fix: In type methods, constructors, and methods, the "errorCode"
  of a thrown error was not propagated properly; no matter what it was
  set to, it always emerged as "NONE".

Changes in V0.91
--------------------------------------------------------------------

* Bug fix: On a system with both 0.9 and 0.81 installed, 
  "package require snit 0.9" would get snit 0.81.  Here's why: to me
  it was clear enough that 0.9 is later than 0.81, but to Tcl the 
  minor version number 9 is less than minor version number 81.
  From now on, all pre-1.0 Snit version numbers will have two
  digits.

* Bug fix: If a method or typemethod had an argument list which was
  broken onto multiple lines, the type definition would fail. It now
  works as expected.

* Added the "expose" statement; this allows you to expose an entire
  component as part of your type's public interface.  See the man page
  and the Snit FAQ list for more information.

* The "info" type and instance methods now take "string match"
  patterns as appropriate.

Changes in V0.9
--------------------------------------------------------------------

For specific changes, please see the file ChangeLog in this directory.
Here are the highlights:

* Snit widgets and widget adaptors now support the Tk option database.

* It's possible set the hull type of a Snit widget to be either a
  frame or a toplevel.

* It's possible to explicitly set the widget class of a Snit widget.

* It's possible to explicitly set the resource and class names for
  all locally defined and explicitly delegated options.

* Option and method names can be excluded from "delegate option *" by
  using the "except" clause, e.g.,

     delegate option * to hull except {-borderwidth -background}

* Any Snit type or widget can define a "type constructor": a body of
  code that's executed when the type is defined.  The type constructor
  is typically used to initialize array-valued type variables, and to
  add values to the Tk option database.

* Components should generally be created and installed using the new
  "install" command.

* snit::widgetadaptor hulls should generally be created and installed
  using the new "installhull using" form of the "installhull" command.

See the Snit man page and FAQ list for more information on these new 
features.


Changes in V0.81
--------------------------------------------------------------------

* All documentation errors people e-mailed to me have been fixed.

* Bug fix: weird type names.  In Snit 0.8, type names like
  "hyphenated-name" didn't work because the type name is used as a
  namespace name, and Tcl won't parse "-" as part of a namespace name
  unless you quote it somehow.  Kudos to Michael Cleverly who both
  noticed the problem and contributed the patch.

* Bug fix: Tcl 8.4.2 incompatibility.  There was a bug in Tcl 8.4.1
  (and in earlier versions, likely) that if the Tcl command "catch"
  evaluated a block that contained an explicit "return", "catch" 
  returned 0.  The documentation evidently indicated that it should
  return 2, and so this was fixed in Tcl 8.4.2.  This broke a bit
  of code in Snit.

Changes in V0.8
--------------------------------------------------------------------

* Note that there are many incompatibilities between Snit V0.8 and
  earlier versions; they are all included in this list.

* Bug fix: In Snit 0.71 and Snit 0.72, if two instances of a
  snit::type are created with the same name, the first instance's
  private data is not destroyed.  Hence, [$type info instances] will
  report that the first instance still exists.  This is now fixed.

* Snit now requires Tcl 8.4, as it depends on the new command
  tracing facility.

* The snit::widgettype command, which was previously deprecated, has
  now been deleted.

* The snit::widget command has been renamed snit::widgetadaptor; its
  usage is unchanged, except that the idiom "component hull is ..." 
  is no longer used to define the hull component.  Instead, use the
  "installhull" command:

        constructor {args} {
            installhull [label $win ...]
            $self configurelist $args
        }

* The "component" command is now obsolete, and has been removed.
  Instead, the "delegate" command implicitly defines an instance
  variable for the named component; the constructor should assign an
  object name to that instance variable.  For example, whereas you
  used to write this: 

    snit::type dog {
        delegate method wag to tail

        constructor {args} {
            component tail is [tail $self.tail -partof self]
        }

        method gettail {} {
            return [component tail]
        }
    }

  you now write this:

    snit::type dog {
        delegate method wag to tail

        constructor {args} {
            set tail [tail $self.tail -partof self]
        }

        method gettail {} {
            return $tail
        }
    }

* There is a new snit::widget command; unlike snit::widgetadaptor,
  snit::widget automatically creates a Tk frame widget as the hull
  widget; the constructor doesn't need to create and set a hull component.

* Snit objects may now be renamed without breaking; many of the
  specific changes which follow are related to this.  However,
  there are some new practices for type authors to follow if they wish
  to write renameable types and widgets.  In particular,

  * In an instance method, $self will always contain the object's
    current name, so instance methods can go on calling other instance
    methods using $self.

  * If the object is renamed, then $self's value will change.  Therefore, 
    don't use $self for anything that will break if $self changes.
    For example, don't pass a callback as "[list $self methodname]".

  * If the object passes "[list $self methodname arg1 arg2]" as a callback, 
    the callback will fail when the object is renamed.  Instead, the 
    object should pass "[mymethod methodname arg1 arg2]".  The [mymethod]
    command returns the desired command as a list beginning with a
    name for the object that never changes.

    For example, in Snit V0.71 you might have used this code to call a
    method when a Tk button is pushed: 

     .btn configure -command [list $self buttonpress]

    This still works in V0.8--but the callback will break if your
    instance is renamed.  Here's the safe way to do it:

     .btn configure -command [mymethod buttonpress]

  * Every object has a private namespace; the name of this namespace
    is now available in method bodies, etc., as "$selfns".  This value is
    constant for the life the object.  Use "$selfns" instead of "$self" if
    you need a unique token to identify the object.

  * When a snit::widget's instance command is renamed, its Tk window
    name remains the same--and is still extremely important.
    Consequently, the Tk window name is now available in snit::widget
    method bodies, etc., as "$win".  This value is constant for the
    life of the object.  When creating child windows, it's best to 
    use "$win.child" rather than "$self.child" as the name of the
    child window. 

* The names "selfns" and "win" may no longer be used as explicit argument
  names for typemethods, methods, constructors, or onconfigure
  handlers.

* procs defined in a Snit type or widget definition used to be able to
  reference instance variables if "$self" was passed to them
  explicitly as the argument "self"; this is no longer the case.

* procs defined in a Snit type or widget definition can now reference
  instance variables if "$selfns" is passed to them explicitly as the
  argument "selfns".  However, this usage is deprecated.

* All Snit type and widget instances can be destroyed by renaming the
  instance command to "".

Changes in V0.72
--------------------------------------------------------------------

* Updated the pkgIndex.tcl file to references snit 0.72 instead of
  snit 0.7.

* Fixed a bug in widget destruction that caused errors like
  "can't rename "::hull1.f": command doesn't exist".

Changes in V0.71
--------------------------------------------------------------------

* KNOWN BUG: The V0.7 documentation implies that a snit::widget can
  serve as the hull of another snit::widget.  Unfortunately, it
  doesn't work.  The fix for this turns out to be extremely
  complicated, so I plan to fix it in Snit V0.8.

  Note that a snit::widget can still be composed of other
  snit::widgets;  it's only a problem when the hull component in
  particular is a snit::widget.

* KNOWN BUG: If you rename a Snit type or instance command (i.e., using
  Tcl's [rename] command) it will no longer work properly.  This is
  part of the reason for the previous bug, and should also be fixed in
  Snit V0.8.

* Enhancement: Snit now preserves the call stack (i.e., the
  "errorInfo") when rethrowing errors thrown by Snit methods,
  typemethods, and so forth.  This should make debugging Snit types
  and widgets much easier.  In Snit V0.8, I hope to clean up the
  call stack so that Snit internals are hidden.

* Bug fix: Option default values were being processed incorrectly.  In
  particular, if the default value contained brackets, it was treated
  as a command interpolation.  For example,

    option -regexp {[a-z]+}

  yield the error that "a-z" isn't a known command.  Credit to Keith
  Waclena for finding this one.

* Bug fix: the [$type info instances] command failed to find
  instances that weren't defined in the global namespace, and found
  some things that weren't instances.  Credit to Keith Waclena for
  finding this one as well.

* Internal Change: the naming convention for instance namespaces
  within the type namespace has changed.  But then, your code
  shouldn't have depended on that anyway.

* Bug fix: snit::widget destruction was seriously broken if the hull
  component was itself a megawidget (e.g., a BWidget).
  Each layer of megawidget code needs its opportunity
  to clean up properly, and that wasn't happening.  In addition, the
  snit::widget destruction code was bound as follows:

    bind $widgetName <Destroy> {....}

  which means that if the user of a Snit widget needs to bind to
  <Destroy> on the widget name they've just wiped out Snit's
  destructor.  Consequently, Snit now creates a bindtag called
  
    Snit<widgettype>

  e.g.,

    Snit::rotext

  and binds its destroy handler to that.  This bindtag is inserted in
  the snit::widget's bindtags immediately after the widget name.

  Destruction is always going to be somewhat tricky when multiple
  levels of megawidgets are involved, as you need to make sure that
  the destructors are called in inverse order of creation.

Changes in V0.7
----------------------------------------------------------------------

* INCOMPATIBILITY: Snit constructor definitions can now have arbitrary
  argument lists, as methods do.  That is, the type's create method 
  expects the instance name followed by exactly the arguments defined
  in the constructor's argument list: 

    snit::type dog {
        variable data
        constructor {breed color} {
            set data(breed) $breed
            set data(color) $color
        }
    }

    dog spot labrador chocolate

  To get the V0.6 behavior, use the argument "args".  That is, the
  default constructor would be defined in this way:

    snit::type dog {
        constructor {args} {
            $self configurelist $args
        }
    }

* Added a "$type destroy" type method.  It destroys all instances of
  the type properly (if possible) then deletes the type's namespace
  and type command.

Changes in V0.6
-----------------------------------------------------------------

* Minor corrections to the man page.

* The command snit::widgettype is deprecated, in favor of
  snit::widget.

* The variable "type" is now automatically defined in all methods,
  constructors, destructors, typemethods, onconfigure handlers, and
  oncget handlers.  Thus, a method can call type methods as "$type
  methodname".

* The new standard instance method "info" is used for introspection on 
  type and widget instances:

  $object info type
     Returns the object's type.

  $object info vars
     Returns a list of the object's instance variables (excluding Snit
     internal variables).  The names are fully qualified.

  $object info typevars
     Returns a list of the object's type's type variables (excluding
     Snit internal variables).  The names are fully qualified.

  $object info options
     Returns a list of the object's option names.  This always
     includes local options and explicitly delegated options.  If
     unknown options are delegated as well, and if the component to
     which they are delegated responds to "$object configure" like Tk
     widgets do, then the result will include all possible unknown
     options which could be delegated to the component.  

     Note that the return value might be different for different
     instances of the same type, if component object types can vary
     from one instance to another.

* The new standard typemethod "info" is used for introspection on
  types:

  $type info typevars
     Returns a list of the type's type variables (excluding Snit
     internal variables).

  $type info instances
     Returns a list of the instances of the type.  For non-widget
     types, each instance will be the fully-qualified instance command
     name; for widget types, each instance will be a widget name.

* Bug fixed: great confusion resulted if the hull component of a
  snit::widgettype was another snit::widgettype.  Snit takes over the
  hull widget's Tk widget command by renaming it to a known name, and
  putting its own command in its place.  The code made no allowance
  for the fact that this might happen more than once; the second time,
  the original Tk widget command would be lost.  Snit now ensures that
  the renamed widget command is given a unique name.

* Previously, instance methods could call typemethods by name, as
  though they were normal procs.  The downside to this was that
  if a typemethod name was the same as a standard Tcl command, the
  typemethod shadowed the standard command in all of the object's
  code.  This is extremely annoying should you wish to define a
  typemethod called "set".  Instance methods must now call typemethods
  using the type's command, as in "$type methodname".

* Typevariable declarations are no longer required in
  typemethods, methods, or procs provided that the typevariables are defined
  in the main type or widget definition.

* Instance variable declarations are no longer required in methods provided
  that the instance variables are defined in the main type or widget
  declaration.

* Instance variable declarations are no longer required in procs,
  provided that the instance variables are defined in the main type or
  widget declaration.  Any proc that includes "self" in its argument
  list will pick up all such instance variables automatically.

* The "configure" method now returns output consistent with Tk's when
  called with 0 or 1 arguments, i.e., it returns information about one
  or all options.  For options defined by Snit objects, the "dbname"
  and "classname" returned in the output will be {}.  "configure" does
  its best to do the right thing in the face of delegation.

* If the string "%AUTO%" appears in the "name" argument to "$type create"
  or "$widgettype create", it will be replaced with a string that
  looks like "$type$n", where "$type" is the type name and "$n" is 
  a counter that's incremented each time a
  widget of this type is created.  This allows the caller to create 
  effectively anonymous instances:

  widget mylabel {...}

  set w [mylabel .pane.toolbar.%AUTO% ...]
  $w configure -text "Some text"

* The "create" typemethod is now optional for ordinary types so long
  as the desired instance name is different than any typemethod name
  for that type.  Thus, the following code creates two dogs, ::spot
  and ::fido.

  type dog {...}

  dog create spot
  dog fido

  If there's a conflict between the instance name and a typemethod, 
  either use "create" explicitly, or fully qualify the instance name:

  dog info -color black           ;# Error; assumes "info" typemethod.
  dog create info -color black    ;# OK
  dog ::info -color black         ;# also OK

* Bug fix: If any Snit method, typemethod, constructor, or onconfigure
  handler defines an explicit argument called "type" or "self", the type
  definition now throws an error, preventing confusing runtime
  behavior.

* Bug fix: If a Snit type or widget definition attempts to define a
  method or option locally and also delegate it to a component, the 
  type definition now throws an error, preventing confusing runtime 
  behavior.

* Bug(?) Fix: Previously, the "$self" command couldn't be used in
  snit::widget constructors until after the hull component was
  defined.  It is now possible to use the "$self" command to call
  instance methods at any point in the snit::widget's
  constructor--always bearing in mind that it's an error to configure
  delegated options or are call delegated methods before creating the
  component to which they are delegated.

Changes in V0.5
------------------------------------------------------------------

* Updated the test suite so that Tk-related tests are only run if
  Tk is available.  Credit Jose Nazario for pointing out the problem.

* For snit::widgettypes, the "create" keyword is now optional when 
  creating a new instance.  That is, either of the following will
  work:

  ::snit::widgettype mylabel { }

  mylabel create .lab1 -text "Using create typemethod"
  mylabel .lab2 -text "Implied create typemethod"

  This means that snit::widgettypes can be used identically to normal
  Tk widgets.  Credit goes to Colin McCormack for suggesting this.

* Destruction code is now defined using the "destructor" keyword
  instead of by defining a "destroy" method.  If you've been 
  defining the "destroy" method, you need to replace it with 
  "destructor" immediately.  See the man page for the syntax.

* widgettype destruction is now handled properly (it was buggy).  
  Use the Tk command "destroy" to destroy instances of a widgettype;
  the "destroy" method isn't automatically defined for widgettypes as
  it is for normal types, and has no special significance even if it
  is defined.

* Added the "from" command to aid in parsing out specific option
  values in constructors.

Changes in V0.4
------------------------------------------------------------------

* Added the "codename" command, to qualify type method and private
  proc names.

* Changed the internal implementation of Snit types and widget types
  to prevent an obscure kind of error and to make it easier to pass
  private procs as callback commands to other objects.  Credit to Rolf
  Ade for discovering the hole.

Changes in V0.3
------------------------------------------------------------------

* First public release.

# TODO: Rework all this!!!

#dim.tclArray <- function (ta)
#{
#  nms <- grep(",", names(ta), value = TRUE)
#  if (length(nms) == 0) return(c(0, 0))
#  c(max(as.numeric(gsub(",.*", "", nms))),
#      max(as.numeric(gsub(".*,", "", nms)))) + 1
#}


#' Edit a matrix or data frame in spreadsheet-like editor
#'
#' A tkTable widget is used to display and edit a matrix or data frame. One
#' can edit entries, add or delete rows and columns, ...
#'
#' @param x A matrix or data frame to edit.
#' @param title The title of the editor window.
#' @param header Do we display a header?
#' @param maxHeight The maximum height of the editor window.
#' @param maxWidth The maximum width of the editor window.
#' @param fontsize The size of the font to use in the editor window.
#' @param ... Further arguments to pass to the function.
#'
#' @return The function is used for its side-effet, that is, to modify a matrix or data
#' frame in a spreadsheet-like editor.
#' @note You need the tkTable widget to use this function.
#' @export
#' @author Jeffrey J. Hallman
#' @seealso [tclSetValue()]
#'
#' @examples
#' \dontrun{
#' # These cannot be run by examples() but should be OK when pasted
#' # into an interactive R session with the tcltk package loaded
#' data(iris)
#' tk2edit(iris)
#' }
tk2edit <- function(x, title = "Matrix Editor", header = NULL,
maxHeight = 600, maxWidth = 800, fontsize = 9, ...) {
  if (!is.tk())
    stop("Package Tk is required but not loaded")
  if (!inherits(tclRequire("Tktable", warn = FALSE), "tclObj"))
    stop("Tcl package 'Tktable' must be installed first")
  .Tcl(paste("option add *Table.font {courier", fontsize, "bold}"))
  old <- options(scipen = 7)
  on.exit(options(old))

  makeCharMat <- function(x) {
    # Make sure it's a character matrix
    mat <- matrix(unlist(x), nrow = nrow(as.matrix(x)))
    dm <- dim(mat)

    # Check for row and column names
    hasRownames <- length(rn <- rownames(x)) > 0
    hasColnames <- length(cn <- colnames(x)) > 0
    # Fake row and column names if they aren't there
    if (!hasRownames)
      rn <- paste("[", 1:nrow(x), ",]", sep = "")
    if (!hasColnames)
      cn <- paste("[,", 1:ncol(x), "]", sep = "")

    # Format the columns
    mat[] <- apply(unclass(mat), 2, format, justify = "right")
    mat <- rbind(cn, mat)
    mat <- cbind(c("", rn), mat)
    mat
  }

  fillTclArrayFromCharMat <- function(ta, cm) {
    # cm[,1] contains column names, while cm[1,] has rownames
    # cm[1,1] is ignored
    for (j in 2:ncol(cm))
      ta[[0, j - 1]] <- as.tclObj(cm[1, j], drop = TRUE)
    for (i in 2:nrow(cm))
      for (j in 1:ncol(cm))
        ta[[i - 1, j - 1]] <- as.tclObj(cm[i, j], drop = TRUE)
  }

  tA <- tclArray()
  cmat <- makeCharMat(x)
  fillTclArrayFromCharMat(tA, cmat)

  tt <- tktoplevel()
  tkwm.title(tt, title)

  colwidths <- apply(cmat, 2, function(x) max(nchar(x)) + 1 )
  nTableCols <- ncol(cmat)
  if ((moreWidth <- 60 - sum(colwidths)) > 0) {
    addEach <- moreWidth %/% length(colwidths)
    if (addEach < 5) {
      colwidths <- colwidths + addEach + 1
    } else {
      nTableCols <- nTableCols + ceiling(moreWidth / 10)
    }
  }

  tktable <- tkwidget(tt, "table", variable = tA,
    rows = nrow(cmat), cols = nTableCols,
    titlerows = 1, titlecols = 1, selecttitle = 1,
    anchor = "e", multiline = 0, selectmode = "extended",
    rowseparator = dQuote("\n"), colseparator = dQuote("\t"),
    background = "white", maxheight = maxHeight, maxwidth = maxWidth,
    xscrollcommand = function(...) tkset(xscr, ...),
    yscrollcommand = function(...) tkset(yscr, ...))
  xscr <- tkscrollbar(tt, orient = "horizontal",
    command = function(...) tkxview(tktable, ...))
  yscr <- tkscrollbar(tt,
    command = function(...) tkyview(tktable, ...))

  # Set column widths
  for (i in 1:ncol(cmat))
    tcl(tktable, "width", i - 1, colwidths[i])

  # Rebind the Backspace key, which somehow gets messed up
  string <- "bind Table <BackSpace> {
    set ::tk::table::Priv(junk) [%W icursor]
      if {[string compare {} $::tk::table::Priv(junk)] && $::tk::table::Priv(junk)} {
    %W delete active [expr {$::tk::table::Priv(junk)-1}]
      }}"
  .Tcl(string)

  # Internal functions for buttons
  activeRow <- function()
    as.numeric(tkindex(tktable, "active", "row"))

  activeCol <- function()
    as.numeric(tkindex(tktable, "active", "col"))

  undoEdits <- function() {
    ta <- tclArray()
    fillTclArrayFromCharMat(ta, cmat)
    assign("tA", ta, inherits = TRUE)
    tkconfigure(tktable, variable = tA)
  }

  finish <- function()
    tkdestroy(tt)

  cancel <- function() {
    undoEdits()
    tkdestroy(tt)
  }

  insertRow <- function() {
    row <- activeRow()
    col <- activeCol()
    tkinsert(tktable, "rows", row, 1)
    newCell <- paste(row + 1, col, sep = ",")
    tkactivate(tktable, newCell)
    tksee(tktable, newCell)
  }

  insertCol <- function() {
    row <- activeRow()
    col <- activeCol()
    tkinsert(tktable, "cols", col, 1)
    newCell <- paste(row, col + 1, sep = ",")
    tkactivate(tktable, newCell)
    tksee(tktable, newCell)
  }

  deleteRow <- function() {
    if ((row <- activeRow()) != 0)
      tkdelete(tktable, "rows", row, 1)
  }

  deleteCol <- function() {
    if ((col <- activeCol()) != 0)
      tkdelete(tktable, "cols", col, 1)
  }

  copyRow <- function() {
    src <- activeRow()
    if (src != 0) {
      insertRow()
      dst <- activeRow()
      for (j in 0:(ncol(tA) - 1))
        tA[[dst, j]] <- tA[[src, j]]
    }
  }

  copyCol <- function() {
    src <- activeCol()
    if (src != 0) {
      insertCol()
      dst <- activeCol()
      for (i in 0:(nrow(tA) - 1))
        tA[[i, dst]] <- tA[[i,src]]
    }
  }

  finishButton <- tkbutton(tt, text = "Finish", command = finish)
  cancelButton <- tkbutton(tt, text = "Cancel", command = cancel)
  undoEditsButton <- tkbutton(tt, text = "Undo Edits", command = undoEdits)
  insertRowButton <- tkbutton(tt, text = "Insert Row", command = insertRow)
  copyRowButton <- tkbutton(tt, text = "Copy Row", command = copyRow)
  deleteRowButton <- tkbutton(tt, text = "Delete Row", command = deleteRow)
  insertColButton <- tkbutton(tt, text = "Insert Col", command = insertCol)
  copyColButton <- tkbutton(tt, text = "Copy Col", command = copyCol)
  deleteColButton <- tkbutton(tt, text = "Delete Col", command = deleteCol)

  # Layout
  if (length(header) > 0) {
    for (label in header)
      tkgrid(tklabel(tt, text = label), columnspan = 7, sticky = "nw")
  }
  tkgrid(tktable, yscr, columnspan = 8)
  tkgrid.configure(tktable, sticky = "news")
  tkgrid.configure(yscr, sticky = "nsw")
  tkgrid(xscr, sticky = "new", columnspan = 8)
  tkgrid(insertRowButton, copyRowButton, deleteRowButton, sticky = "news")
  tkgrid(insertColButton, copyColButton, deleteColButton,
      "x", cancelButton, undoEditsButton, finishButton, sticky = "news")
  tkgrid.columnconfigure(tt, 3, weight = 1)
  tkgrid.rowconfigure(tt, length(header), weight = 1)
  tkactivate(tktable, "0,0")
  tktag.configure(tktable, "active", background = "lightyellow2")
  tktag.configure(tktable, "title", state = "normal")

  tkgrab.set(tt)
  tkfocus(tt)
  tkwait.window(tt)

  outMat <- matrix("", nrow = nrow(tA), ncol = ncol(tA))

  for (i in 1:nrow(outMat)) {
    for (j in 1:ncol(outMat)) {
      val <- tA[[i - 1,j - 1]]
      if (is.null(val)) {
        val <- ""
      } else {
        val <- tclvalue(val)
      }
      outMat[i,j] <- val
    }
  }

  # Recover row and column names
  rn <- outMat[, 1][-1]
  cn <- outMat[1, ][-1]
  outMat <- outMat[-1, -1, drop = FALSE]

  # Ignore bad and/or NA row and column names
  badRownames <- c(grep("\\[.*\\]", rn), (1:length(rn))[is.na(rn)])
  if (length(badRownames) != length(rn)) {
    rn[badRownames] <- ""
    rownames(outMat) <- rn
  }
  badColnames <- c(grep("\\[.*\\]", cn), (1:length(cn))[is.na(cn)])
  if (length(badColnames) != length(cn)) {
    cn[badColnames] <- ""
    colnames(outMat) <- cn
  }
  mode(outMat) <- mode(x)
  Sys.sleep(0.1)
  outMat
}

#' Use DDE (Dynamic Data Exchange) under Windows
#'
#' DDE is the first Microsoft's attempt to make an inter-application mechanism.
#' It is now superseeded by (D)Com, but it is still available (although declared
#' as unsupported). Being simpler than Com, DDE is interesting for simple tasks.
#' Applications like Word or Excel provide services one can access through DDE
#' (see examples). This code if left for backward compatibility, and also, just
#' in case you will find some use of it. But for new projects in general, you
#' should not use this any more.
#'
#' @param topic The 'topic' to reach or expose. A DDE server is accessed as
#' service'|'topic'. In the case of [tk2dde()], a non null topic activates
#' the DDE server, and a null topic deactivate it.
#' @param service The name of the service to reach. In `tk2dde.services`, if
#' both service and topic are empty, the list of all available DDE service is
#' returned, otherwise, only available topics for a given service are listed.
#' @param command A string with the command to run in the external application
#' (syntax depends on the server).
#' @param async Is a command run asynchroneously (returns immediately, before
#' the command is processed), or not?
#' @param item The concerned item (usually a variable name, a range in a
#' worksheet, etc...).
#' @param data The new value for the item.
#' @param binary Should the return be treated as binary data or not?
#'
#' @note
#' This is only available under Windows. Trying to use these functions under
#' other platforms raises an error. Under Windows, R is automatically configured
#' as a DDE server with name 'TclEval|SciViewsR' when this package is loaded.
#'
#' @export
#' @rdname tk2dde
#' @author Philippe Grosjean
#' @seealso [tk2reg.get()]
#' @keywords utilities
#'
#' @examples
#' \dontrun{
#' # These cannot be run by examples() but should be OK when pasted
#' # into an interactive R session with the tcltk package loaded
#'
#' # Examples of DDE - Windows only
#'
#' ### Examples using wish ###
#' # Start a Wish84 console side-by-side with R.
#' # (to get wish, you need to install ActiveTcl from
#' # http://www.activestate.com/Products/ActiveTcl/)
#' # Once it is done, start 'Wish84' from the start menu)
#' # Register the Wish console as a DDE server, that is, type in it
#' # (% is the Tcl prompt, do not type it!):
#' # % package require dde
#' # % dde servername wish
#'
#' ### In R:
#' tk2dde("R") # Return 0 if succeed
#' tk2dde.services()
#' # Evaluate some string in wish
#' tk2dde.exec("TclEval", "wish", "{puts {Hello World!}}")
#' # Give a value to a variable in wish
#' tk2dde.poke("TclEval", "wish", "myvar", "{This is a string!}")
#' # Note that you must surround strings with curly braces in Tcl!
#' tk2dde.poke("TclEval", "wish", "mynumvar", c(34.56, 78.9))
#'
#' # In wish, check that vars exist and have correct value
#' # % puts $myvar
#' # % puts $mynumvar
#'
#' # Get the value of one variable from wish into R
#' tk2dde.request("TclEval", "wish", "myvar")
#' tk2dde.request("TclEval", "wish", "mynumvar")
#' # Note that you do not know here if it is a string, a number, or so...
#' # You have to know and convert yourself!
#'
#' # Now, the other way: execute a R function from wish
#' # You first need to register a R function for callback
#' # (For the moment, only functions without arguments are supported!)
#' doDDE <- function() cat("DDE execute!")	# A simple function
#' tclFun(doDDE)
#' # And in wish
#' # % dde execute TclEval R doDDE
#'
#' # Once you have defined a variable using tclVar, you can get or change it
#' # from the dde server. However, tclVar gives cryptic names like ::RTcl1.
#' # So we prefer to use tclVarName()
#' myvar2 <- tclVarName("myvar2", "this is a test...")
#' tclvalue(myvar2) # This is the way we access to this variable in R
#'
#' # In wish you get the value and change it:
#' # % dde request TclEval R myvar2
#' # Again, dde poke does not work and must be replaced by an execute command
#' # This does not work (???)
#' # % dde poke TclEval R myvar2 {yes! and it works...}
#' # ... but this is fine
#' # % dde execute TclEval R {set myvar2 {yes! and it works...}}
#'
#' # And in R...
#' tclvalue(myvar2)
#'
#' ### DDE at the command line with execdde.exe ###
#' # You can also change the value of a variable, or run a command in R from
#' # the command line using execdde.exe:
#' # - Download execdde.exe from http://www.sciviews.org/SciViews-R/execdde.zip
#' # - Unzip it and copy 'execdde.exe' somewhere in your path,
#' # - Start a DOS window
#' # - Enter the following commands ('>' is the prompt, do not type it):
#' # > execdde -s TclEval -t R -c doDDE > NUL
#' # > if errorlevel 1 echo An error occurs... branch accordingly in your batch!
#' # > execdde -s TclEval -t R -c "set myvar2 'ok from execdde'" > NUL
#'
#' # And in R:
#' tclvalue(myvar2)
#' # Note: thanks to separate event loops, it works also when R calculates...
#'
#' ### Manipulating Microsoft Excel ###
#' # Start Excel with a blank workbook, then...
#'
#' # Change values in Excel from R:
#' tk2dde.poke("Excel", "Sheet1", "R1C1:R2C1", c("5.7", "6.34"))   # Some data
#' tk2dde.poke("Excel", "Sheet1", "R3C1", "= A1 + A2")             # A formula
#'
#' # Read values in Excel (note that results of formulas are returned)
#' Res <- tk2dde.request("Excel", "Sheet1", "R1C1:R3C1")
#' Res
#' as.numeric(Res)
#'
#' }
tk2dde <- function(topic = NULL) {
  # Initialize a tcltk dde server with name 'TclEval|topic'
  .tk2dde.require()

  # If topic is NULL, just get my server name
  if (is.null(topic))
    return(tclvalue(.Tcl("dde servername {}")))

  # Otherwise topic must be character
  topic <- topic[1]
  if (!is.character(topic) || topic == "")
    stop("'topic' must be a non null character string!")

  # Verify if I am not already registered under this topic
  if (tclvalue(.Tcl("dde servername {}")) == topic)
    return(0)  # OK

  # Check that this server name does not exist yet
  if (length(grep(paste("[{]TclEval ", topic, "[}]", sep = ""),
    as.character(.Tcl("dde services TclEval {}")))) > 0)
    return(1)  # This server name already exists => return 1 and don't set!

  # Register me as a dde server with this topic name
  .Tcl(paste("dde servername", topic))
  # Check that the server is set correctly
  # (if not, return 2 to warn that a problem occurred)
  if (tclvalue(.Tcl("dde servername {}")) == topic) {
    return(0)
  } else {
    return(2)
  }
}

#' @export
#' @rdname tk2dde
tk2dde.exec <- function(service, topic, command, async = FALSE) {
  # Execute a command in the 'service|topic' dde server
  .tk2dde.require()

  if (!is.character(service) || !is.character(topic) || !is.character(command))
    stop("'service', 'topic' and 'command' must be character strings!")
  if (async[1] == TRUE) async <- "-async" else async <- ""

  # Execute the command in a try(), to nicely catch the error
  # class is "try-error" if an error occurs, otherwise, returns ""
  res <- (try(tclvalue(.Tcl(paste("dde execute ", async, " ",
    as.character(service[1]), " ", as.character(topic[1]), " ",
    as.character(command[1]), sep = "")))))
  res
}

#' @export
#' @rdname tk2dde
tk2dde.poke <- function(service, topic, item, data) {
  # Set a value (data) to 'item' in the 'service|topic' dde server's app
  .tk2dde.require()

  if (!is.character(service) || !is.character(topic))
    stop("'service' and 'topic' must be character strings!")
  if (!is.character(item))
    stop("'item' must be character strings!")
  # In Tcl, if 'data' is a character string, enclose it in curly braces
  data <- paste("{", paste(as.character(data), collapse = "\n"), "}", sep = "")

  # For some reasons, dde poke does not seem to work with a TclEval serve...
  # use dde execute instead
  if (service == "TclEval") {
    Cmd <- paste("{set ", as.character(item[1]), " ", data, "}", sep = "")
    # This would not work with all kind of data!!!
    # Also, if it is a vector, matrix, or array, it does not work properly!
    return(tk2dde.exec(service, topic, Cmd, async = TRUE))
  }

  # Poke the data within a try(), to nicely catch the error
  # class is "try-error" if an error occurs, otherwise, returns ""
  res <- (try(as.character(.Tcl(paste("dde poke", as.character(service[1]),
    as.character(topic[1]), as.character(item[1]), data)))))
  res
}

#' @export
#' @rdname tk2dde
tk2dde.request <- function(service, topic, item, binary = FALSE) {
  # Get the value for 'item' in 'service|topic' dde server
  .tk2dde.require()

  if (!is.character(service) || !is.character(topic))
    stop("'service' and 'topic' must be character strings!")
  if (!is.character(item))
    stop("'item' must be character strings!")
  if (binary[1] == TRUE) binary <- "-binary" else binary <- ""

  # Request the value in a try(), to nicely catch the error
  # class is "try-error" if an error occurs, otherwise, returns ""
  res <- (try(as.character(.Tcl(paste("dde request ", binary, " ",
    as.character(service[1]), " ", as.character(topic[1]), " ",
    as.character(item[1]), sep = "")))))
  res
}

#' @export
#' @rdname tk2dde
tk2dde.services <- function(service = "", topic = "") {
  # List the 'service|topic' dde currently available
  .tk2dde.require()

  # Check arguments
  if (!is.character(service) || !is.character(topic))
    stop("'service' and 'topic' must be character strings!")
  service <- as.character(service[1])
  if (service == "")
    service <- "{}"  # This is an empty string in Tcl
  topic <- as.character(topic[1])
  if (topic == "")
    topic <- "{}"    # This is an empty string in Tcl

  # Get the list of all 'service|topic' dde servers currently running
  as.character(.Tcl(paste("dde services", service, topic)))
}

.tk2dde.require <- function() {
  if (.Platform$OS.type != "windows")
    stop("This is a Windows-specific function!")
  # Make sure tcl/tk dde is operational
  if (!capabilities("tcltk"))
    stop("This version of R cannot use Tcl/Tk!")
  res <- tclRequire("dde", warn = TRUE)
  if (inherits(res, "tclObj"))
    res <- tclvalue(res)
  if (res[1] == FALSE)
    stop("Unable to find the 'dde' Tcl/tk package!")
  res  # The package version number
}

# TODO:
# - use a try for tk2reg.get()
# - Change double call for a try() for tk2reg.keys(), tk2reg.type()
#   and tk2reg.values()
# - Add "none to the type of supported formats?

### TODO: implement ::ico::getIconByName, ::ico::getFileIcon & ::ico::writeIcon
### TODO: gif files are acceptable too for tk2ico.set(), example:
### Image <- tclVar()
### tcl("image", "create", "photo", Image, file = "myfile.gif")
### tcl("wm", "iconphoto", tt, Image) instead of tk2ico.set

#' Manipulate the registry under Windows
#'
#' These functions access the Windows registry in a secure way (most errors
#' are handled gracefully), and ensures correct conversion back and forth
#' for atomic strings ('sz' and 'expand\\\\_') and numbers ('dword' and
#' 'dword\\\\_big\\\\_endian'), and for vectors of strings ('multi\\\\_sz').
#'
#' @param keyname The name of the key.
#' @param valuename A value in this key.
#' @param data The data to place in this value.
#' @param type The type of value in the registry. By default, it is 'sz', that
#' is, an atomic string.
#'
#' @return
#' Functions that should return registry value(s) or key(s) return them in a
#' character string, or they return `NA` if the key/value is not found in the
#' registry.
#'
#' [tk2reg.broadcast()], [tk2reg.delete()], [tk2reg.deletekey()],
#' [tk2reg.set()] and [tk2reg.setkey()] return `TRUE` in case of success and
#' `FALSE` otherwise.
#'
#' [tk2reg.get()] should handle correctly the types 'sz', 'expand\\\\_sz' and
#' multi\\\\_sz' (note that 'expand\\\\_sz' string is NOT expanded!), as well as
#' dword' and 'dword\\\\_big\\\\_endian' that are converted into numeric values.
#' Other types are not converted and the Tcl expression is returned ('objTcl'
#' class) untransformed.
#'
#' [tk2reg.set()] currently works with 'sz', 'expand\\\\_sz', 'multi\\\\_sz',
#' dword' and 'dword\\\\_big\\\\_endian' types. A couple of other types are
#' accepted by the function... but they are not tested ('binary', 'link',
#' resource\\\\_list').
#'
#' @note
#' For Windows only. These functions issue an error when they are called
#' under other platforms. Take care while manipulating the Windows registry!
#' You can easily lock the system completely, if you delete important items,
#' especially if you are logged as administrator on your computer. Make a backup
#' of your registry first before experimenting with these function!!!
#'
#' @export
#' @rdname tk2reg
#' @author Philippe Grosjean
#' @seealso [tk2dde.exec()], [tk2ico.create()]
#' @keywords utilities
#'
#' @examples
#' \dontrun{
#' # These cannot be run by examples() but should be OK when pasted
#' # into an interactive R session with the tcltk package loaded
#'
#' ### Examples of tk2reg - registry manipulation under Windows
#' # Rem: HKEY_LOCAL_MACHINE, HKEY_USERS, HKEY_CLASSES_ROOT, HKEY_CURRENT_USER,
#' #      HKEY_CURRENT_CONFIG, HKEY_PERFORMANCE_DATA, HKEY_DYN_DATA
#' Rkey <- "HKEY_CURRENT_USER\\\\Software\\\\R-core\\\\R"   # The R key
#' Rkey <- paste(Rkey, "\\\\", R.version$major, ".", R.version$minor, sep = "")
#' Rsubkey <- paste(Rkey, "subkey", sep = "\\\\")        # A subkey
#'
#' # Get all subkeys for Software in the local machine
#' tk2reg.keys("HKEY_LOCAL_MACHINE\\\\Software")
#'
#' # Get all names in the R key
#' tk2reg.values(Rkey)
#'
#' # Get the path for the current R version
#' tk2reg.get(Rkey, "InstallPath")
#'
#' # Create a subkey (explore the registry with regedit.exe to see it)
#' tk2reg.setkey(Rsubkey)
#' # Add a couple of keys in it
#' tk2reg.set(Rsubkey, "test", "a key added in the registry!", type = "sz")
#' tk2reg.set(Rsubkey, "test exp", "\%SystemRoot\%\\\\system32", type = "expand_sz")
#' tk2reg.set(Rsubkey, "test multi", LETTERS[1:5], type = "multi_sz")
#' tk2reg.set(Rsubkey, "test dword", 1024, type = "dword")
#' tk2reg.set(Rsubkey, "test big end", 1024, type = "dword_big_endian")
#'
#' # Get the type of a value
#' tk2reg.type(Rsubkey, "test")
#' tk2reg.type(Rsubkey, "test exp")
#' tk2reg.type(Rsubkey, "test multi")
#' tk2reg.type(Rsubkey, "test dword")
#' tk2reg.type(Rsubkey, "test big end")
#'
#' # Get a value in a key
#' tk2reg.get(Rsubkey, "test")
#' tk2reg.get(Rsubkey, "test exp")
#' tk2reg.get(Rsubkey, "test multi")
#' tk2reg.get(Rsubkey, "test dword")
#' tk2reg.get(Rsubkey, "test big end")
#'
#' # Delete a name in a key (take care: dangerous!)
#' tk2reg.delete(Rsubkey, "test")
#' # Delete a whole key (take care: very dangerous!)
#' tk2reg.deletekey(Rsubkey)
#'
#' # An alternate way to get the path
#' tk2reg.get(paste("HKEY_LOCAL_MACHINE", "SYSTEM", "CurrentControlSet",
#'   "Control", "Session Manager", "Environment", sep = "\\\\"), "path")
#'
#' # Make sure that currently running apps are warned of your changes in the registry
#' tk2reg.broadcast()
#'
#' # Delete temporary variables
#' rm(list = c("Rkey", "Rsubkey"))
#' }
tk2reg.broadcast <- function() {
  # Used to warn running apps that something changes in the registry
  # Use this when you change an environment variable
  .tk2reg.require()
  res <- tclvalue(.Tcl("catch {registry broadcast \"Environment\"}"))
  (res == "0")  # "0" if OK, "1" otherwise
}

#' @export
#' @rdname tk2reg
tk2reg.delete <- function(keyname, valuename) {
  # Delete a registry value in a key (take care when using this!)
  .tk2reg.require()
  keyname <- as.character(keyname[1])
  valuename <- as.character(valuename[1])
  res <- tclvalue(.Tcl(paste("catch {registry delete {", keyname, "} {",
    valuename, "}}", sep = "")))  # return "0" if OK, "1" otherwise
  (res == "0")
}

#' @export
#' @rdname tk2reg
tk2reg.deletekey <- function(keyname) {
  # Completely delete a registry key (take care when using this!)
  .tk2reg.require()
  keyname <- as.character(keyname[1])
  res <- tclvalue(.Tcl(paste("catch {registry delete {", keyname, "}}",
    sep = "")))  # Return "0" if OK (even if already deleted) or "1"
  (res == "0")
}

#' @export
#' @rdname tk2reg
tk2reg.get <- function(keyname, valuename) {
  # Get the content of a key
  .tk2reg.require()
  keyname <- as.character(keyname[1])
  valuename <- as.character(valuename[1])
  # First get the type of this registry key
  Type <- tk2reg.type(keyname, valuename)
  if (is.na(Type))
    return(NA)  # The key does not exists
  # The key is found... retrieve its data
  res <- .Tcl(paste("registry get {", keyname, "} {",
    valuename, "}", sep = ""))
  # Convert according to its type...
  res <- switch(Type,
    sz = tclvalue(res),        # A single string
    expand_sz = tclvalue(res),    # This string is NOT expanded!
    multi_sz = as.character(res),  # A vector of strings
    dword = as.numeric(res),    # Numbers,... check very large numbers!
    dword_big_endian = as.numeric(res),  # Is this correct???
    res)  # Other types are probably not handled well!
  res
}

#' @export
#' @rdname tk2reg
tk2reg.keys <- function(keyname) {
  # Get a list of all subkeys in a key
  .tk2reg.require()
  keyname <- as.character(keyname[1])
  # First check if the command succeeds
  res <- tclvalue(.Tcl(paste("catch {registry keys {", keyname, "}}",
    sep = "")))  # Return "0" if OK, "1" otherwise
  if (res != "0")
    return(NA)  # Indicate that keyname is probably inexistant
  # Now run the command unprotected
  as.character(.Tcl(paste("registry keys {", keyname, "}", sep = "")))
}

#' @export
#' @rdname tk2reg
tk2reg.set <- function(keyname, valuename, data,
type = c("sz", "expand_sz", "multi_sz", "dword", "dword_big_endian")) {
  # Set a registry key value
  .tk2reg.require()
  keyname <- as.character(keyname[1])
  valuename <- as.character(valuename[1])
  data <- as.character(data)
  if (length(data) > 1)  # Collapse into one string, using {} as separators
    data <- paste(data, collapse = "\n")
  type <- type[1]
  if (!(type %in% c("sz", "expand_sz", "multi_sz", "dword", "dword_big_endian",
    "binary", "link", "resource_list", "none")))
    stop("Unrecognized 'type'!")
  res <- tclvalue(.Tcl(paste("catch {registry set {", keyname, "} {",
    valuename, "} {", data, "} {", type, "}}" , sep = "")))
  (res == "0")  # Because "0" if OK, and "1" otherwise
}

#' @export
#' @rdname tk2reg
tk2reg.setkey <- function(keyname) {
  # Set a registry key
  keyname <- as.character(keyname[1])
  .tk2reg.require()
  res <- tclvalue(.Tcl(paste("catch {registry set {", keyname, "}}",
    sep = "")))  # Return "0" if OK, "1" otherwise
  (res == "0")
}

#' @export
#' @rdname tk2reg
tk2reg.type <- function(keyname, valuename) {
  # Get the type of a key...
  .tk2reg.require()
  keyname <- as.character(keyname[1])
  valuename <- as.character(valuename[1])
  # First test it to see if the command succeeds (i.e., if the key exists)
  res <- tclvalue(.Tcl(paste("catch {registry type {", keyname, "} {",
    valuename, "}}", sep = "")))  # return "0" if OK, "1" otherwise
  if (res != "0")
    return(NA)  # The key is probably missing
  # Run the command unprotected now
  tclvalue(.Tcl(paste("registry type {", keyname, "} {", valuename, "}", sep = "")))
}

#' @export
#' @rdname tk2reg
tk2reg.values <- function(keyname) {
  # Get a list of all values in a key
  keyname <- as.character(keyname[1])
  .tk2reg.require()
  # First check if the command succeeds
  res <- tclvalue(.Tcl(paste("catch {registry values {", keyname, "}}",
    sep = "")))  # Returns "0" if OK, "1" otherwise
  if (res != "0")
    return(NA)  # The key probably does not exist!
  # We issue the command now without protection
  as.character(.Tcl(paste("registry values {", keyname, "}", sep = "")))
}

.tk2reg.require <- function() {
  # Make sure tcl/tk registry is operational
  if (.Platform$OS.type != "windows")
    stop("This is a Windows-specific function!")
  if (!capabilities("tcltk"))
    stop("This version of R cannot use Tcl/Tk!")
  # This should be installed by default with the tcltk package under Windows
  res <- tclRequire("registry", warn = TRUE)
  if (inherits(res, "tclObj"))
    res <- tclvalue(res)
  if (res[1] == FALSE)
    stop("Unable to find the 'registry' Tcl/tk package!")
  res  # The package version number
}

.onLoad <- function(libname, pkgname) {
  libdir <- file.path(libname, pkgname, "tklibs")

  # A slightly modified version of addTclPath() that works also within SciViews
  addTclPath <- function(path = ".") {
    if (.Platform$OS.type == "windows")
      path <- gsub("\\", "/", path, fixed = TRUE)
    # Modified by GregznaV (Pull Request #2)
    #a <- tclvalue(tcl("set", "::auto_path"))
    #paths <- strsplit(a, " ", fixed = TRUE)[[1L]]
    #paths <- as.character(tcl("set", "::auto_path"))
    paths <- as.character(tcl("set", "auto_path"))
    if (!path %in% paths) {
      tcl("lappend", "auto_path", path)
    } else {
      # Added by GregznaV (Pull Request #2)
      # To have a consistent output if the path is not changed:
      tcl("set", "auto_path")
    }
    invisible(paths)
  }
  res <- addTclPath(libdir)  # extend the Tcl/Tk path

  # Load Tcl and Tk translation catalogs
  res <- tclRequire("msgcat")
  if (inherits(res, "tclObj")) {
    .Tcl("namespace import msgcat::*")
    .Tcl("mcload [file join $::tcl_library msgs]")
    # In case there is no display available, this fails -> fail silently
    try(.Tcl("mcload [file join $::tk_library msgs]"), silent = TRUE)

    # Make sure that Tcl/Tk locale is the same one as current R locale
    lang <- getLanguage()
    if (lang != "") {  # Set the same language for Tcl/Tk
      try(setLanguage(lang), silent = TRUE)
    }
  }

  if (is.tk()) {
    # Here is how we could install the supplementary material in Tcl/Tk

    # This is for a better management of scrollbars in listbox, text, canvas
    suppressWarnings(try(tclRequire("autoscroll"), silent = TRUE)) # Version 1.1
    try(tcl("source", file.path(libdir, "scrolledWidget.tcl")), silent = TRUE)

    #tclRequire("choosefont")     # Version 0.2
    #tclRequire("ctext")          # Version 3.1
    #tclRequire("cursor")         # Version 0.3.1
    #tclRequire("mclistbox")      # Version 1.2
    #tclRequire("swaplist")       # Version 0.2
    #tclRequire("tablelist")      # Version 7.6
    #Not provided any more -> tclRequire("Tktable")       # Version 2.9

    # The following code is not implemented as Tcl package... just source it
    try(tcl("source", file.path(libdir, "notebook1.3", "notebook.tcl")), silent = TRUE)
    try(tcl("source", file.path(libdir, "tree1.7", "tree.tcl")), silent = TRUE)

    # Do we try to load the tile widgets? (only if Tcl./Tk < 8.5)
    if (as.numeric(tclvalue("::tcl_version")) < 8.5) {
###      tcl("source", file.path(libdir, "fonts.tcl"))
      # Define fonts used in Tk (note: must be done AFTER loading tile!)
      # Default values for system fonts are calculated by tile...
      # but they should be computed from the system, actually
      # We collect back those values calculated by tile and possibly override
      # them with better values
###      tk2font.setstyle(system = TRUE, default.styles = TRUE, text = TRUE)
### TODO: reflect possible changes to other graphical toolkits (how?)
    } else {# There is a bug in mclistbox with Tcl/Tk 8.5
      # Patch by Christiane Raemsch, slightly modified by Ph. Grosjean
      # This is essentially the listbox procedure, but with an additional
      # focus argument required by mclistbox
      .Tcl('proc ::tk::ListboxBeginSelect {w el {focus 0}} {
        variable ::tk::Priv
        if {[$w cget -selectmode] eq "multiple"} {
          if {[$w selection includes $el]} {
            $w selection clear $el
          } else {
            $w selection set $el
          }
        } else {
          $w selection clear 0 end
          $w selection set $el
          $w selection anchor $el
          set Priv(listboxSelection) {}
          set Priv(listboxPrev) $el
        }
        event generate $w <<ListboxSelect>>
        if {$focus && [winfo exists $w]} {
          focus $w
        }
      }')
    }

    # Load additional ttk themes - No: load only on demand!
    # Not done any more on startup, done on demand in tk2theme() now
    #try(tclRequire("ttk::theme::plastik"), silent = TRUE)
    #try(tclRequire("ttk::theme::keramik"), silent = TRUE)
    #try(tclRequire("ttk::theme::keramik_alt"), silent = TRUE)
    #try(tclRequire("ttk::theme::clearlooks"), silent = TRUE)
    #try(tclRequire("ttk::theme::radiance"), silent = TRUE)

    # Which ttk theme should we use?
    # If the user specified a default theme, use it
    if (!.loadTheme()) {
      # ...otherwise, try to guess the best default value
      themes <- try(tk2theme.list(), silent = TRUE)
      if (!inherits(themes, "try-error")) {
        if ("aqua" %in% themes) { # This must be aquaTk on a Mac
          try(tk2theme("aqua"), silent = TRUE)
        } else if ("vista" %in% themes) { # This must be Vista or Win 7
          try(tk2theme("vista"), silent = TRUE)
        } else if ("xpnative" %in% themes) { # This must be XP
          try(tk2theme("xpnative"), silent = TRUE)
        } else if ("winnative" %in% themes) { # This must be a pre-XP windows
          try(tk2theme("winnative"), silent = TRUE)
        } else if (.isUbuntu()) {
          #try(tk2theme("radiance"), silent = TRUE)
          # We also load clearlooks by default in Ubuntu
          try(tk2theme("clearlooks"), silent = TRUE)
          # Special treatment for Ubuntu: change fonts to Ubuntu and Ubuntu mono
          # and use white text on black for tooltips

          # Again, Tk 8.5/8.6 does a better job by default now than 8.4
          # So, we don't need this any more!?
          #tkfont.configure("TkDefaultFont", family = "Ubuntu", size = 11)
          #tkfont.configure("TkMenuFont", family = "Ubuntu", size = 11)
          #tkfont.configure("TkCaptionFont", family = "Ubuntu", size = 10)
          #tkfont.configure("TkSmallCaptionFont", family = "Ubuntu", size = 9)
          #tkfont.configure("TkTooltipFont", family = "Ubuntu", size = 9)
          #tkfont.configure("TkMenuFont", family = "Ubuntu", size = 11)
          #tkfont.configure("TkHeadingFont", family = "Ubuntu", size = 12)
          #tkfont.configure("TkIconFont", family = "Ubuntu", size = 11)
          #tkfont.configure("TkTextFont", family = "Ubuntu", size = 11)
          #tkfont.configure("TkFixedFont", family = "Ubuntu Mono", size = 11)
          res <- tclRequire("tooltip")
          if (inherits(res, "tclObj")) {
            .Tcl(paste("set ::tooltip::labelOpts [list -highlightthickness 0",
              "-relief solid -bd 1 -background black -fg white]"))
          }
        } else {# A modern "default" theme that fits not too bad in many situations
          suppressWarnings(try(tk2theme("clearlooks"), silent = TRUE))
        }
      }
   }
    # Save default font as TkSysDefaultFont
    tk2font.set("TkSysDefaultFont", tk2font.get("TkDefaultFont"))
  }

  # Windows only
  if (.Platform$OS.type == "windows") {
    try(tclRequire("dde"), silent = TRUE) # Version 1.2.2
    # Not loaded automatically!
    #tclRequire("registry")  # Version 1.1.3
    # Support for winico.dll is drop from version 1.2-1!
    #    if (nzchar(r_arch <- .Platform$r_arch))
    #    tcl("load", file.path(libname, pkgname, "libs", r_arch, "Winico06.dll"))
    #  else
    #    tcl("load", file.path(libname, pkgname, "libs", "Winico06.dll"))
    # Also register the DDE server as TclEval|R
    try(tk2dde("R"), silent = TRUE)
  }
}

.onUnload <- function(libpath) {
  # Remove all currently scheduled tasks
  tclTaskDelete(id = NULL)
}

.saveTheme <- function()
  cat(tk2theme(), "\n", sep = "", file = "~/.Rtk2theme")

.loadTheme <- function() {
  if (file.exists("~/.Rtk2theme")) {
    theme <- try(readLines("~/.Rtk2theme")[1], silent = TRUE)
    if (inherits(theme, 'try-error')) return(FALSE)
    # Try changing the tk2theme according to this value
    res <- try(tk2theme(theme), silent = TRUE)
    !inherits(res, "try-error")
  } else FALSE
}

.isUbuntu <- function() {
  # Note: take care not to call 'cat' on Windows: it is usually *not* there!
  if (.Platform$OS.type == "windows" || grepl("^mac", .Platform$pkgType))
    return(FALSE)  # This is either Windows or Mac OS X!
  # On Ubuntu, there is an lsb-release file, but read it just to make sure
  file.exists("/etc/lsb-release") &&
    any(grepl("[Uu]buntu", readLines("/etc/lsb-release")))
}

.mergeList <- function(l1, l2) {
  # For named lists, overwrite items of l1 present in l2
  nms <- names(l2)
  # Deal with named items
  if (length(nms)) {
    named <- nms != ""
    if (any(named)) {
      l2n <- l2[named]
      nmsn <- nms[named]
      for (i in 1:length(nmsn)) l1[[nmsn[i]]] <- l2n[[nmsn[i]]]
    }
    # Keep only non named items in l2
    l2 <- l2[!named]
  }
  # Deal with non named items in l2
  if (length(l2)) { # Unnamed list
    n1 <- length(l1)
    n2 <- length(l2)
    for (i in 1:n2) l1[[n1 + i]] <- l2[[i]]
  }
  l1
}

.configStd <- function(x, lstval) {
  # These config parameters are considered as data
  # Image
  if (!is.null(lstval$image)) {
    tkconfigure(x, image = lstval$image)
    lstval$image <- NULL
  }
  # Text
  if (!is.null(lstval$text)) {
    tkconfigure(x, text = lstval$text)
    lstval$text <- NULL
  }
  # Textvariable
  if (!is.null(lstval$textvariable)) {
    tkconfigure(x, textvariable = lstval$textvariable)
    lstval$textvariable <- NULL
  }
  # Values
  if (!is.null(lstval$values)) {
    values(x) <- lstval$values
    lstval$values <- NULL
  }
  # Value
  if (!is.null(lstval$value)) {
    value(x) <- lstval$value
    lstval$value <- NULL
  }
  # Selection
  if (!is.null(lstval$selection)) {
    selection(x) <- lstval$selection
    lstval$selection <- NULL
  }
  # Label (not a Tk attribute)
  if (!is.null(lstval$label)) {
    label(x) <- lstval$label
    lstval$label <- NULL
  }
  # Tag (not a Tk attribute)
  if (!is.null(lstval$name)) {
    tag(x) <- lstval$tag
    lstval$tag <- NULL
  }
  # Tooltip
  if (!is.null(lstval$tip)) {
    tip(x) <- lstval$tip
    lstval$tip <- NULL
  }
  # Disabled (is tk 'state' parameter indeed)
  if (!is.null(lstval$disabled)) {
    disabled(x) <- lstval$disabled
    lstval$disabled <- NULL
  }
  # Return modified value list
  lstval
}

.wraplength <- function(w, width) {
  # Calculate wraplength required for tk2label widgets
  # width is expressed in characters, but wraplength must be given in pixels
  # This is stupid and requires additional computation to calculate the
  # width in pixel of an average character, like "0" to do the conversion!
  # Get the average size of one character in the current font used

  # If width is not set, just return a large value for wraplength
  if (!length(width)) return(1000)

  # Get the font and measure it
  font <- tclvalue(tkcget(w, "-font"))
  if (font == "") font <- tk2style("tk2label", "font")
  if (font == "") {
    charsize <- 8 # Use an everage value
  } else charsize <- as.numeric(tkfont.measure(tkfont.actual(font), "0"))

  # Optimal wraplength is width * charsize
  width * charsize
}

.TempEnv <- function() {
  pos <-  match("SciViews:TempEnv", search())
  if (is.na(pos)) {  # Must create it
    `SciViews:TempEnv` <- list()
    Attach <- function(...) get("attach", mode = "function")(...)
    Attach(`SciViews:TempEnv`, pos = length(search()) - 1)
    rm(`SciViews:TempEnv`)
    pos <- match("SciViews:TempEnv", search())
  }
  pos.to.env(pos)
}

.assignTemp <- function(x, value, replace.existing = TRUE)
  if (replace.existing || !exists(x, envir = .TempEnv(), mode = "any",
    inherits = FALSE))
    assign(x, value, envir = .TempEnv())

.getTemp <- function(x, default = NULL, mode = "any", item = NULL) {
  if (is.null(item)) Mode <- mode else Mode <- "any"
  if  (exists(x, envir = .TempEnv(), mode = Mode, inherits = FALSE)) {
    dat <- get(x, envir = .TempEnv(), mode = Mode, inherits = FALSE)
    if (is.null(item)) return(dat) else {
      item <- as.character(item)[1]
      if (inherits(dat, "list") && item %in% names(dat)) {
        dat <- dat[[item]]
        if (mode != "any" && mode(dat) != mode) dat <- default
          return(dat)
        } else {
          return(default)
        }
    }
  } else {# Variable not found, return the default value
    return(default)
  }
}

# TODO:
# - Add a catch {} in tclFun and handle it
# - A tclFunDispose() function to delete the Tcl equivalent of a function
# - Add a try construct in tclVarExists and tclVarFind
# - better manage catch{} in tclVarName


#' Manipulate R variables and functions from tcl and back
#'
#' These functions are intended to provide a better "duality" between the name
#' of variables in both R and tcl, including for function calls. It is possible
#' to define a variable with the same name in R and tcl (the content is
#' identical, but copied and coerced in the two respective environments). It is
#' also possible to get the value of a tcl variable from R, and to call a R
#' function from within tcl. These features are provided in the tcltk package,
#' but Tcl variable usually have different internal names as R equivalents.
#'
#' @param names Transform names so that they are valid for variables in Tcl.
#' @param unique Should these names be unique in the vector?
#' @param f An R function. currently, do no support functions with arguments.
#' @param name The name of a variable.
#' @param value The value to place in a variable.
#' @param pattern A pattern to search for.
#' @param init Initial value to use when creating the variable.
#' @param keep.existing If the tcl variable already exist, should we keep its
#' content?
#'
#' @details
#' These functions are similar to [tcltk::tclVar()] from package tcltk, except
#' for the following change: here, it is possible to propose a name for the
#' created tcl variable, or to set or retrieve the content of a tcl variable
#' that is not mirrored in R.
#'
#' @return
#' Most of these functions return a 'tclVar' object.
#' @author Philippe Grosjean
#' @seealso [tk2edit()], [tcltk::tclVar()]
#' @keywords utilities
#' @export
#' @rdname tclVarFun
#'
#' @examples
#' \dontrun{
#' # These cannot be run by examples() but should be OK when pasted
#' # into an interactive R session with the tcltk package loaded
#'
#' # Tcl functions and variables manipulation
#' tclVarExists("tcl_version")
#' tclVarExists("probably_non_existant")
#' tclVarFind("tcl*")
#'
#' # Using tclVarName() and tclGetValue()...
#' # intented for better match between R and Tcl variables
#' Test <- tclVarName("Test", "this is a test!")
#' # Now 'Test' exist both in R and in Tcl... In R, you need to use
#' tclvalue(Test) # to retrieve its content
#' # If a variable already exists in Tcl, its content is preserved using
#' # keep.existing = TRUE
#'
#' # Create a variable in Tcl and assign "just a test..." to it
#' tclSetValue("A_Variable", "just to test...")
#' # Create the dual variable with same name
#' A_Variable <- tclVarName("A_Variable", "something else?")
#' tclvalue(A_Variable) # Content of the variable is not changed!
#'
#' # If you want to retrieve the content of a Tcl variable,
#' # but do not want to create a reference to it in R, use:
#'
#4 # Create a Tcl variable, not visible from R
#' tclSetValue("Another_Variable", 1:5)
#' tclGetValue("Another_Variable") # Get its content in R (no conversion!)
#' tclSetValue("Another_Variable", paste("Am I", c("happy", "sad"), "?"))
#' tclGetValue("Another_Variable") # Get its content in R (no conversion!)
#' }
makeTclNames <- function(names, unique = FALSE) {
  # Make valid Tcl variable names (allow_ = TRUE by default in R >= 2.0.0)
  names <- make.names(names, unique = unique)
  # There is a problem if the variable starts with a dot => prepend it with 'X'
  .names <- grep("^\\.", names)
  names[.names] <- paste("X", names[.names], sep = "")
  # Although it is accepted, there could be problems with variable names
  # containing dots, so, replace them with '_'
  gsub("\\.", "_", names)
}

### TODO: change this to use closure functions instead!!!
#' @export
#' @rdname tclVarFun
tclFun <- function(f, name = deparse(substitute(f))) {
  # Register a simple R function (without arguments) as a callback in Tcl,
  # and give it the same name under Tcl)
  # Indeed, .Tcl.callback(f) does the job... but it gives criptic names
  # like R_call 0x13c7168

  # Check that 'f' is a function with no arguments (cannot handle them yet)
  if (!is.function(f)) stop("'f' must be a function!")
  if (!is.null(formals(f))) stop("The function used cannot (yet) have arguments!")
  # Make sure the name of the function is valid
  if (!is.character(name)) stop("'name' must be a character string!") else
  name <- make.names(name[1])

  res <- .Tcl.callback(f)
  # Make sure this is correct (R_call XXXXXXXX)
  if (length(grep("R_call ", res) > 0)) {
    # Create a proc with the same name in Tcl
    .Tcl(paste("proc ", name, " {} {", res, "}", sep = ""))
  }
  # Rem: if you delete the R 'f' function, the Tcl 'f' function still works!
  # You have to explicitly delete the Tcl function

  # Return the R_call XXXXXXXX string, as .Tcl.callback() does
  res
}

#' @export
#' @rdname tclVarFun
tclGetValue <- function(name) {
  # Get the value stored in a plain Tcl variable
  if (!is.character(name))
    stop("'name' must be a character!")
  name <- makeTclNames(name[1]) # The usual name conversion

  # Create a temporary dual variable with tclVar() (name does not mather)
  Temp <- tclVar(init = "")

  # Copy the content of the var of interest to it
  res <- tclvalue(.Tcl(paste("catch {set ", as.character(Temp), " $", name,
    "}", sep = "")))  # Return "0" if OK, "1" otherwise
  if (res != "0")
    stop(gettextf("Error when getting the value in the '%s' Tcl variable",
      name))

  # Get the content of the temporary variable
  tclvalue(Temp) # (Temp will be destroyed when the function exits)
}

#' @export
#' @rdname tclVarFun
tclSetValue <- function(name, value) {
  # This is the opposite of tclGetValue() and it is a wrapper
  # for 'set name value' Tcl command
  if (!is.character(name))
    stop("'name' must be a character!")
  name <- makeTclNames(name[1]) # The usual name conversion

  # Create a temporary dual variable with tclVar() (name does not mather)
  Temp <- tclVar(init = value)

  # Copy the content of this variable to the tcl variable 'name'
  res <- tclvalue(.Tcl(paste("catch {set ", name, " $", as.character(Temp),
    "}", sep = "")))
  if (res != "0")
    stop(gettextf("Error when changing the value of the '%s' Tcl variable",
      name))

  # (Temp is destroyed when the function exits)
  invisible(name)  # Return the name of the Tcl variable invisibly
}

#' @export
#' @rdname tclVarFun
tclVarExists <- function(name)
  as.integer(tcl("info", "exists", name)) == 1

#' @export
#' @rdname tclVarFun
tclVarFind <- function(pattern)
  as.character(tcl("info", "vars", pattern))

#' @export
#' @rdname tclVarFun
tclVarName <- function(name, init = "", keep.existing = TRUE) {
  # tclVar gives names like ::RtclX automatically...
  # We need to define names ourselve. This is what tclVarName does
  # If keep existing == TRUE and the variable is already defined, then
  # we keep its content, instead of initializing it with "init"
  if (!is.character(name))
    stop("'name' must be a character!")
  name <- makeTclNames(name[1])  # Make sure the name is correct

  # Temporary save potential content of the Tcl variable elsewhere
  # (catch in case the variable does not exist)
  if (isTRUE(keep.existing))
    .Tcl(paste("catch {set ZZZTempRvariable $", name, "}", sep = ""))

  # Create the new dual Tcl-R variable
  l <- list(env = new.env())
  assign(name, NULL, envir = l$env)
  reg.finalizer(l$env, function(env) tcl("unset", ls(env)))
  class(l) <- "tclVar"
  tclvalue(l) <- init

  # Possibly restore the content of the variable, if keep.existing == TRUE
  if (isTRUE(keep.existing)) {
    .Tcl(paste("catch {set", name, "$ZZZTempRvariable}"))
    # Remove the temporary variable
    .Tcl("unset -nocomplain ZZZTempRvariable")
  }
  l
}

#' Additional Tk dialog boxes
#'
#' Tk dialog boxes to select a font, unicode characters or a list of ordered
#' items.
#'
#' @param ... Further arguments passed to the dialog box.
#' @param parent The Tk toplevel dialog box that will be the parent of the
#' configuration dialog box.
#' @param widget A widget that can accept a unicode character. For
#' `tk2unicode_bind()` it must be a `tk2text` or a `tk2entry` widget.
#'
#' @return
#' The selection made in the dialog box if `OK` is clicked, `""` otherwise for
#' [tk2chooseFont()].
#'
#' The [tk2unicode_select()] dialog pastes the selected unicode character in the
#' designed widget, but returns nothing. The [tk2unicode_config()] changes the
#' configuration for the unicode composer, but returns nothing. If you decide to
#' do so, it saves the config on a file. This is done app-by-app, and the
#' default app name is `"R"`. You can change it by setting a different
#' value in the option `"tk2app"`, i.e., `options(tk2app = "myApp")`.
#' The `tk2unicode_bind()` is also invoked for its side-effect to install
#' required bindings to enable the unicode composer engine for the given widget
#' and it returns nothing.
#'
#' @export
#' @rdname tk2dialogfonts
#' @author Philippe Grosjean
#' @seealso [tk2text()], [tk2listbox()], [tk2list.insert()]
#' @keywords utilities
#'
#' @examples
#' \dontrun{
#' library(tcltk2)
#' # These cannot be run by examples() but should be OK when pasted
#' # into an interactive R session with the tcltk package loaded
#'
#' # Font selection
#' tk2chooseFont()
#' tk2chooseFont(font = "{courier} 9", title = "Choose a fixed font",
#'   fonttype = "fixed", style = 4, sizetype = "all")
#' tk2chooseFont(font = "Verdana 12 bold italic underline overstrike",
#'   fonttype = "prop", style = 2, sizetype = "point")
#'
#' # Easy unicode character entry
#' tt <- tktoplevel()
#' txt <- tk2text(tt, width = 60, height = 20)
#' tkpack(txt)
#' e <- tk2entry(tt, width = 50)
#' tkpack(e)
#' # Get an unicode character for the text widget
#' tk2unicode_select(txt)
#' # and for the entry widget
#' tk2unicode_select(e)
#'
#' # Bind the composer to both the text and the entry widgets
#' # and display the configuration box
#' # Once done, try the compose key + m + u, or compose + " + a
#' # or any othert sequence in both widgets
#' # or hit the compose key twice
#' tk2unicode_bind(txt)
#' tk2unicode_bind(e)
#' tk2unicode_config(tt)
#' }
tk2chooseFont <- function(...) {
  if (!is.tk())
    stop("Package Tk is required but not loaded")
  tclRequire("choosefont")
  # Make sure message translations are correctly loaded
  try(tcl("mcload", system.file("tklibs", "choosefile", "msgs",
    package = "tcltk2")), silent = TRUE)
  tcl("choosefont::choosefont", ...)
}

# Unicode character input
.tk2unicode_file <- function(app = getOption("tk2app", "R"))
  file.path("~", paste0(".khimrc.", as.character(app)[1]))

.tk2unicode_load <- function() {
  # Try to get current configuration
  cfg <- try(tcl("::khim::getConfig"), silent = TRUE)
  if (inherits(cfg, "try-error")) {
    # Try loading the khim package
    res <- tclRequire("khim")
    if (!inherits(res, "tclObj"))
      return()
    # If a config file exists, load it now
    cfgfile <- .tk2unicode_file()
    if (file.exists(cfgfile))
      tcl("source", cfgfile)
    # finally get the updated config
    cfg <- tcl("::khim::getConfig")
  }
  # Make sure message translations are correctly loaded
  try(tcl("mcload", system.file("tklibs", "khim", "msgs",
    package = "tcltk2")), silent = TRUE)
  tclvalue(cfg)
}

#' @export
#' @rdname tk2dialogfonts
tk2unicode_config <- function(parent) {
  if (!inherits(parent, "tkwin"))
    stop("'parent' must be a 'tkwin' object")

  # Make sure khim is loaded and get its current config
  cfg <- .tk2unicode_load()

  # Display the configuration dialog box
  .Tcl(paste0("::khim::getOptions ", parent$ID, ".khim"))

  # Get the new config and compare it with the old one
  cfg2 <- tclvalue(tcl("::khim::getConfig"))
  if (cfg2 != cfg) {
    # Ask to save the new config
    msg <- tclmc("Do you want to save this configuration on disk?",
      domain = "khim")
    res <- tkmessageBox(
      message = msg, icon = "question", type = "yesno")
    if (tclvalue(res) == "yes") {
      cfgfile <- .tk2unicode_file()
      cat(cfg2, file = cfgfile)
    }
  }
}

#' @export
#' @rdname tk2dialogfonts
tk2unicode_select <- function(widget) {
  .tk2unicode_load()
  tcl("::khim::FocusAndInsertSymbol", widget$ID)
}

#' @export
#' @rdname tk2dialogfonts
tk2unicode_bind <- function(widget) {
  if (!inherits(widget, c("tk2text", "tk2entry")))
    stop("You can bind the unicode composer to tk2text() or tk2entry() widgets only")
  # Make sure evertything is loaded and configured correctly
  .tk2unicode_load()
  # Create the binding
  if (inherits(widget, "tk2text")) {
    tkbindtags(widget, paste0(widget$ID , " KHIM Text ",
      widget$env$parent$ID, " all"))
  } else {# This must be a tk2entry widget
    tkbindtags(widget, paste0(widget$ID , " KHIM Entry ",
      widget$env$parent$ID, " all"))
  }
}

#' @details
#' The \{tcltk2\} package provides additions to the \{tcltk\} base package.
#'
#' Many Tk widgets are added. Also, features of the Tcl language that can be
#' useful also in R are made accessible through R functions.
#'
#' @keywords internal
"_PACKAGE"

#' @importFrom tcltk tcl .Tcl tclRequire tclObj as.tclObj tclVar tclArray tclvalue tclvalue<- tclServiceMode .Tcl.callback tktoplevel tkwidget tkfont.names tkfont.configure tkfont.families tkfont.measure tkfont.actual tkconfigure tkcget tkcurselection tkselection.clear tkselection.set tksee tkset tksize tkwm.title tkxview tkyview tkindex tkdestroy tkinsert tkactivate tkbutton tkgrid tklabel tkgrid.configure tkgrid.columnconfigure tkgrid.rowconfigure tktag.configure tkgrab.set tkfocus tkwait.window tkwm.iconbitmap tkbind tkadd tkmessageBox tkscrollbar tkdelete tkselect tkbindtags tkget
# The following block is used by usethis to automatically manage
# roxygen namespace tags. Modify with care!
## usethis namespace: start
## usethis namespace: end
NULL

#' Manipulate Tk fonts
#'
#' Get or set fonts used by Tk widgets from within R.
#'
#' @param font The name of one or several cached Tk font.
#' @param what A list of font characteristics to get: 'family', 'size', 'bold',
#' italic', 'underline' and/or 'overstrike'. By default, everything except
#' underline' and 'overstrike'.
#' @param settings Settings of fonts. There are two possible forms: (1) a vector
#' of character strings of same length as font with Tk fonts description like
#' -family Times -size 12 -weight bold', for instance, or (2) a list of
#' font characteristics (list with components 'family', 'size', 'bold', 'italic',
#' 'underline' and 'overstrike').
#' @param text Do we synchronise text Tk fonts (text, titles, and fixed-font
#' text) with current settings in `.Fonts` inside the `SciViews:TempEnv`
#' environment?
#' @param system Do we synchronise system Tk fonts (widgets, window caption,
#' menus, tooltips, ...) with current system configuration? This is highly
#' platform dependent. Currently, system settings are gathered only under
#' Windows, thanks to the `winSystemFonts()` function.
#' @param default.styles Do we add `.fontsStyleXXX` in the `SciViews:TempEnv`
#' environment, where `XXX` is one of the four default styles: 'Classic',
#' 'Alternate', 'Presentation' or 'Fancy'.
#'
#' @return
#' [tk2font.get()] retrieves a list with font characteristics (same format
#' as the `settings =` argument) for the first Tk font found in its `font =`
#' argument, or `""` if the font is not found. [tk2font.set()] changes current
#' font settings or, possibly, create the Tk font.
#' [tk2font.setstyle()] changes the current Tk fonts settings according to
#' actual system and/or text configuration fonts.
#'
#' @export
#' @rdname tk2fonts
#' @author Philippe Grosjean
#' @seealso [tk2chooseFont()]
#' @keywords utilities
#'
#' @examples
#' \dontrun{
#' # These cannot be run by examples() but should be OK when pasted
#' # into an interactive R session with the tcltk package loaded
#' # Refresh both text and system Tk fonts
#' tk2font.setstyle(system = TRUE, default.styles = TRUE)
#' # Get characteristics of the default font
#' tk2font.get("TkDefaultFont")
#' }
tk2font.get <- function(font, what = c("family", "size", "bold", "italic")) {
  # font is the TkFont name to use, in case of several items, other ones
  # are secondary, tertiary, ... options
  # what indicate what characteristic of the font to report in the list
  # 'family', 'size', 'bold', 'italic', 'underline', 'overstrike' (last two rarely used)
  if (!is.tk())
    return("")
  allTkFonts <- as.character(tkfont.names())
  for (fnt in font) {
    if (fnt %in% allTkFonts)
      break
  }
  if (!fnt %in% allTkFonts) {
    return("")
    #if (length(font) == 1) {
    #  stop("'", font, "' is not currently defined in Tk")
    #} else {
    #  stop("'", paste(font, collapse = "', '"), "' are not currently defined in Tk")
    #}
  }
  fontspec <- as.character(tkfont.configure(fnt))
  res <- list()
  if (length(fontspec) != 12)
    return(res)  # There is a problem here!
  if ("family" %in% what)
    res$family <- fontspec[2]
  if ("size" %in% what)
    res$size <- as.numeric(fontspec[4])
  if ("bold" %in% what)
    res$bold <- (fontspec[6] == "bold")
  if ("italic" %in% what)
    res$italic <- (fontspec[8] == "italic")
  if ("underline" %in% what)
    res$underline <- (fontspec[10] == "1")
  if ("overstrike" %in% what)
    res$overstrike <- (fontspec[12] == "1")
  res
}

#' @export
#' @rdname tk2fonts
tk2font.set  <- function(font, settings) {
  ### TODO: allow for multiple fonts specifications => take first one available
  # font is the name of the TkFont to create/change
  # settings is a list with font characteristics
  if (!is.tk())
    return(NULL)
  font <- as.character(font)
  l <- length(font)
  if (!is.list(settings) && !is.character(settings))
    stop("'settings' must be a list or a character string")
  # If settings is a character string,
  # it is assumed to be a text description of a Tk font
  if (is.character(settings)) {
    # Do not recycle... make sure that lengths match
    if (length(settings) != l)
      stop("length of 'font' and 'settings' do not match")
    for (i in 1:l) {
      .Tcl(paste("catch {font create ", font[i], "}", sep = ""))
      .Tcl(paste("catch {font configure ", font[i], " ", settings[i], "}", sep = ""))
    }
  } else {# This is a list of font characteristics
    # Do not recycle... make sure that lengths match
    if (l > 1) {
      if (length(settings) != l)
        stop("length of 'font' and 'settings' do not match")
    } else {# Is it the list of characteristics, or a lit containing it?
      if (any(names(settings) %in%
        c("family", "size", "bold", "italic", "underline", "overstrike")))
        settings <- list(settings)
    }
    fntfamilies <- as.character(tkfont.families())
    for (i in 1:l) {
      # Construct the font descriptor
      fntl <- as.list(settings[[i]])
      fnt <- " "
      if (!is.null(fntl$family)) {
        # Look for the first font family provided that is available
        fntfamily <- fntl$family
        if (length(fntfamily) > 1) {
          fntexists <- fntfamily %in% fntfamilies
          if (any(fntexists)) {
            fntfamily <- fntfamily[fntexists][1]
          } else {
            fntfamily <- fntfamily[1]  # No fonts found... take first one
          }
        }
        fnt <- paste(fnt, "-family {", fntfamily, "}", sep = "")
      }
      if (!is.null(fntl$size))
        fnt <- paste(fnt, "-size", fntl$size)
      if (!is.null(fntl$bold)) {
        value <- if (fntl$bold == TRUE) "bold" else "normal"
        fnt <- paste(fnt, "-weight", value)
      }
      if (!is.null(fntl$italic)) {
        value <- if (fntl$italic == TRUE) "italic" else "roman"
        fnt <- paste(fnt, "-slant", value)
      }
      if (!is.null(fntl$underline))
        fnt <- paste(fnt, "-underline",
          as.numeric(fntl$underline == TRUE))
      if (!is.null(fntl$overstrike))
        fnt <- paste(fnt, "-overstrike",
          as.numeric(fntl$overstrike == TRUE))
      # Possibly create the font in Tk
      .Tcl(paste("catch {font create ", font[i], "}", sep = ""))
      if (fnt != " ")
        .Tcl(paste("catch {font configure ", font[i], fnt, "}", sep = ""))
    }
  }
  res <- font %in% as.character(tkfont.names())
  names(res) <- font
  res
}

#' @export
#' @rdname tk2fonts
tk2font.setstyle <- function(text = TRUE, system = FALSE, default.styles = FALSE) {
  # Set default fonts according to currently defined style
  # .SystemFonts and .Fonts must be defined in SciViews:TempEnv!

  if (!is.tk()) {
    warning("Package Tk is required but not loaded")
    return(NULL)
  }

  # This is a copy of assignTemp(), getTemp() and existsTemp() functions from
  # svMisc, so that we do not link to this package
  TempEnv <- function() {
    pos <-  match("SciViews:TempEnv", search())
    if (is.na(pos)) { # Must create it
      `SciViews:TempEnv` <- list()
      Attach <- function(...)
        get("attach", mode = "function")(...)
      Attach(`SciViews:TempEnv`, pos = length(search()) - 1)
      rm(`SciViews:TempEnv`)
      pos <- match("SciViews:TempEnv", search())
    }
    pos.to.env(pos)
  }

  assignTemp <- function(x, value, replace.existing = TRUE)
    if (replace.existing || !exists(x, envir = TempEnv(), mode = "any", inherits = FALSE))
      assign(x, value, envir = TempEnv())

  existsTemp <- function(x, mode = "any")
    exists(x, envir = TempEnv(), mode = mode, inherits = FALSE)

  getTemp <- function(x, default = NULL, mode="any") {
    if  (exists(x, envir = TempEnv(), mode = mode, inherits = FALSE)) {
        return(get(x, envir = TempEnv(), mode = mode, inherits = FALSE))
    } else {# Variable not found, return the default value
        return(default)
    }
  }

  if (system) {# Set system fonts
    # We collect back system fonts settings (other values may be imposed by Tk)
    defaultclassic <- tk2font.get("TkClassicDefaultFont")
    if (defaultclassic == "")
      defaultclassic <- tk2font.get("TkDefaultFont")
    sysfonts <- list(
      defaultclassic = defaultclassic,
      default = tk2font.get("TkDefaultFont"),
      caption = tk2font.get("TkCaptionFont"),
      smallcaption = tk2font.get(c("TkSmallCaptionFont", "TkCaptionFont")),
      menu = tk2font.get(c("TkMenuFont", "TkDefaultFont")),
      status = tk2font.get(c("TkStatusFont", "TkTooltipFont")),
      tooltip = tk2font.get("TkTooltipFont"),
      heading = tk2font.get("TkHeadingFont"),
      icon = tk2font.get(c("TkIconFont", "TkDefaultFont"))
    )
    # Make sure these are correctly defined
    assignTemp(".SystemFonts", sysfonts)
    res <- TRUE
  } else res <- character(0)

  if (default.styles) {# Define default styles
    # These are the four default Font themes one can use
    assignTemp(".FontsStyleClassic", list(
      Text = list(family = c("Times New Roman", "Times"), size = -12),
      Title = list(family = c("Arial", "Helvetica"), size = -14, bold = TRUE),
      BigTitle = list(family = c("Arial", "Helvetica"), size = -16, bold = TRUE),
      Fixed = list(family = c("Courier New", "Courier"), size = -12)
    ))

    assignTemp(".FontsStyleAlternate", list(
      Text = list(family = "Georgia", alt.family = "Times", size = -12),
      Title = list(family = c("Trebuchet MS", "Trebuchet"),
        alt.family = "Helvetica", size = -14, bold = TRUE),
      BigTitle = list(family = c("Trebuchet MS", "Trebuchet"),
        alt.family = "Helvetica", size = -16, bold = TRUE),
      Fixed = list(family = "Andale Mono", alt.family = "Courier", size = -12)
    ))

    assignTemp(".FontsStylePresentation", list(
      Text = list(family = "Verdana", alt.family = "Helvetica", size = -12),
      Title = list(family = "Verdana", alt.family = "Helvetica", size = -14,
        bold = TRUE),
      BigTitle = list(family = "Verdana", alt.family = "Helvetica", size = -16,
        bold = TRUE),
      Fixed = list(family = "Lucida Console", alt.family = "Courier",
        size = -12)
    ))

    assignTemp(".FontsStyleFancy", list(
      Text = list(family = c("Trebuchet MS", "Trebuchet"),
        alt.family = "Helvetica", size = -12),
      Title = list(family = c("Comic Sans MS", "Comic Sans"),
        alt.family = "Helvetica", size = -14, bold = TRUE),
      BigTitle = list(family = c("Comic Sans MS", "Comic Sans"),
        alt.family = "Helvetica", size = -16, bold = TRUE),
      Fixed = list(family = "Lucida Console", alt.family = "Courier",
        size = -12)
    ))
  }

  if (text) {# Set text, titles and fixed fonts
    # Determine which font style we currently use
    curStyle <- getTemp(".FontsStyle", default = "Classic", mode = "character")
    curSFonts <- getTemp(paste(".FontsStyle", curStyle, sep = ""),
      default = getTemp(".FontsStyleClassic"))
    assignTemp(".Fonts", curSFonts)

    # Create corresponding fonts in Tk (note, we create bold, italic, and
    # bolditalic equivalents for TkTextFont and TkFixedFont
    Fonts <- list()
    Fonts$Text <- curSFonts$Text
    Fonts$Text$bold <- FALSE
    Fonts$Text$italic <- FALSE
    Fonts$TextBold <- Fonts$Text
    Fonts$TextBold$bold <- TRUE
    Fonts$TextItalic <- Fonts$Text
    Fonts$TextItalic$italic <- TRUE
    Fonts$TextBoldItalic <- Fonts$TextBold
    Fonts$TextBoldItalic$italic <- TRUE
    Fonts$Title <- curSFonts$Title
    Fonts$BigTitle <- curSFonts$BigTitle
    Fonts$Fixed <- curSFonts$Fixed
    Fonts$Fixed$bold <- FALSE
    Fonts$Fixed$italic <- FALSE
    Fonts$FixedBold <- Fonts$Fixed
    Fonts$FixedBold$bold <- TRUE
    Fonts$FixedItalic <- Fonts$Fixed
    Fonts$FixedItalic$italic <- TRUE
    Fonts$FixedBoldItalic <- Fonts$FixedBold
    Fonts$FixedBoldItalic$italic <- TRUE

    FNames <- c("TkTextFont", "TkTextBoldFont", "TkTextItalicFont",
      "TkTextBoldItalicFont", "TkTitleFont", "TkBigTitleFont", "TkFixedFont",
      "TkFixedBoldFont", "TkFixedItalicFont", "TkFixedBoldItalicFont")
    res <- c(res, tk2font.set(FNames, Fonts))
  }
  # Check the results
  if (system && any(!res))
    warning("One or several Tk fonts not set: '",
      paste(names(res)[!res], collapse = "', '", "'"))
  res
}

#' A series of methods applicable to tk2widget or tk2cfglist objects
#'
#' Tk2widgets can be used as tcltk widgets, but they propose also an
#' object-oriented interaction through these different methods.
#'
#' @param x A tk2widget object.
#' @param ... A series of named arguments corresponding to parameters and values
#' to use for the configuration for `tk2cfglist()`, or reserved arguments for
#' future use for the other function (not used yet).
#' @param value A value to assign to the object's method.
#'
#' @return
#' Depends on the function. The `is.xxx()` function return `TRUE` or `FALSE` if
#' the object is of the right class or not. The assignations form return the
#' assigned value. The direct form return the item.
#'
#' @export
#' @rdname tk2methods
#' @author Philippe Grosjean
#' @seealso [tk2button()], [tk2tip()]
#' @keywords utilities
is.tk2widget <- function(x)
  return(inherits(x, "tk2widget"))

#' @exportS3Method
#' @rdname tk2methods
print.tk2widget <- function(x, ...) {
  if (disabled(x)) txt <- " (disabled)" else txt <- ""
  cat("A tk2widget of class '", class(x)[1], "'", txt, "\n", sep = "")
  cat("State: ", state(x), "\n", sep = "")
  cursize <- size(x)
  if (cursize > 0)
    cat("Size: ", cursize, "\n", sep = "")
  val <- value(x)
  if (!is.null(val)) {
    cat("Value:\n")
    print(value(x))
  }
  invisible(x)
}

#' @export
#' @rdname tk2methods
tk2cfglist <- function(...) {
  res <- list(...)
  class(res) <- c("tk2cfglist", class(res))
  res
}

#' @exportS3Method
#' @rdname tk2methods
print.tk2cfglist <- function(x, ...) {
  if (!length(x)) {
    cat("An empty tk2widget cfglist\n")
  } else {
    cat("A tk2widget cfglist with:\n\n")
    print(unclass(x))
  }
  invisible(x)
}

#' @export
#' @rdname tk2methods
state <- function(x, ...)
  UseMethod("state")

#' @exportS3Method
#' @rdname tk2methods
state.tk2widget <- function(x, ...) {
  if (any(grepl("-state ", as.character(tkconfigure(x))))) {
    as.character(tkcget(x, "-state", ...))
  } else {
    "normal"
  }
}

# TODO: a state.tk2listbox, because there is no state property defined for it!

#' @export
#' @rdname tk2methods
label <- function(x, ...)
  UseMethod("label")

#' @exportS3Method
#' @rdname tk2methods
label.tk2widget <- function(x, ...)
  x$env$label

#' @export
#' @rdname tk2methods
`label<-` <- function(x, value)
  UseMethod("label<-")

#' @export
#' @rdname tk2methods
`label<-.tk2widget` <- function(x, value) {
  x$env$label <- as.character(value)[1]
  x
}

#' @export
#' @rdname tk2methods
tag <- function(x, ...)
  UseMethod("tag")

#' @exportS3Method
#' @rdname tk2methods
tag.tk2widget <- function(x, ...)
  x$env$tag

#' @export
#' @rdname tk2methods
`tag<-` <- function(x, value)
  UseMethod("tag<-")

#' @export
#' @rdname tk2methods
`tag<-.tk2widget` <- function(x, value) {
  x$env$tag <- value
  x
}

#' @export
#' @rdname tk2methods
disabled <- function(x, ...)
  UseMethod("disabled")

#' @exportS3Method
#' @rdname tk2methods
disabled.tk2widget <- function(x, ...)
  (state(x) == "disabled")

#' @export
#' @rdname tk2methods
`disabled<-` <- function(x, value)
  UseMethod("disabled<-")

#' @export
#' @rdname tk2methods
`disabled<-.tk2widget` <- function(x, value) {
  if (isTRUE(value)) state <- "disabled" else state <- "normal"
  tkconfigure(x, state = state)
  x
}

#' @export
#' @rdname tk2methods
values <- function(x, ...)
  UseMethod("values")

#' @exportS3Method
#' @rdname tk2methods
values.tk2widget <- function(x, ...)
  NULL # Default value, for widgets that do not support this!

#' @exportS3Method
#' @rdname tk2methods
values.tk2listbox <- function(x, ...)
  as.character(tkget(x, 0, "end"))

#' @export
#' @rdname tk2methods
`values<-` <- function(x, value)
  UseMethod("values<-")

#' @export
#' @rdname tk2methods
`values<-.tk2widget` <- function(x, value)
  stop("This tk2widget does not seem to support values")

#' @export
#' @rdname tk2methods
`values<-.tk2listbox` <- function(x, value) {
  # Save current selection
  cursel <- selection(x)
  tclServiceMode(FALSE)
  on.exit(tclServiceMode(TRUE))
  isDisabled <- disabled(x)
  on.exit(disabled(x) <- isDisabled, add = TRUE)
  if (isDisabled)
    disabled(x) <- FALSE
  # Change items (no attempt to match them -possible future improvement!-)
  tkdelete(x, 0, "end")
  for (item in as.character(value))
    tkinsert(x, "end", item)
  # Try to reapply selection
  for (sel in cursel)
    tkselection.set(x, sel - 1)
  x
}

#' @export
#' @rdname tk2methods
value <- function(x, ...)
  UseMethod("value")

#' @exportS3Method
#' @rdname tk2methods
value.tk2widget <- function(x, ...)
  NULL # Default value is NULL for tk2widgets

#' @exportS3Method
#' @rdname tk2methods
value.tk2listbox <- function(x, ...)
  values(x)[selection(x)]

#' @export
#' @rdname tk2methods
`value<-` <- function(x, value)
  UseMethod("value<-")

#' @export
#' @rdname tk2methods
`value<-.tk2widget` <- function(x, value)
  stop("This tk2widget does not seem to support setting its value")

#' @export
#' @rdname tk2methods
`value<-.tk2listbox` <- function(x, value) {
  items <- items(x)
  if (length(items) > 0)
    selection(x) <- (1:length(items))[items %in% value]
  x
}

#' @export
#' @rdname tk2methods
selection <- function(x, ...)
  UseMethod("selection")

#' @exportS3Method
#' @rdname tk2methods
selection.tk2widget <- function(x, ...)
  NULL # For tk2widgets that do not support selection

#' @exportS3Method
#' @rdname tk2methods
selection.tk2listbox <- function(x, ...)
  (as.integer(tkcurselection(x)) + 1)

#' @export
#' @rdname tk2methods
`selection<-` <- function(x, value)
  UseMethod("selection<-")

#' @export
#' @rdname tk2methods
`selection<-.tk2widget` <- function(x, value)
  stop("This tk2widget does not seem to support setting its selection")

#' @export
#' @rdname tk2methods
`selection<-.tk2listbox` <- function(x, value) {
  # Prepare
  tclServiceMode(FALSE)
  on.exit(tclServiceMode(TRUE))
  isDisabled <- disabled(x)
  on.exit(disabled(x) <- isDisabled, add = TRUE)
  if (isDisabled) disabled(x) <- FALSE
  # Clear selection only
  if (is.null(value) || length(value) < 1) {
    tkselection.clear(x, 0, "end")
    return(x)
  }
  # Check data
  value <- sort(as.integer(round(value)))
  if (value[1] < 1)
    stop("Selections must be indices > 0")
  if (value[length(value)] > size(x)) return(x)
  # Change selection
  tkselection.clear(x, 0, "end")
  if (tclvalue(tkcget(x, "-selectmode")) == "single" && length(value) > 1) {
    warning("Single selection mode... only lowest selection used")
    tkselection.set(x, value[1] - 1)
  } else {
    for (sel in value)
      tkselection.set(x, sel - 1)
  }
  if (!isDisabled) tksee(x, value[1] - 1)
  x
}

#' @export
#' @rdname tk2methods
#' @param index The zero-based index of the item to make visible.
visibleItem <- function(x, index, ...)
  UseMethod("visibleItem")

#' @exportS3Method
#' @rdname tk2methods
visibleItem.tk2widget <- function(x, index, ...)
  stop("This tk2widget does not seems to support the visibleItem method")

#' @exportS3Method
#' @rdname tk2methods
visibleItem.tk2listbox <- function(x, index, ...) {
  # Index must be a positive integer
  index <- as.integer(round(index))
  if (is.null(index) || length(index) < 1 || index[1] < 1)
    stop("index must be a postive integer")
  tksee(x, index[1] - 1) # Because Tcl uses 0-based indices
  return(NULL)
}

#' @export
#' @rdname tk2methods
size <- function(x, ...)
  UseMethod("size")

#' @exportS3Method
#' @rdname tk2methods
size.tk2widget <- function(x, ...)
  0L # By default, a tk2widget has values of zero size (NULL)

#' @exportS3Method
#' @rdname tk2methods
size.tk2listbox <- function(x, ...)
  as.integer(tksize(x))

#' @export
#' @rdname tk2methods
config <- function(x, ...)
  UseMethod("config")

#' @exportS3Method
#' @rdname tk2methods
#' @param cfglist a list containing one or more named items, with the name
#'   being a Tcl/Tk property and items being the new value for the property.
config.tk2widget <- function(x, cfglist, ...) {
  # Compile a list of arguments
  args <- list(...)
  if (!missing(cfglist))
    args <- .mergeList(as.list(cfglist, args))

  # Prepare an empty object
  res <- list()
  class(res) <- c("tk2cfglist", class(res))

  # No arguments provided... query a sublist of parameters
  if (length(args) == 0) {
    # Return the complete config (but not the data!)
    params <- tk2configList(x)
    if (!length(params))
      return(res)
  } else {
    # Separate named (set) from unnamed (query only) arguments
    params <- names(args)
    if (is.null(params)) {# No named arguments, only queries
      params <- as.character(args)
    } else {
      # For those named arguments, change the config
      res <- (config(x) <- args[params != ""])
      # ... and query the others
      params <- as.character(args[params == ""])
    }
  }

  # Retrieve values for the queries
  if (length(params)) {
    for (i in 1:length(params)) {
      p <- tclvalue(tkcget(x, paste("-", params[i], sep = "")))
      if (!is.null(p) && p != "")
        res[[params[i]]] <- p
    }
  }
  res
}

#' @exportS3Method
#' @rdname tk2methods
config.tk2label <- function(x, cfglist, ...) {
  # wrap is special here... => how to deal with it???
  # TODO...
  config.tk2widget(x, cfglist, ...)
}

# TODO: config.tk2listbox()

#' @export
#' @rdname tk2methods
`config<-` <- function(x, value)
  UseMethod("config<-")

#' @export
#' @rdname tk2methods
`config<-.tk2widget` <- function(x, value) {
  # The default function deleguates to tkconfigure, except for a few things
  value <- .configStd(x, value)
  value$widget <- x
  do.call(tkconfigure, value)
  x
}

#' @export
#' @rdname tk2methods
`config<-.tk2label` <- function(x, value) {
  # Standard treatment
  value <- .configStd(x, value)

  if (!is.null(value$wrap)) {
    # wrap is not a ttk option but we use it here for convenience
    wrap <- value$wrap
    value$wrap <- NULL
  } else wrap <- NULL

  # For the other parameters, apply tkconfigure() with them
  value$widget <- x
  do.call(tkconfigure, value)

  # Do we still have to apply wrap?
  width <- abs(as.integer(tclvalue(tkcget(x, "-width"))))
  if (!is.null(wrap)) {
    if (isTRUE(wrap)) wraplength <- .wraplength(x, width) else wraplength <- 0
    tkconfigure(x, wraplength = wraplength)
  }
  # Reapply width to get correct text wrapping (bug in ttk::label?)
  if (length(width))
    tkconfigure(x, width = width)
  x
}

# TODO: `config<-.tk2listbox`

# tk2widgets.R - Support for the ttk widgets
# Copyright (c), Philippe Grosjean (phgrosjean@sciviews.org)
# Licensed under LGPL 3 or above
#
# Changes:
# - 2012-01-09: reworked tk2label to provide "full options"
#
# - 2012-01-07: listbox now behaves like a ttk widget, although it is not!
#
# - 2009-06-30: only use ttk (no tile or plain widgets)
#
# - 2007-01-01: first version (for tcltk2_1.0-0)
#
# To do:
# - Rework all this and add new widgets like sizegrip, tkplot, ...

### autoscroll
#tclRequire("autoscroll")
#tt <- tktoplevel()
#scrl <- tkscrollbar(tt, orient = "v", command = function(...) tkyview(txt, ...))
#txt <- tktext(tt, highlightthickness = 0, yscrollcommand = function(...) tkset(scrl, ...))
#tkpack(scrl, side = "right", fill = "y")
#tkpack(txt, side = "left", fill = "both", expand = 1)
#tcl("::autoscroll::autoscroll", scrl)


# Management of locales and message translation using msgcat


#' A series of versatile using either themable ttk widgets
#'
#' A series of widgets you can use in your Tk windows/dialog boxes.
#'
#' @param parent The parent window.
#' @param tip A tooltip to display for this widget (optional).
#' @param label A single character string used to label that widget (optional).
#' @param tag Ay object that you would like to associate with this widget
#' (optional).
#' @param cfglist A named list with configuration parameters and values to
#' apply.
#' @param wrap Do we wrap long lines in the widget?
#' @param values A character vector with values to use to populate the widget.
#' @param value A character vector with current value for the widget, or
#' currently selected values, if multiple selection is allowed. Takes precedence
#' on `selection`.
#' @param selection A numeric (indices) vector with current selection.
#' @param selectmode The selection mode for this widget. `extended` is the
#' usual choice for multiselection `tk2listbox()`.
#' @param height The height of the widget.
#' @param scroll Do we add scrollbars? Possible values are `"x"`, `"y"`,
#' `"both"` or `"none"`; can be abridged.
#' @param autoscroll Do we automatically hide scrollbars if not needed? Possible
#' values are the same as for the `scroll` argument.
#' @param enabled Is the widget enabled or disabled?
#' @param text The text to display in the widget.
#' @param justify How text is justified?
#' @param tabs The tabs to create in the notebook widget.
#' @param width The desired width. Use a negative value to use `aspect` instead.
#' @param aspect Sets the aspect ratio of the widget (100 = square, 200 = twice
#' as large, 50 = twice as tall). Only used if `width` is negative.
#' @param orientation Either `"horizontal"` or `"vertical"`.
#' @param activebackground Color to use for active background of menu items (if
#' not provided, a reasonable default value is used).
#' @param activeforeground Color to use for active foreground of menu items (if
#' not provided, a reasonable default value is used).
#' @param ... Further arguments passed to the widget.
#'
#' @return
#' The reference to the created widget.
#'
#' @note You need Tk 8.5 or above to use these widgets.
#'
#' @author Philippe Grosjean
#' @export
#' @rdname tk2widgets
#' @seealso [is.ttk()]
#' @keywords utilities
#'
#' @examples
#' \dontrun{
#' # These cannot be run by examples() but should be OK when pasted
#' # into an interactive R session with the tcltk package loaded
#'
#' # A tk2notebook example
#' tt2 <- tktoplevel()
#' nb <- tk2notebook(tt2, tabs = c("Test", "Button"))
#' tkpack(nb, fill = "both", expand = 1)
#' tb1 <- tk2notetab(nb, "Test")
#' lab <- tk2label(tb1, text = "Nothing here.")
#' tkpack(lab)
#' tb2 <- tk2notetab(nb, "Button")
#' but <- tk2button(tb2, text = "Click me", command = function() tkdestroy(tt2))
#' tkgrid(but)
#' tk2notetab.select(nb, "Button")
#' tk2notetab.text(nb) # Text of the currently selected tab
#'
#' # A simple tk2panedwindow example
#' tt2 <- tktoplevel()
#' pw <- tk2panedwindow(tt2, orient = "vertical")
#' lpw.1 <- tk2text(pw)
#' lpw.2 <- tk2text(pw)
#' tkadd(pw, lpw.1)#, minsize = 100)
#' tkadd(pw, lpw.2)#, minsize = 70)
#' but <- tk2button(tt2, text = "OK", width = 10,
#'   command = function() tkdestroy(tt2))
#' tkpack(pw, fill = "both", expand = "yes")
#' tkpack(but)
#' # Resize the window and move the panel separator with the mouse
#'
#' # A tk2combobox example
#' tt2 <- tktoplevel()
#' cb <- tk2combobox(tt2)
#' tkgrid(cb)
#' # Fill the combobox list
#' fruits <- c("Apple", "Orange", "Banana")
#' tk2list.set(cb, fruits)
#' tk2list.insert(cb, "end", "Scoubidou", "Pear")
#' tk2list.delete(cb, 3)   # 0-based index!
#' tk2list.size(cb)
#' tk2list.get(cb)   # All items
#' # Link current selection to a variable
#' Fruit <- tclVar("Pear")
#' tkconfigure(cb, textvariable = Fruit)
#' # Create a button to get the content of the combobox
#' but <- tk2button(tt2, text = "OK", width = 10,
#'   command = function() {tkdestroy(tt2); cat(tclvalue(Fruit), "\n")})
#' tkgrid(but)
#'
#' # An example of a tk2spinbox widget
#' tt2 <- tktoplevel()
#' tspin <- tk2spinbox(tt2, from = 2, to = 20, increment = 2)
#' tkgrid(tspin)
#' # This widget is not added yet into tcltk2!
#' #tdial <- tk2dial(tt2, from = 0, to = 20, resolution = 0.5, width = 70,
#' #	tickinterval = 2)
#' #tkgrid(tdial)
#' tbut <- tk2button(tt2, text = "OK", width = 10,
#'   command = function() tkdestroy(tt2))
#' tkgrid(tbut)
#'
#' # A tk2mclistbox example
#' tt2 <- tktoplevel()
#' mlb <- tk2mclistbox(tt2, width = 55, resizablecolumns = TRUE)
#' # Define the columns
#' tk2column(mlb, "add", "name", label = "First name", width = 20)
#' tk2column(mlb, "add", "lastname", label = "Last name", width = 20)
#' tk2column(mlb, "add", "org", label = "Organisation", width = 15)
#' tkgrid(mlb)
#' # Fill the multicolumn list (we can use a vector, or a matrix of character strings)
#' item1 <- c("Bryan", "Oackley", "ChannelPoint")
#' items <- matrix(c("John", "Ousterhout", "Scriptics", "Steve", "Miller", "TclTk inc."),
#'   ncol = 3, byrow = TRUE)
#' tk2insert.multi(mlb, "end", item1)
#' tk2insert.multi(mlb, "end", items)
#' # TODO: bind events
#' # Ex: .listbox label bind date <ButtonPress-1> "sortByDate %W"
#' # See the example.tcl in .\libs\mclistbox1.02 for a more complex example
#' # Create a button to close the dialog box
#' but <- tk2button(tt2, text = "OK", width = 10,
#'   command = function() tkdestroy(tt2))
#' tkgrid(but)
#'
#' # A simple tk2table example (Tktable is required here!)
#' myRarray <- c("Animal", "\"sphinx moth\"", "oyster", "Type", "insect", "mollusk")
#' dim(myRarray) <- c(3, 2)
#' for (i in (0:2))
#'   for (j in (0:1))
#'     .Tcl(paste("set tclarray(", i, ",", j, ") ", myRarray[i+1, j+1], sep = ""))
#' tt2 <- tktoplevel()
#' table1 <- tk2table(tt2, variable = "tclarray", rows = "3", cols = "2",
#'   titlerows = "1", selectmode = "extended", colwidth = "25", background = "white")
#' tkpack(table1)

#' # A tablelist example
#' tt <- tktoplevel()
#' tlist <- tk2tablelist(tt, columntitles = c("First column", "Second column"),
#'   stretch = "all", expand = 1)
#' tkpack(tlist, fill = "both")
#' tkinsert(tlist, "end", c("first row", "another value"))
#' tkinsert(tlist, "end", c("another row", "bla bla"))
#' tbut <- tk2button(tt, text = "Done", command = function () tkdestroy(tt))
#' tkpack(tbut)
#' }
tk2button <- function(parent, tip = "", ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  w <- tkwidget(parent, "ttk::button", ...)
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2button", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2canvas <- function(parent, tip = "", ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
### TODO: use autoscroll here!
  # Default background to fieldbackground
  if (any(names(list(...)) == "background")) {
    w <- tkwidget(parent, "canvas", ...)
  } else {
    background <- tclvalue(.Tcl("ttk::style lookup TEntry -fieldbackground"))
    if (background == "")
      background <- "white"
    w <- tkwidget(parent, "canvas", background = background, ...)
  }
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2canvas", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2checkbutton <- function(parent, tip = "", ...) {
### TODO: associate with a variable and set both states values
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  w <- tkwidget(parent, "ttk::checkbutton", ...)
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2checkbutton", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2combobox <- function(parent, tip = "", ...) {
### TODO: associate the list and results with a variable and intialize the widget
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  w <- tkwidget(parent, "ttk::combobox", ...)
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2combobox", "tk2widget", class(w))
  w
}

### TODO: a centry widget

#' @export
#' @rdname tk2widgets
tk2entry <- function(parent, tip = "", ...) {
### TODO: add cut/copy/paste/select all/clear context menu
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  w <- tkwidget(parent, "ttk::entry", cursor = "xterm", ...)
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2entry", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2frame <- function(parent, ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  w <- tkwidget(parent, "ttk::frame", ...)
  class(w) <- c("tk2frame", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2label <- function(parent, tip, label, tag, cfglist, wrap = FALSE, ...) {
  # Also image, text, textvariable, label & tag for data
  # width = -10, compound = "left", justify = "left", wrap = FALSE for config
  # Special treatment from wrap (wraplength)
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  # If a config list is provided, rework arguments
  if (!missing(cfglist)) {
    args <- .mergeList(cfglist, list(...))
    args$parent <- parent
    if (!missing(tip))
      args$tip <- tip
    if (!missing(label))
      args$label <- label
    if (!missing(tag))
      args$tag <- tag
    return(do.call(tk2label, args))
  }
  # Create the widget and the corresponding R object
  w <- tkwidget(parent, "ttk::label", ...)
  class(w) <- c("tk2label", "tk2widget", class(w))
  if (!missing(tip))
    tip(w) <- tip
  if (!missing(label))
    label(w) <- label
  if (!missing(tag))
    tag(w) <- tag
  # Special treatment for 'wrap' argument that does not exists in ttk::label
  if (!missing(wrap)) {
    # We need width
    width <- abs(as.integer(tclvalue(tkcget(w, "-width"))))
    if (isTRUE(wrap)) wraplength <- .wraplength(w, width) else wraplength <- 0
    tkconfigure(w, wraplength = wraplength)
    # If width is not reapplied after wraplength, the text is not always
    # wrapped in the widget (is this a bug?)
    if (wraplength > 0 && length(width))
      tkconfigure(w, width = width)
  }
  w
}

#' @export
#' @rdname tk2widgets
tk2labelframe <- function(parent, ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  w <- tkwidget(parent, "ttk::labelframe", ...)
  class(w) <- c("tk2labelframe", "tk2widget", class(w))
  w
}

## TODO: I need to rework this on the same scheme as tk2label

#' @export
#' @rdname tk2widgets
tk2listbox <- function(parent, values, value, selection,
selectmode = c("extended", "single", "browse", "multiple"), height = 5,
tip = "", scroll = "both", autoscroll = "x", enabled = TRUE, ...) {
  # Check conditions and arguments
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  selectmode <- match.arg(selectmode)
  scrolls <- c("x", "y", "both", "none")
  scroll <- match.arg(scroll, scrolls)
  autoscroll <- match.arg(autoscroll, scrolls)
  if (missing(values) || length(values) == 0) {
    values <- NULL
  } else {
    values <- as.character(values)
  }
  # We provide either value or selection... translate value into selection
  if (missing(selection))
    selection <- NULL
  if (!missing(value)) {
    if (is.null(values) || length(values) < 1) {
      selection <- NULL
    } else {
      selection <- (1:length(values))[values %in% as.character(value)]
    }
  }
  if (length(selection) == 0 || is.null(values) || length(values) < 1) {
    selection <- NULL
  } else {
    selection <- sort(as.integer(round(selection)))
    if (selection[1] < 1)
      stop("Numerical selections must be indices > 0")
    if (selection[1] > length(values))
      selection <- NULL
    if (selectmode == "single" && length(selection) > 1) {
      warning("Only lowest selection used in single selection mode")
      selection <- selection[1]
    }
  }

  # Location of the widget depends if we add scrollbars or not
  background <- tclvalue(.Tcl("ttk::style lookup TEntry -fieldbackground"))
  if (background == "")
    background <- "white"
  if (scroll == "none") {
    w <- tkwidget(parent, "listbox", font = "TkDefaultFont",
      borderwidth = 1, relief = "sunken", activestyle = "dotbox",
      selectmode = selectmode, height = height, exportselection = 0,
      background = background, ...)
  } else {# We need to create a tk2frame as parent of the listbox
    wf <- tk2frame(parent)
    w <- tkwidget(wf, "listbox", font = "TkDefaultFont",
      borderwidth = 1, relief = "sunken", activestyle = "dotbox",
      selectmode = selectmode, height = height, exportselection = 0,
      background = background, ...)
  }
  # Make it react to tk2theme changes, and integrate the listbox as much
  # as possible with current ttk theme
  #restyleListbox <- function (W) {
  #  # Restyle the listbox according to current ttk style
  #  # Note: font is set to TkDefaultFont => already managed there!
  #  tkconfigure(W,
  #    foreground = tk2style("tk2entry", "foreground",
  #      default = "#000000"),
  #    borderwidth = tk2style("", "borderwidth", default = 0),
  #    disabledforeground = tk2style("tk2entry", "foreground",
  #      "disabled", default = "#a3a3a3"),
  #    highlightbackground = tk2style("tk2entry", "selectbackground",
  #      default = "#c3c3c3"),
  #    highlightcolor = tk2style("tk2entry", "selectbackground",
  #      default = "#c3c3c3"),
  #    selectbackground = tk2style("tk2entry", "selectbackground",
  #      default = "#c3c3c3"),
  #    selectforeground = tk2style("tk2entry", "selectforeground",
  #      default = "#ffffff")
  #  )
  #}
  # Restyle it now
  #restyleListbox(w)
  restyleListbox <- function(W) {
    background <- tclvalue(.Tcl("ttk::style lookup TEntry -fieldbackground"))
    if (background == "")
      background <- "white"
    tkconfigure(W, background = background)
  }

  # If there are values and/or selections, populate the list now
  for (item in values)
    tkinsert(w, "end", item)
  if (!is.null(selection)) {
    for (sel in selection)
      tkselection.set(w, sel - 1) # Because Tcl uses 0-based indexing!
    tksee(w, selection[1]) # Ensure that the first selected item is visible
  }

  # Possibly add a tooltip
  if (tip != "") tk2tip(w, tip)

  ## Do we add scrollbars?
  if (scroll == "none") {
    ## Apply bindings to original listbox
    tkbind(w, "<<ThemeChanged>>", restyleListbox)
    tkbind(w, "<1>", function(W) tkfocus(W)) # Needed for mouseweel action
    # Do we disable it?
    if (!isTRUE(enabled))
      tkconfigure(w, state = "disabled")
    # Done... just return the widget
    class(w) <- c("tk2listbox", "tk2widget", class(w))
    return(w)
  } else {
    # Add (autohide) scrollbar(s)
    tcl("scrolledWidget", w, wf, scroll, autoscroll)
    # Apply bindings to frame container
    tkbind(wf, "<<ThemeChanged>>", restyleListbox)
    tkbind(wf, "<1>", function(W) tkfocus(W)) # Needed for mouseweel action
    class(wf) <- c("tk2listbox", "tk2widget", class(w))
    # Do we disable it?
    if (!isTRUE(enabled))
      tkconfigure(wf, state = "disabled")
    return(wf)
  }
}

#' @export
#' @rdname tk2widgets
tk2mclistbox <- function(parent, tip ="", ...) {
### TODO: a tile equivalent of this widget
### TODO: or adjust the header: font, color and frame
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  res <- tclRequire("mclistbox")
  if (!inherits(res, "tclObj"))
    stop("Impossible to load the Tcl mclistbox package; check your Tcl/Tk installation")

  background <- tclvalue(.Tcl("ttk::style lookup TEntry -fieldbackground"))
  if (background == "")
    background <- "white"
  w <- tkwidget(parent, "mclistbox::mclistbox", font = "TkDefaultFont",
    background = background, ...)
  tkconfigure(w, relief = "flat")
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2mclistbox", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2menu <- function(parent, activebackground, activeforeground, ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  w <- tkwidget(parent, "menu", ...)
  if (missing(activebackground))
    activebackground <- tk2style("tk2button", "selectbackground")
  if (activebackground == "")
    activebackground = "darkblue" # Default value
  if (missing(activeforeground))
    activeforeground <- tk2style("tk2button", "selectforeground")
  if (activeforeground == "") activeforeground = "white" # Default value
    tkconfigure(w, activebackground = activebackground,
      activeforeground = activeforeground)
  class(w) <- c("tk2menu", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2menubutton <- function(parent, tip = "", ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  w <- tkwidget(parent, "ttk::menubutton", ...)
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2menubutton", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2message <- function(parent, text = "", justify = c("left", "center", "right"),
width = -1, aspect = 150, tip = "", ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  justify <- as.character(justify[1])
  w <- tkwidget(parent, "message", text = text, justify = justify,
    width = width, aspect = aspect, font = "TkDefaultFont", ...)
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2message", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2notebook <- function(parent, tabs, ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  w <- tkwidget(parent, "ttk::notebook", ...)
  # Add pages
  tabs <- as.character(tabs)
  for (tab in tabs) {
    tframe <- tk2frame(w)
    tkadd(w, tframe, text = tab, sticky = "nsew")
  }
  tk2notetraverse(w)  # Enable keyboard traversal for this notebook
  class(w) <- c("tk2notebook", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2panedwindow <- function(parent, orientation = c("horizontal", "vertical"), ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  orientation <- as.character(orientation[1])
  w <- tkwidget(parent, "ttk::panedwindow", orient = orientation, ...)
  class(w) <- c("tk2panedwindow", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2progress <- function(parent, orientation = c("horizontal", "vertical"),
tip = "", ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  orientation <- as.character(orientation[1])
  w <- tkwidget(parent, "ttk::progressbar", ...)
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2progress", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2radiobutton <- function(parent, tip = "", ...) {
### TODO: associate with a variable and set both states values
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  w <- tkwidget(parent, "ttk::radiobutton", ...)
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2radiobutton", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2scale <- function(parent, orientation = c("horizontal", "vertical"),
tip = "", ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  orientation <- as.character(orientation[1])
  w <- tkwidget(parent, "ttk::scale", orient = orientation, ...)
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2scale", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2scrollbar <- function(parent, orientation = c("horizontal", "vertical"), ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  orientation <- as.character(orientation[1])
  w <- tkwidget(parent, "ttk::scrollbar", orient = orientation, ...)
  class(w) <- c("tk2scrollbar", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2separator <- function(parent, orientation = c("horizontal", "vertical"), ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  orientation <- as.character(orientation[1])
  w <- tkwidget(parent, "ttk::separator", orient = orientation, ...)
  class(w) <- c("tk2separator", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2spinbox <- function(parent, tip = "", ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  # Default background to fieldbackground
  if (any(names(list(...)) == "background")) {
    w <- tkwidget(parent, "spinbox", font = "TkDefaultFont",
      relief = "solid", borderwidth = 1, ...)
  } else {
    background <- tclvalue(.Tcl("ttk::style lookup TEntry -fieldbackground"))
    if (background == "")
      background <- "white"
    w <- tkwidget(parent, "spinbox", font = "TkDefaultFont",
      relief = "solid", borderwidth = 1, background = background, ...)
  }

  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2spinbox", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2text <- function(parent, tip = "", ...) {
### TODO: autohide scrollbars
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")

  # Default background to fieldbackground
  if (any(names(list(...)) == "background")) {
    w <- tkwidget(parent, "text", font = "TkTextFont", ...)
  } else {
    background <- tclvalue(.Tcl("ttk::style lookup TEntry -fieldbackground"))
    if (background == "")
      background <- "white"
    w <- tkwidget(parent, "text", font = "TkTextFont", background = background, ...)
  }

  tkconfigure(w, relief = "flat")
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2text", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2ctext <- function(parent, tip = "", ...) {
### TODO: autohide scrollbars
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  tclRequire("ctext")

  # Default background to fieldbackground
  if (any(names(list(...)) == "background")) {
    w <- tkwidget(parent, "ctext", font = "TkFixedFont", ...)
  } else {
    background <- tclvalue(.Tcl("ttk::style lookup TEntry -fieldbackground"))
    if (background == "")
      background <- "white"
    w <- tkwidget(parent, "ctext", font = "TkFixedFont", background = background, ...)
  }

  tkconfigure(w, relief = "flat")
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2ctext", "tk2widget", class(w))
  w
}

### TODO: rework this, using ttk::treeview

#' @export
#' @rdname tk2widgets
tk2tree <- function(parent, tip = "", ...) {
### TODO: better icons!
  # Reasonable default icons for files and folders
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  images <- as.character(tcl("image", "names"))
  if (!"Tree:dir" %in% images)
    .Tcl("image create photo Tree:dir -data {R0lGODdhEAAQAPIAAAAAAHh4eLi4uPj4APj4+P///wAAAAAAACwAAAAAEAAQAAADPVi63P4wLkKCtTTnUsXwQqBtAfh910UU4ugGAEucpgnLNY3Gop7folwNOBOeiEYQ0acDpp6pGAFArVqthQQAO///}")
  if (!"Tree:file" %in% images)
    .Tcl("image create photo Tree:file -data {R0lGODdhEAAQAPIAAAAAAHh4eLi4uPj4+P///wAAAAAAAAAAACwAAAAAEAAQAAADPkixzPODyADrWE8qC8WN0+BZAmBq1GMOqwigXFXCrGk/cxjjr27fLtout6n9eMIYMTXsFZsogXRKJf6uP0kCADv/}")

### TODO: correct support of font
  w <- tkwidget(parent, "Tree:create")  #, font = "TkDefaultFont", ...)
  tkconfigure(w, relief = "flat")
  if (tip != "")
    tk2tip(w, tip)
  class(w) <- c("tk2tree", "tk2widget", class(w))
  w
}

#' @export
#' @rdname tk2widgets
tk2table <- function(parent, ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  if (inherits(tclRequire("Tktable", warn = FALSE), "tclObj")) {
    w <- tkwidget(parent, "table", font = "TkDefaultFont", ...)
    class(w) <- c("tk2table", "tk2widget", class(w))
    return(w)
  } else {
    stop("Tcl package 'Tktable' must be installed first")
  }
}

#' @export
#' @rdname tk2widgets
tk2tablelist <- function(parent, ...) {
  if (!is.ttk())
    stop("Tcl/Tk >= 8.5 is required")
  if (inherits(tclRequire("tablelist_tile", warn = FALSE), "tclObj")) {
    # Default background to fieldbackground
    if (any(names(list(...)) == "background")) {
      w <- tkwidget(parent, "tablelist::tablelist", font = "TkDefaultFont", ...)
    } else {
      background <- tclvalue(.Tcl("ttk::style lookup TEntry -fieldbackground"))
      if (background == "")
        background <- "white"
      w <- tkwidget(parent, "tablelist::tablelist",
        font = "TkDefaultFont", background = background, ...)
    }

    class(w) <- c("tk2tablelist", "tk2widget", class(w))
    return(w)
  } else {
    stop("Tcl package 'tablelist' must be installed first")
  }
}

# Management of locales and message translation using msgcat


#' Change or get the language used in R and Tcl/Tk, strings translation in Tcl
#'
#' The function changes dynamically the language used by both R (messages only)
#' and Tcl/Tk, retrieves its current value, and manage string translation in
#' Tcl.
#'
#' @param lang An identification for the targeted language, for instance, \"en\"
#' for English, \"en_US\" for american English, \"fr\" for French, \"de\" for
#' German, \"it\" for Italian, etc. Facultative argument for [tclmclocale()].
#' @param msg A single character string with the message to translate.
#' @param translation The corresponding version in `lang`. Substitutions markers
#' like \%s for strings, or \%d for numbers are allowed (same syntax as
#' [base::gettextf()]). These translations are added in the Tcl catalog in the
#' main domain, i.e., you don't need to give a domain name with [tclmc()] to
#' retrieve the translation.
#' @param fmt A single character vector of format string.
#' @param ... Values to be passed into \code{fmt} for the substitution.
#' @param domain The 'domain", i;e., Tcl namespace where the translation is
#' defined. Use `NULL` (the default) or `""` for the main domain where
#' translations using [tclmcset()]	are stored.
#'
#' @return
#' [setLanguage()] returns `TRUE` if language was successfully changed in
#' Tcl/Tk, `FALSE` otherwise. [getLanguage()] returns a string with current
#' language in use for R, or an empty string if it cannot determinate which is
#' the language currently used, and a `tcl.language` attribute with the
#' different catalogs that are used in priority order (ending with `""` for no
#' translation, i.e., Tcl translations do not return an error, but the initial
#' string if the item is not found in the catalog).
#' [tclmclocale()] allows to change and get language for Tcl only, without
#' changing anything for R.
#'
#' The two functions [tclmcset()] and [tclmc()] allow to record and retrieve the
#' translation of strings in the main R domain. Moreover, [tclmc()] also allows
#' to retrieve translations of Tcl strings in other Tcl namespaces (a.k.a.,
#' domains), see the examples.
#'
#' @note You need the msgcat Tcl package to use this (but it is provided with
#' all recent distributions of Tcl/Tk by default).
#'
#' @author Philippe Grosjean
#' @export
#' @keywords utilities
#'
#' @examples
#' # What is the language used by Tcl?
#' tclmclocale()
#'
#' # Define a simple translation in French and German
#' tclmcset("de", "Yes", "Ja")
#' tclmcset("fr", "Yes", "Oui")
#'
#' # Determine which language is currently in use in R
#' (oldlang <- getLanguage())
#' if (oldlang != "") {
#'   # Switch to English; test a command that issues a warning and a Tcl string
#'   setLanguage("en_US")
#'   1:3 + 1:2
#'   tclmc("Yes")
#'
#'   # Switch to German and test
#'   setLanguage("de")
#'   1:3 + 1:2
#'   tclmc("Yes")
#'
#'   # Switch to Belgian French and test
#'   setLanguage("fr_BE")
#'   1:3 + 1:2
#'   tclmc("Yes")
#'
#'   # A more complex trnaslation message with a substitution
#'   tclmcset("fr", "Directory contains %d files",
#'     "Le repertoire contient %d fichiers")
#'   tclmc("Directory contains %d files", 9)
#'   # or from a R/Tcl variable...
#'   nfiles <- tclVar(12)
#'   tclmc("Directory contains %d files", tclvalue(nfiles))
#'
#'   # Retrieve a translation defined in the "tk" domain
#'   tclmc("Replace existing file?", domain = "tk")
#'
#'   # Tcl dialog boxes are translated according to the current language
#'   \dontrun{
#'     tkgetOpenFile()
#'   }
#'
#'   # Restore previous language
#'   setLanguage(oldlang)
#' }
setLanguage <- function(lang) {
  # Change locale for both R and Tcl/Tk
  Sys.setLanguage(substring(lang, 1, 2)) #Sys.setenv(LANGUAGE = lang)
  Sys.setenv(LANG = lang)
  #try(Sys.setlocale("LC_MESSAGES", lang), silent = TRUE)  # Fails on Windows!
  res <- tclRequire("msgcat")
  if (inherits(res, "tclObj")) {
    .Tcl("namespace import msgcat::*")
    # If the tcl.language attribute is defined, use it
    tcllang <- attr(lang, "tcl.language")
    if (!is.null(tcllang) && tcllang[1] != "") {
      lang <- tcllang[1] # Use only first item
    } else {
      # Tcl does not accept locales like en_US.UF-8: must be en_us only
      lang <- tolower(sub("^([^.]+)\\..*$", "\\1", lang))
    }
    if (lang == "c") {
      tclmclocale("en") # Use English by default
    } else {
      tclmclocale(lang)
    }
    TRUE
  } else {
    FALSE
  }
}

#' @export
#' @rdname setLanguage
getLanguage <- function() {
  # Try to recover current language used for messages and GUI stuff in R
  lang <- Sys.getenv("LANGUAGE")
  if (lang == "")
    lang <- Sys.getlocale("LC_MESSAGES")
  # This is a bad hack that probably does not work all the time, but at least,
  # it works under Windows for getting "fr" for French language
  if (lang == "")
    lang <- tolower(substr(Sys.getlocale("LC_COLLATE"), 1, 2))

  # Try to get language information from Tcl
  tcllang <- try(as.character(tcl("mcpreferences")), silent = TRUE)
  attr(lang, "tcl.language") <- tcllang

  lang
}

#' @export
#' @rdname setLanguage
tclmclocale <- function(lang) {
  if (missing(lang)) {
    as.character(tcl("mclocale"))
  } else {
    # Make sure lang is made compatible to Tcl
    lang <- tolower(sub("^([^.]+)\\..*$", "\\1", lang))
    as.character(tcl("mclocale", lang))
  }
}

#' @export
#' @rdname setLanguage
tclmcset <- function(lang, msg, translation)
  invisible(tclvalue(tcl("mcset", lang, msg, translation)))

#' @export
#' @rdname setLanguage
tclmc <- function(fmt, ..., domain = NULL) {
  if (is.null(domain) || domain == "") {
    # Simpler form
    tclvalue(tcl("mc", fmt, ...))
  } else {
    # Need to evaluate in 'domain' Tcl namespace
    transl <- .Tcl(paste0("namespace eval ", domain, " {set ::Rtransl [mc {",
      fmt, "}]}"))
    sprintf(tclvalue(transl), ...)
  }
}

### TODO: implement ::ico::getIconByName, ::ico::getFileIcon & ::ico::writeIcon
### TODO: gif files are acceptable too for tk2ico.set(), example:
### Image <- tclVar()
### tcl("image", "create", "photo", Image, file = "myfile.gif")
### tcl("wm", "iconphoto", tt, Image) instead of tk2ico.set

#' Manipulate icons under Windows
#'
#' Create, load and work with Windows icons. Change icons for Windows. These
#' functions are only useful for Windows, but they silently return `NULL` on
#' other platforms for writing compatible code (Windows icons instructions can
#' be simply ignored).
#'
#' @param iconfile A file with a .ico, or .exe extension, containing one or more
#' Windows icons
#' @param file A file having icon resources (.exe, or .dll).
#' @param res The name of the resource from where the icon should be extracted.
#' @param size Te size of the icon to use. For windows icons, 16 should be fine
#' usually.
#' @param win A Tk window, or an integer representing the handle (HWND) of a
#' foreign window whose icon will be changed (take care, the function returns
#' `TRUE` even if the handle is wrong!
#' @param icon A icon object.
#'
#' @return
#' An icon object, which is a reference to an image resource in Tcl. Its classes
#' are `c("tclObj", "tclIcon")`. Do not forget to destroy it using
#' [tk2ico.destroy()] when you do not need it any more!
#' If [tk2ico.load()] fails, it returns `NULL` instead of a Tcl object.
#'
#' @note
#' This is Windows-specific. It is implemented using the ico Tcl package.
#'
#' @export
#' @rdname tk2ico
#' @author Philippe Grosjean
#' @seealso [tk2dde.exec()], [tk2reg.get()]
#' @keywords utilities
#'
#' @examples
#' \dontrun{
#' # These cannot be run by examples() but should be OK when pasted
#' # into an interactive R session with the tcltk package loaded
#'
#' ### Examples of tk2ico - icon manipulation under Windows
#' tt2 <- tktoplevel()
#' # Load a system icon (there are: "application", "asterisk", "error",
#' # "exclamation", "hand", "question", "information", "warning", and "winlogo".
#' Warn <- tk2ico.load(res = "warning")
#' # Change the icon of my window tt2
#' tk2ico.set(tt2, Warn)
#' # Do not forget to destroy icon to free resource when not needed any more
#' tk2ico.destroy(Warn)
#' rm(Warn)
#'
#' ### Otherwise, the list of icons in a file are:
#' tk2ico.list()
#' # and for a given icon, the various sizes are:
#' tk2ico.sizes(res = 4)
#'
#' ### One can set icon of a window from an .ico or .exe file directly
#' tk2ico.setFromFile(tt, default = file.path(R.home("bin"), "Rgui.exe"))
#'
#' tk2ico.setFromFile(tt2, system.file("gui", "SciViews.ico", package = "tcltk2"))
#'
#' ### When done, dispose of the window and clean the workspace
#' tkdestroy(tt2)
#' rm(tt2)
#' }
tk2ico.create <- function(iconfile, res = 0, size = 16) {
  if (length(iconfile) != 1)
    stop("'iconfile' must be of length one!")
  if (!file.exists(iconfile <- as.character(iconfile)))
    stop(gettextf("File '%s' not found!", iconfile))

  # Just use tk2ico.load() with different default args)
  tk2ico.load(file = iconfile, res = res, size = size)
}

#' @export
#' @rdname tk2ico
tk2ico.destroy <- function(icon) {
  if (!is.tk())
    return(NULL)
  if (.Platform$OS.type != "windows")
    return(NULL)

  if (!inherits(icon, "tclIcon"))
    stop("'icon' is not a \"tclIcon\" object!")

  res <- tclvalue(.Tcl(paste("catch {image delete ", icon, "}", sep = "")))
  (res == "0")  # Return "0" if OK, "1" otherwise
}

#' @export
#' @rdname tk2ico
tk2ico.list <- function(file = "shell32.dll") {
  # Make sure that the 'ico' package is loaded
  .tk2ico.require()

  if (length(file) != 1)
    stop("'file' must be of length one!")

  # If the file is not found directly, try using Sys.which()
  if (!file.exists(file)) {
    File <- Sys.which(file)
    if (!file.exists(File))
      stop("file '", file, "' not found")
  } else File <- file

  cmd <- paste("::ico::icons {", File, "}", sep = "")
  res <- try(iconlist <- .Tcl(cmd), silent = TRUE)
  if (inherits(res, "try-error")) # Tcl error message is unreadable!
    stop("Unable to list the icon resources in 'file'!")
  as.character(iconlist)
}

#' @export
#' @rdname tk2ico
tk2ico.sizes <- function(file = "shell32.dll", res = "application") {
  # Make sure that the 'ico' package is loaded
  .tk2ico.require()

  if (length(file) != 1)
    stop("'file' must be of length one!")
  if (length(res) != 1)
    stop("'res' must be of length one!")
  # For compatibility reasons, res can be "application", "asterisk", "error",
  # "exclamation", "hand", "question", "information", "warning" or "winlogo"
  # but need to be changed into corresponding ID
  res <- as.character(res)[1]
  res <- switch(res,
    application = "154",
    asterisk = "173",
    error = "28",
    exclamation = "154",
    hand = "29",
    question = "263",
    information = "1001",
    warning = "200",
    winlogo = "47",
    res)

  # If the file is not found directly, try using Sys.which()
  if (!file.exists(file)) {
    File <- Sys.which(file)
    if (!file.exists(File))
      stop("file '", file, "' not found")
  } else File <- file

  cmd <- paste("::ico::iconMembers {", File, "} ", res,  sep = "")
  res <- try(iconsizes <- .Tcl(cmd), silent = TRUE)
  if (inherits(res, "try-error")) # Tcl error message is unreadable!
    stop("Unable to list sizes for the icon resource 'res' in 'file'!")
  iconsizes <- as.character(iconsizes)
  iconsizes <- unique(sub("^[^ ]+ ([0-9]+) .+$", "\\1", iconsizes))
  as.integer(iconsizes)
}

#' @export
#' @rdname tk2ico
tk2ico.load <- function(file = "shell32.dll", res = "application", size = 16) {
  # Make sure that the 'ico' package is loaded
  .tk2ico.require()

  if (length(file) != 1)
    stop("'file' must be of length one!")
  if (length(res) != 1)
    stop("'res' must be of length one!")
  # For compatibility reasons, res can be "application", "asterisk", "error",
  # "exclamation", "hand", "question", "information", "warning" or "winlogo"
  # but need to be changed into corresponding ID
  res <- as.character(res)[1]
  res <- switch(res,
    application = "154",
    asterisk = "173",
    error = "28",
    exclamation = "154",
    hand = "29",
    question = "263",
    information = "1001",
    warning = "200",
    winlogo = "47",
    res)

  # If the file is not found directly, try using Sys.which()
  if (!file.exists(file)) {
    File <- Sys.which(file)
    if (!file.exists(File))
      stop("file '", file, "' not found")
  } else File <- file

  # The old winico code!
  #cmd <- paste("winico load ", res, " {", file, "}", sep = "")
  cmd <- paste("::ico::getIcon {", File, "} ", res, " -res ", size, sep = "")
  res <- try(icon <- .Tcl(cmd), silent = TRUE)
  if (inherits(res, "try-error")) # Tcl error message is unreadable!
    stop("Unable to load the icon resource, 'file' or 'res' is wrong!")

  if (inherits(icon, "tclObj"))
    class(icon) <- c(class(icon), "tclIcon")
  icon
}

#' @export
#' @rdname tk2ico
tk2ico.setFromFile <- function(win, iconfile) {
  # iconfile can be either an .ico file, or an .exe
  # This is the simplest way to set a tk window icon
  tkwm.iconbitmap(win, iconfile)
}

#' @export
#' @rdname tk2ico
tk2ico.set <- function(win, icon) {
  # Integer for win is not supported any more
  if (inherits(win, "integer"))
    stop("Integers for argument win are not supported any more in tcltk > 1.2-0")

  # Make sure that the 'ico' package is loaded
  .tk2ico.require()

  if (!inherits(win, "tkwin") || length(win) < 1)
    stop("'win' is not a \"tkwin\" object")
  if (!inherits(icon, "tclIcon"))
    stop("'icon' is not a \"tclIcon\" object!")

  # Change the icon of a Tk window
  tcl("wm", "iconphoto", win, icon)
}

.tk2ico.require <- function() {
  if (.Platform$OS.type != "windows")
    stop("This is a Windows-specific function!")
  # Make sure tcl/tk dde is operational
  if (!capabilities("tcltk"))
    stop("This version of R cannot use Tcl/Tk!")
  if (!is.tk())
    stop("Tk is required")
  res <- tclRequire("ico", warn = TRUE)
  if (inherits(res, "tclObj"))
    res <- tclvalue(res)
  if (res[1] == FALSE)
    stop("Unable to find the 'ico' Tcl/tk package!")
  res  # The package version number
}

#' Schedule and manage delayed tasks
#'
#' Tcl allows fo scheduling execution of code on the next event loop or after a
#' given time (`after` Tcl command). `tclTaskXxx()` functions use it to schedule
#' execution of R code with much control from within R (central management of
#' scheduled tasks, possibility to define redoable tasks, use of S3 objects to
#' keep track of tasks information. The `tclAfterXxx()` functions are low-level
#' access to the Tcl `after` command.
#'
#' @param wait Time in ms to delay the task (take care: approximate value,
#' depends on when event loops are triggered). Using a value lower or equal to
#' zero, the task is scheduled on the next event loop.
#' @param fun Name of the R function to run (you may not supply arguments to
#' this function, otherwise it is not scheduled properly; take care of scoping,
#' since a copy of the function will be run from within Tcl).
#' @param expr An expression to run after 'wait'.
#' @param id The R identifier of the task to schedule, if this id contains `#`,
#' then, it is replaced by next available number, but you cannot schedule more
#' than a thousand tasks with the same name (the system will give up well
#' before, anyway). If `NULL` in [tclTaskGet()], retrieve the list of all
#' existing tasks.
#' @param all If `id = NULL`, `all = TRUE` indicate to list all tasks, including
#' hidden ones (with id starting with a dot).
#' @param redo Should the task be rescheduled n times, indefinitely
#' (`redo = TRUE`) or not (`redo = FALSE`, default, or a value <= 0).
#' @param task A Tcl task timer, or its name in Tcl (in the form of
#' 'after#xxx').
#' @param x A 'tclTask' object.
#' @param ... Further argument to the `print()` method.
#'
#' @return
#' The `tclAfterXxx()` functions return a 'tclObj' with the result of the
#' corresponding Tcl function. [tclAfter()] returns the created Tcl timer in
#' this object. If 'task' does not exists, [tclAfterInfo()] returns `NULL`.
#'
#' [tclTaskGet()] returns a 'tclTask' object, a list of such objects, or `NULL`
#' if not found.
#'
#' The four remaining `tclTaskXxx()` functions return invisibly `TRUE` if the
#' process is done successfully, `FALSE` otherwise.
#' [tclTaskRun()] forces running a task now, even if it is scheduled later.
#'
#' @export
#' @rdname tclTask
#' @author Philippe Grosjean
#' @seealso [tclFun()], [base::addTaskCallback()], [base::Sys.sleep()]
#' @keywords utilities
#'
#' @examples
#' \dontrun{
#' # These cannot be run by examples() but should be OK when pasted
#' # into an interactive R session with the tcltk package loaded
#'
#' # Run just once, after 1 sec
#' test <- function () cat("==== Hello from Tcl! ====\n")
#' tclTaskSchedule(1000, test())
#' Sys.sleep(2)
#'
#' # Run ten times a task with a specified id
#' test2 <- function () cat("==== Hello again from Tcl! ====\n")
#' tclTaskSchedule(1000, test2(), id = "test2", redo = 10)
#' Sys.sleep(1)
#'
#' # Run a function with arguments (will be evaluated in global environment)
#' test3 <- function (txt) cat(txt, "\n")
#' msg <- "==== First message ===="
#' tclTaskSchedule(1000, test3(msg), id = "test3", redo = TRUE)
#' Sys.sleep(2)
#' msg <- "==== Second message ===="
#' Sys.sleep(2)
#'
#' # Get info on pending tasks
#' tclTaskGet() # List all (non hidden) tasks
#' tclTaskGet("test2")
#' # List all active Tcl timers
#' tclAfterInfo()
#'
#' # Change a task (run 'test3' only once more, after 60 sec)
#' tclTaskChange("test3", wait = 60000, redo = 1)
#' Sys.sleep(1)
#' # ... but don't wait so long and force running 'test3' right now
#' tclTaskRun("test3")
#'
#' Sys.sleep(3)
#' # finally, delete all pending tasks
#' tclTaskDelete(NULL)
#' }
tclAfter <- function(wait, fun) {
  # This is the basic Tcl command, do prefer tclTaskSchedule()!
  wait <- as.integer(wait)[1]
  if (wait <= 0) wait <- "idle"  # Schedule task on next event loop
  # Check fun
  if (!is.function(fun))
    stop("'fun' must be a function")
  # Install a new Tcl timer
  tcl("after", wait, fun)
}

#' @export
#' @rdname tclTask
tclAfterCancel <- function(task) {
  # Cancel a Tcl timer (no effect if the timer does not exist)
  tcl("after", "cancel", as.character(task)[1])
}

#' @export
#' @rdname tclTask
tclAfterInfo <- function(task = NULL) {
  # Get info about a Tcl timer, or list all current ones (using task = NULL)
  if (is.null(task)) {
    return(tcl("after", "info"))
  } else {
    # First check that task exists
    task <- as.character(task)[1]
    ok <- tclvalue(.Tcl(paste("catch {after info ", task, "}", sep = "")))
    if (ok == 0) {
      return(tcl("after", "info", task))
    } else return(NULL)
  }
}

#' @export
#' @rdname tclTask
print.tclTask <- function(x, ...) {
  # Look when the task is run
  if (x$wait == "idle") {
    cat("tclTask '", x$id, "' scheduled on next event loop\n", sep = "")
  } else {
    cat("tclTask '", x$id, "' scheduled after ", x$wait, " ms ", sep = "")
    # Determine how much is remaining
    rem <- x$started + x$wait - proc.time()["elapsed"] * 1000
    if (rem <= 0) {
      cat("(elapsed)\n")
    } else {
      cat("(", as.integer(rem), " remaining)\n", sep = "")
    }
  }
  # Look if it is rescheduled
  if (isTRUE(x$redo)) {
    cat("Rescheduled forever\n")
  } else if (x$redo == FALSE || x$redo <= 0) {
    cat("Not rescheduled\n")
  } else if (x$redo == 1) {
    cat("Rescheduled once\n")
  } else {
    cat("Rescheduled", x$redo, "times\n")
  }
  # Print the command to be executed
  cat("runs:\n")
  print(x$expr)
  invisible(x)
}

#' @export
#' @rdname tclTask
tclTaskSchedule <- function(wait, expr, id = "task#", redo = FALSE) {
  # Schedule a task to be executed after 'wait' ms
  # If 'wait' is <= 0, schedule for execution on the next event loop
  # Id is the task name to use (if the task already exists, it is deleted
  # and replaced by the new definition)

  wait <- as.integer(wait)[1]
  if (wait <= 0) wait <- "idle"  # Schedule task on next event loop

  Tasks <- .getTclTasks()
  TNames <- ls(Tasks, all.names = TRUE)

  id <- as.character(id)[1]
  # If 'id' contains '#', replace it by a number (first one available)
  # but don't allow more than 1000 tasks with same name (to avoid bad
  # situations with buggy code like infinite loops or so)
  if (grepl("#", id)) {
    for (i in 1:1000) {
      Id <- sub("#", i, id)
      if (!Id %in% TNames) break
    }
    if (Id %in% TNames)
      stop("Too many tclTasks!")
  } else {
    # Delete the task if it already exists
    if (id %in% TNames) tclTaskDelete(id)
    Id <- id
  }

  if (!isTRUE(redo)) {
    redo <- as.integer(redo)[1]
    if (redo <= 0) redo <- FALSE
  }

  # Schedule the task, but don't run expr directly, but through tclTaskRun()
  # Note: if I use tcl("after", wait, tclTaskRun(Id), R is blocked until the
  # task is done. Here, I must provide the name of a function without args)
  task <- .makeTclTask(id = Id, wait = wait)

  # Create a tclTask object containing all info about this task
  res <- list(task = task, id = Id, expr = substitute(expr),
    started = proc.time()["elapsed"] * 1000, wait = wait,
    redo = redo)
  class(res) <- c("tclTask", class(res))

  # Add this task to the list
  Tasks[[Id]] <- res

  invisible(res)
}

#' @export
#' @rdname tclTask
tclTaskRun <- function(id) {
  # Execute the code associated with a given task and detemine if the task
  # should be rescheduled again (repeat argument)
  id <- as.character(id)[1]

  Tasks <- .getTclTasks()
  Task <- Tasks[[id]]
  if (is.null(Task)) {
    warning("tclTask '", id, "' is not found")
    return(invisible(FALSE))
  }
  # Make sure to indicate that we run it once
  if (!is.logical(Task$redo)) {
    Task$redo <- Task$redo - 1
    if (Task$redo < 1) Task$redo <- FALSE
  }
  # Update the original object too
  Tasks[[id]] <- Task

  # Run the code associate with this task
  eval(Task$expr, envir = .GlobalEnv)

  # Should we delete this task (if repeat is FALSE), or reschedule it?
  # Note, we read Task again, in case fun() would have changed something there!
  Task <- Tasks[[id]]
  # Make sure the tcl timer is destroyed (in case tclTaskRun() is
  # triggered otherwise)
  tclTaskDelete(id)
  if (Task$redo) {
    # Reschedule the task
    Task$task <- .makeTclTask(id = id, wait = Task$wait)
    # and update information in .tclTasks
    Tasks[[id]] <- Task
  }
  invisible(TRUE)
}

#' @export
#' @rdname tclTask
tclTaskGet <- function(id = NULL, all = FALSE) {
  # If id is NULL, list all scheduled tasks, otherwise, give info about a
  # particular scheduled task
  if (is.null(id)) {
    return(ls(.getTclTasks(), all.names = all))
  } else {
    ## Get the data associated with a scheduled task
    return(.getTclTasks()[[id]])
  }
}

#' @export
#' @rdname tclTask
tclTaskChange <- function(id, expr, wait, redo) {
  # Change a characteristic of a scheduled task
  # Is there something to change?
  if (missing(expr) && missing(wait) && missing(redo))
    return(invisible(FALSE))
  # Get task and change it
  Tasks <- .getTclTasks()
  Task <- Tasks[[id]]
  if (is.null(Task)) {
    warning("tclTask '", id, "' is not found")
    return(invisible(FALSE))
  }
  if (!missing(expr)) Task$expr <- substitute(expr)
  if (!missing(wait )) {
    wait <- as.integer(wait)[1]
    if (wait <= 0) wait <- "idle"  # Schedule task on next event loop
    Task$wait <- wait
  }
  if (!missing(redo)) {
    if (!isTRUE(redo)) {
      redo <- as.integer(redo)[1]
      if (redo <= 0) redo <- FALSE
    }
    Task$redo <- redo
  }
  # Delete the task and recreate it with the new parameters
  tclTaskDelete(id)
  Task$task <- .makeTclTask(id = id, wait = Task$wait)

  # Update Tasks
  Tasks[[id]] <- Task

  invisible(TRUE)
}

#' @export
#' @rdname tclTask
tclTaskDelete <- function(id) {
  Tasks <- .getTclTasks()
  # Remove a previously scheduled task (if id s NULL, then, remove all tasks)
  if (is.null(id)) {
    # Delete all current tasks
    for (Task in ls(Tasks, all.names = TRUE))
      tclAfterCancel(Tasks[[Task]]$task)
    # Eliminate .tclTasks environment from SciViews:TempEnv
    rm(list = ".tclTasks", envir = .TempEnv())
  } else {
    # Delete only one task
    Task <- Tasks[[id]]
    if (!is.null(Task)) {# The task exists
      tclAfterCancel(Task$task)
      rm(list = id, envir = Tasks)
    }
  }
}

.getTclTasks <- function() {
  # Retrieve references to all scheduled tasks
  res <- .getTemp(".tclTasks", default = NULL)
  if (is.null(res)) {
    res <- new.env(parent = .TempEnv())
    .assignTemp(".tclTasks", res)
  }
  res
}

.makeTclTask <- function(id, wait) {
  run <- function()
    eval(parse(text = paste('tclTaskRun("', id, '")', sep = "")))
  tclAfter(wait, run)
}

#' Display and manage tooltips in Tk widgets
#'
#' [tk2tip()] provides a simple mechanism to display tooltips on Tk widgets when
#' the mouse cursor hoovers on top of them.
#'
#' @param widget The widget to which a tooltip is attached.
#' @param message The message of the tooltip (\code{""} to remove the tooltip).
#' @param x A tk2widget object.
#' @param ... Further arguments to the method (unused, but reserved for
#' future use).
#' @param value The message of the tooltip, or \code{""} to remove the tip.
#'
#' @return
#' The current tip or `NULL` depending on the function.
#'
#' @note
#' This implementation is done in pure Tcl code.
#'
#' @export
#' @author Philippe Grosjean
#' @seealso [tk2button()], [label()]
#' @keywords utilities
#'
#' @examples
#' \dontrun{
#' # These cannot be run by examples() but should be OK when pasted
#' # into an interactive R session with the tcltk package loaded
#'
#' # Using plain Tcl/Tk label and button (tk2XXX equivalent have built-in
#' # tooltip features)
#' tt <- tktoplevel()
#' lb <- tklabel(tt, text = "Move mouse over me, or over the button to see tooltip")
#' tkgrid(lb)
#' tk2tip(lb, "A tooltip for the label \ndisplayed on two lines")
#' but <- tkbutton(tt, text = "Exit", width = 10,
#'   command = function() tkdestroy(tt))
#' tkgrid(but)
#' tk2tip(but, "Exit from this dialog box")
#'
#' # To test tk2killtip(), move mouse on top of a widget
#' # so that the tip is visible, and force killing it manually using:
#' tk2killtip()
#' # Move again to the widget: the tip is displayed again.
#'
#' # With tk2widgets, the tip() method can also be used:
#' lb2 <- tk2label(tt, text = "Move also over me to see the tooltip")
#' tkgrid(lb2)
#' tip(lb2) # No tip yet
#' tip(lb2) <- "Now the tooltip is there!"
#' # Move the mouse over that last label
#'
#' tip(lb2) # Yes, this is my tooltip
#' tip(lb2) <- NULL # To eliminate the tooltip for this widget
#' }
tk2tip <- function(widget, message) {
  if (!is.tk())
    stop("Package Tk is required but not loaded")
  if (is.null(message))
    message <- ""
  res <- tclRequire("tooltip")
  if (inherits(res, "tclObj")) {
    res <- tcl("tooltip::tooltip", widget, message)
    # Store tip text in the object (use NULL instead of "" for no tip)
    if (message == "")
      message <- NULL
    widget$env$tip <- message
  } else {
    stop("cannot find tcl package 'tooltip'")
  }
  invisible(res)
}

#' @export
#' @rdname tk2tip
tk2killtip <- function() {
  if (!is.tk())
    stop("Package Tk is required but not loaded")
  invisible(tcl("tooltip::hide"))
}

#' @export
#' @rdname tk2tip
tip <- function(x, ...)
  UseMethod("tip")

#' @export
#' @rdname tk2tip
tip.tk2widget <- function(x, ...)
  x$env$tip

#' @export
#' @rdname tk2tip
`tip<-` <- function(x, value)
  UseMethod("tip<-")

#' @export
#' @rdname tk2tip
`tip<-.tk2widget` <- function(x, value) {
  tk2tip(x, value)
  x
}

#' A list selector that allows to select and arrange items freely
#'
#' The swaplist is perfect to select and arrange items in a given order from a
#' fixed initial set of possible items.
#'
#' @param items A vector with all items.
#' @param selection A vector with preselected items (must be a subset of `items`).
#' @param title The title of the dialog box, by default, "Select items".
#' @param ... Further parameters passed to swaplist, see its tcl man page:
#' https://core.tcl-lang.org/tklib/doc/trunk/embedded/www/tklib/files/modules/swaplist/swaplist.html.
#'
#' @return A vector with the selected items in the chosen order.
#' @export
#' @seealso [tk2listbox()], [tk2tablelist()]
#'
#' @examples
#' \dontrun{
#' library(tcltk2)
#' # tk2swaplist() makes its use super-easy
#' tk2swaplist(1:9, selection = c(1, 3, 5))
#'
#' # Use of the swaplist on your own
#' tclRequire("swaplist")
#' tt <- tktoplevel()
#' opts <- tclVar()
#' sl <- tcl("swaplist::swaplist", tt, opts, 1:9, c(1, 3, 5))
#' cat("You choose:", tclvalue(opts), "\n")
#' rm(opts, sl, tt)
#' }
tk2swaplist <- function(items, selection, title = "Select items", ...) {
  win <- tktoplevel()
  res <- try(tclRequire("swaplist"), silent = TRUE)
  if (inherits(res, "try-error"))
    stop("swaplist Tcl package not available")
  sel <- tclVar()
  res <- tcl("swaplist::swaplist", win, sel, items, selection,
    title = title, ...)
  if (tclvalue(res) == 0) { # User cancelled
    res <- character(0)
  } else {
    res <- tclObj(sel)
  }
  if (is.ordered(items))
    return(ordered(as.character(res), levels = levels(items)))
  if (is.factor(items))
    return(factor(as.character(res), levels = levels(items)))
  switch(typeof(items),
    integer = as.integer(res),
    double = as.numeric(res),
    logical = as.logical(res),
    complex = as.complex(res),
    as.character(res)
  )
}

# TODO:
# - Rework all this...
# - Style option of Tile widgets?
# - Implement style element options ...

#' Tk commands associated with the tk2XXX widgets
#'
#' These commands supplement those available in the tcltk package to ease
#' manipulation of tk2XXX widgets.
#'
#' @param widget The widget to which these actions apply.
#' @param action Which kind of action?
#' @param ... Further arguments to the action.
#' @param where Where are these item added in the list (by default, at the end).
#' @param items The items to add (either a vector for a single line, or a matrix
#' for more items).
#' @param first The 0-based first index to consider in the list.
#' @param last The 0-based last index to consider in the list, or `"end"` for
#' using the last element of the list.
#' @param index The 0-based index where to insert items in the list.
#' @param nb A tk2notebook widget ('tclObj' object).
#' @param tab The name (text) of a tab in a notebook.
#' @param state The new state of the widget, or the state to inquiry.
#' @param theme A theme to use (character string).
#' @param class The class of the tk2widget (either the Tk class, like `TButton`,
#' or the name of the function that creates it, like [tk2button()].
#' @param style A character string with the name of the style to retrieve.
#' @param default The default value to return in case this style is not found.
#' @param x Either a tk2widget object, or a character string with its class
#' name.
#'
#' @details
#' [tk2column()] manipulate columns of a tk2mclistbox widget,
#' [tk2insert.multi()] is used to insert multiple field entries in a
#' tk2mclistbox widget,
#' [is.tk()] determines if the tk package is loaded (on some platforms it is
#' possible to load the tcltk package without tk, for instance, in batch mode).
#' [is.ttk()] determines if 'ttk' widgets (styled widgets) used by the
#' `tk2XXX()` functions are available (you need Tk >= 8.5).
#'
#' @return
#' Nothing, these functions are used for their side-effect of changing the state
#' of Tk widgets
#'
#' @note
#' In comparison with traditional Tk widgets, ttk proposes an advances mechanism
#' for styling the widgets with "themes". By default, it adapts to the current
#' platform (for instance, under Windows, all widgets take the appearance of
#' Windows themed widgets (even with custom themes applied!). Usual Tk widgets
#' are ALWAYS displayed in old-looking fashion under Windows. If you want, you
#' can switch dynamically to a different theme among those available (list them
#' using [tk2theme.list()], and switch to another one with `tk2theme(newtheme)`.
#' This is most useful to see how your GUI elements and dialog boxes look like
#' on foreign systems. If you prefer, let's say, a Unix look of the R GUI
#' elements under Windows, these functions are also useful. If you are more
#' adventurous, you can even design your own themes (see the tile documentation
#' on the Tcl wiki).
#'
#' @export
#' @rdname tk2commands
#' @author Philippe Grosjean
#' @seealso [tk2button()], [tk2tip()]
#' @keywords utilities
#'
#' @examples
#' \dontrun{
#' # These cannot be run by examples() but should be OK when pasted
#' # into an interactive R session with the tcltk package loaded
#'
#' tt <- tktoplevel()
#' # A label with a image and some text
#' file <- system.file("gui", "SciViews.gif", package = "tcltk2")
#'
#' # Make this a tk2image function...
#' Image <- tclVar()
#' tkimage.create("photo", Image, file = file)
#'
#' tlabel <- tk2label(tt, image = Image,
#'   text = "A label with an image")
#' tkpack(tlabel)
#' config(tlabel, compound = "left")
#'
#' tlabel2 <- tk2label(tt, text = "A disabled label")
#' tkpack(tlabel2)
#' disabled(tlabel2) <- TRUE
#'
#' fruits <- c("Apple", "Orange", "Banana")
#' tcombo <- tk2combobox(tt, values = fruits)
#' tkpack(tcombo)
#' tkinsert(tcombo, 0, "Apple")
#'
#' # Buttons
#' tbut <- tk2button(tt, text = "Enabled")
#' tbut2 <- tk2button(tt, text = "Disabled")
#' tkpack(tbut, tbut2)
#' tkconfigure(tbut2, state = "disabled")
#'
#' tcheck <- tk2checkbutton(tt, text = "Some checkbox")
#' tcheck2 <- tk2checkbutton(tt, text = "Disabled checkbox")
#' tkconfigure(tcheck2, state = "disabled")
#' tcheck3 <- tk2checkbutton(tt, text = "Disabled and selected")
#' tkpack(tcheck, tcheck2, tcheck3)
#' cbValue <- tclVar("1")
#' tkconfigure(tcheck3, variable = cbValue)
#' tkconfigure(tcheck3, state = "disabled")
#'
#' tradio <- tk2radiobutton(tt, text = "Some radiobutton")
#' tradio2 <- tk2radiobutton(tt, text = "Disabled and checked")
#' tkpack(tradio, tradio2)
#' tkconfigure(tradio2, state = "checked")
#' tkconfigure(tradio2, state = "disabled")
#'
#' # Menu allowing to change ttk theme
#' topMenu <- tkmenu(tt)           # Create a menu
#' tkconfigure(tt, menu = topMenu) # Add it to the 'tt' window
#' themes <- tk2theme.list()
#' themeMenu <- tkmenu(topMenu, tearoff = FALSE)
#' if ("alt" %in% themes) tkadd(themeMenu, "command", label = "alt",
#'   command = function() tk2theme("alt"))
#' if ("aqua" %in% themes) tkadd(themeMenu, "command", label = "aqua",
#'   command = function() tk2theme("aqua"))
#' if ("clam" %in% themes) tkadd(themeMenu, "command", label = "clam",
#'   command = function() tk2theme("clam"))
#' tkadd(themeMenu, "command", label = "clearlooks",
#'   command = function() tk2theme("clearlooks"))
#' if ("classic" %in% themes) tkadd(themeMenu, "command", label = "classic",
#'   command = function() tk2theme("classic"))
#' if ("default" %in% themes) tkadd(themeMenu, "command", label = "default",
#'   command = function() tk2theme("default"))
#' tkadd(themeMenu, "command", label = "keramik",
#'   command = function() tk2theme("keramik"))
#' tkadd(themeMenu, "command", label = "plastik",
#'   command = function() tk2theme("plastik"))
#' tkadd(themeMenu, "command", label = "radiance (fonts change too)!",
#'   command = function() tk2theme("radiance"))
#' if ("vista" %in% themes) tkadd(themeMenu, "command", label = "vista",
#'   command = function() tk2theme("vista"))
#' if ("winnative" %in% themes) tkadd(themeMenu, "command", label = "winnative",
#'   command = function() tk2theme("winnative"))
#' if ("xpnative" %in% themes) tkadd(themeMenu, "command", label = "xpnative",
#'   command = function() tk2theme("xpnative"))
#' tkadd(themeMenu, "separator")
#' tkadd(themeMenu, "command", label = "Quit", command = function() tkdestroy(tt))
#' tkadd(topMenu, "cascade", label = "Theme", menu = themeMenu)
#' tkfocus(tt)
#' }
tk2column <- function(widget, action = c("add", "configure", "delete", "names",
"cget", "nearest"), ...) {
  Action <- action[1]
  tcl(widget, "column", Action, ...)
}

#' @export
#' @rdname tk2commands
tk2list.set <- function(widget, items) {
  # Set a list of values for a widget (e.g., combobox)
  if (inherits(widget, "tk2combobox")) {
    # ttk::combobox uses -values parameter
    tkconfigure(widget, values = as.character(items))
  } else {
    # Try to use the defaul method
    # First, clear the list
    tcl(widget, "list", "delete", 0, "end")
    ## Then, insert all its elements
    items <- as.character(items)
    for (item in items)
      tcl(widget, "list", "insert", "end", item)
  }
}

#' @export
#' @rdname tk2commands
tk2list.insert <- function(widget, index = "end", ...) {
  # Insert one or more items in a list
  if (inherits(widget, "tk2combobox")) {
    # ttk::combobox uses -values parameter
    Items <- as.character(unlist(list(...)))
    if (length(Items) < 1)
      return()  # Nothing to insert
    List <- as.character(tcl(widget, "cget", "-values"))
    if (length(List) < 2 && List == "") {
      # The list in empty, simply add these items
      List <- Items
    } else if (index == "end" || index > length(List) - 1) {
      List <- c(List, Items)
    } else if (index == 0) {
      # Insert items at the beginning of the list
      List <- c(Items, List)
    } else {
      # Insert items inside the list
      List <- c(List[1:index], Items, List[(index + 1):length(List)])
    }
    # Reassign this modified list to the combobox
    tkconfigure(widget, values = List)
  } else {
    tcl(widget, "list", "insert", index, ...)
  }
}

#' @export
#' @rdname tk2commands
tk2list.delete <- function(widget, first, last = first) {
  # Delete one or more items from a list
  if (inherits(widget, "tk2combobox")) {
    # ttk::combobox uses -values parameter
    List <- as.character(tcl(widget, "cget", "-values"))
    if (length(List) < 2 && List == "")
      return(List)  # The list in empty
    if (last == "end") last <- length(List) else last <- last + 1
    List <- List[-((first + 1):last)]
    # Reassign this modified list to the combobox
    tkconfigure(widget, values = List)
  } else {
    tcl(widget, "list", "delete", first, last)
  }
}

#' @export
#' @rdname tk2commands
tk2list.get <- function(widget, first = 0, last = "end") {
  # Get the list of elements in a widget (e.g., combobox)
  if (inherits(widget, "tk2combobox")) {
    # ttk::combobox uses -values parameter
    List <- as.character(tcl(widget, "cget", "-values"))
    if (length(List) < 2 && List == "")
      return(List)
    if (last == "end") last <- length(List) else last <- last + 1
    return(List[(first + 1):last])
  } else {
    as.character(tcl(widget, "list", "get", first, last))
  }
}

#' @export
#' @rdname tk2commands
tk2list.size <- function(widget) {
  # Get the length of the list of elements in a widget (e.g., combobox)
  if (inherits(widget, "tk2combobox")) {
    # ttk::combobox uses -values parameter
    List <- as.character(tcl(widget, "cget", "-values"))
    return(length(List))
  } else {
    as.numeric(tcl(widget, "list", "size"))
  }
}

#' @export
#' @rdname tk2commands
tk2state.set <- function(widget, state = c("normal", "disabled", "readonly")) {
  # Change the state of a widget
  state <- as.character(state[1])
  tkconfigure(widget, state = state)
}

#' @export
#' @rdname tk2commands
tk2insert.multi <- function(widget, where = "end", items) {
  # We insert one or several lines in a multicolumn widget
  items <- as.matrix(items)
  # A vector is coerced into a column matrix and we want a row matrix here
  if (ncol(items) == 1) items <- t(items)
  # Convert the matrix into [list {el1} {el2} {el3}] [list {el4}, {el5}, {el6}], ...
  makeTclList <- function(x)
    paste("[list {", paste(x, collapse = "} {"), "}]", sep = "")
  TclList <- paste(apply(items, 1, makeTclList), collapse = "\\\n")
  .Tcl(paste(widget, "insert", where, TclList))
}

#' @export
#' @rdname tk2commands
tk2notetraverse <- function(nb)
  invisible(tcl("ttk::notebook::enableTraversal", nb))

#' @export
#' @rdname tk2commands
tk2notetab <- function(nb, tab) {
  if (inherits(nb, "tk2notebook")) {
    # We need the tab index, so, look for it
    ntab <- as.numeric(tcl(nb, "index", "end"))
    if (ntab < 1)
      return(NULL)
    tabidx <- -1
    for (i in 0:(ntab - 1)) {
      if (tclvalue(tcl(nb, "tab", i, "-text")) == tab) {
        tabidx <- i
        break
      }
    }
    if (tabidx > -1) {
      tabid <- paste(nb$ID, tabidx + 1, sep = ".")
      # Create a simili tkwin object referring to this page
      w <- list()
      w$ID <- tabid
      w$env <- new.env()
      w$env$num.subwin <- 0
      w$env$parent <- nb
      class(w) <- c("tk2notetab", "tk2container", "tkwin")
      return(w)
    } else return(NULL)  # Tab not found!
  } else stop("'nb' must be a 'tk2notebook' object")
}

#' @export
#' @rdname tk2commands
tk2notetab.select <- function(nb, tab) {
  # Select a tab in a notebook
  if (inherits(nb, "tk2notebook")) {
    # Tile notebook
    # We need the tab index, so, look for it
    ntab <- as.numeric(tcl(nb, "index", "end"))
    if (ntab < 1)
      return(invisible(FALSE))
    tabidx <- -1
    for (i in 0:(ntab - 1)) {
      if (tclvalue(tcl(nb, "tab", i, "-text")) == tab) {
        tabidx <- i
        break
      }
    }
    if (tabidx > -1) {
      tkselect(nb, tabidx)
      return(invisible(TRUE))
    } else return(invisible(FALSE))
  } else stop("'nb' must be a 'tk2notebook' object")
}

#' @export
#' @rdname tk2commands
tk2notetab.text <- function(nb) {
  # Select a tab in a notebook
  if (inherits(nb, "tk2notebook")) {
    return(tclvalue(tcl(nb, "tab", "current", "-text")))
  } else stop("'nb' must be a 'tk2notebook' object")
}

# Themes management
#' @export
#' @rdname tk2commands
tk2theme.elements <- function()
  as.character(.Tcl("ttk::style element names"))

#' @export
#' @rdname tk2commands
tk2theme.list <- function()
  as.character(.Tcl("ttk::style theme names"))

#' @export
#' @rdname tk2commands
tk2theme <- function(theme = NULL) {
  if (is.null(theme)) {# Get it
    res <- getOption("tk2theme")
  } else {# Set it to theme
    # First, check if the theme is already loaded... or try loading it
    loadedThemes <- tk2theme.list()
    if (!theme %in% loadedThemes) {
      # Could be plastik, keramik, keramik_alt, clearlooks, radiance
      res <- try(tclRequire(paste0("ttk::theme::", theme)), silent = TRUE)
      if (inherits(res, "try-error"))
        stop("Ttk theme ", theme, " is not found")
    }
    # Themes (like radiance) change TkDefaultFont => reset it for the others
    if (theme == "radiance") {
      tkfont.configure("TkDefaultFont", family = "Ubuntu", size = 11)
    } else {
      tk2font.set("TkDefaultFont", tk2font.get("TkSysDefaultFont"))
    }
    # Change theme
    .Tcl(paste("ttk::style theme use", theme))
    # And save current theme in option "tk2theme"
    options(tk2theme = theme)
    # Make sure to homogenize background for old tk widgets (suggested by Milan Bouchet-Valat)
    # Note: foreground not defined for plastik and keramik => workaround
    fg <- tclvalue(.Tcl("ttk::style lookup TLabel -foreground"))
    if (fg == "") fg <- "#000000"
    afg <- tclvalue(.Tcl("ttk::style lookup TLabel -foreground active"))
    if (afg == "") afg <- "#000000"
    ffg <- tclvalue(.Tcl("ttk::style lookup TLabel -foreground focus"))
    if (ffg == "") ffg <- "#000000"
    hfg <- tclvalue(.Tcl("ttk::style lookup TLabel -foreground hover"))
    if (hfg == "") hfg <- "#000000"
    .Tcl(paste("tk_setPalette",
      "background",
        tclvalue(.Tcl("ttk::style lookup TLabel -background")),
      "foreground", fg,
      "activeBackground",
        tclvalue(.Tcl("ttk::style lookup TLabel -background active")),
      "activeForeground", afg,
      "disabledForeground",
        tclvalue(.Tcl("ttk::style lookup TLabel -foreground disabled")),
      "highlightBackground", "white",
        #tclvalue(.Tcl("ttk::style lookup TLabel -background focus")),
      "highlightColor", ffg,
      "insertBackground", afg,
      "selectBackground",
        tclvalue(.Tcl("ttk::style lookup TText -selectbackground")),
      "selectForeground",
        tclvalue(.Tcl("ttk::style lookup TText -selectforeground")),
      "selectColor",
        tclvalue(.Tcl("ttk::style lookup TText -selectforeground")),
      "throughColor", hfg),
      "fieldBackground",
        tclvalue(.Tcl("ttk::style lookup TEntry -fieldbackground")))

    # Set menu font the same as label font
    font <- tclvalue(.Tcl("ttk::style lookup TLabel -font"))
    if (!length(font) || font == "")
      font <- "TkDefaultFont"
    tk2font.set("TkMenuFont", tk2font.get(font))

    # Return the theme
    res <- theme
  }
  res
}
# Note: to change a style element: .Tcl('ttk::style configure TButton -font "helvetica 24"')
# Create a derived style: ttk::style configure Emergency.TButton -font "helvetica 24" -foreground red -padding 10
# Changing different states:
#ttk::style map TButton \
#  -background [list disabled #d9d9d9  active #ececec] \
#  -foreground [list disabled #a3a3a3] \
#  -relief [list {pressed !disabled} sunken] \
#  ;

# Function to look for a ttk style
#' @export
#' @rdname tk2commands
tk2style <- function(class, style, state = c("default", "active",
"disabled", "focus", "!focus", "pressed", "selected", "background", "readonly",
"alternate", "invalid", "hover", "all"), default = NULL) {
  # Get a ttk style in the current theme
  # Class is either the TTk class, or the tk2 function name
  # TODO: add tk2toolbutton and tk2sizegrip!
  class <- switch(class,
    tk2button = "TButton",
    tk2label = "TLabel",
    tk2toolbutton = "Toolbutton",
    tk2menubutton = "TMenubutton",
    tk2checkbutton = "TCheckbutton",
    tk2radiobutton = "TRadiobutton",
    tk2entry = "TEntry",
    tk2combobox = "TCombobox",
    tk2notebook = "TNotebook",
    tk2labelframe = "TLabelframe",
    tk2scrollbar = "TScrollbar",
    tk2scale = "TScale",
    tk2progress = "TProgressbar",
    #tk2spinbox = "TSpinbox",
    tk2tree = "Treeview",
    tk2frame = "TFrame",
    tk2panedwindow = "TPanedwindow",
    tk2separator = "TSeparator",
    #"TSizegrip",
    as.character(class)[1] # Supposed to be the ttk class
    # Not ttk widgets: tk2canvas, tk2ctext, tk2edit, tk2listbox,
    # tk2mclistbox, tk2menu, tk2menuentry, tk2spinbox, tk2table
  )
  style = paste("-", as.character(style)[1], sep = "")
  state = match.arg(state)
  if (is.null(default))
    default <- ""

  # styles creates a named vector (items in even elements, labels = odd)
  styles <- function(x) {
    st <- as.character(x)
    l <- length(st)
    if (l == 0)
      return(character(0))
    if (l == 1)
      return(c(default = st))
    if (l %% 2 > 0)
      stop("Didn't get an even number of items: ", st)
    stnames <- st[seq(1, l - 1, by = 2)]
    st <- st[seq(2, l, by = 2)]
    names(st) <- stnames
    return(st)
  }

  # First look at the map for this class
  res <-  styles(tcl("ttk::style", "map", class, style))
  res2 <-  styles(tcl("ttk::style", "map", ".", style))
  res <- c(res, res2[!names(res2) %in% names(res)])
  res2 <-  styles(tcl("ttk::style", "configure", class, style))
  res <- c(res, res2[!names(res2) %in% names(res)])
  res2 <-  styles(tcl("ttk::style", "configure", ".", style))
  res <- c(res, res2[!names(res2) %in% names(res)])
  if (length(res) == 0)
    res <- c(default = default)

  # If state != "all", try to resolve the right state
  if (state != "all") {
    # If the given state is there, use it
    if (state %in% names(res)) {
      return(res[state])
    } else if ("default" %in% names(res)) {
      return(res["default"])
    } else {
      return(c(default = as.character(default)[1]))
    }
  } else return(res)
}

#' @export
#' @rdname tk2commands
tk2dataList <- function(x) {
  # List data parameters for a given tk2widget
  # Data manage the content of the widgets
  # Common items are label, tag, and tip
  # image: widgets that can display images
  # text, textvariable: display text
  # values, value and selection
  # command: the command to run
  # validate, validatecommand, invalidcommand: validation mechanism
  # variable: varaible associated with value
  # postcommand: specific to comboboxes, to fill them!
  # onvalue & offvalue: specific to checkbutton
  # default: specific for button (default button in a dialog box)
  # show: specific to entry for password... clash with treeview show => ???
  # mode, maximum, value: for progressbars
  # from, to, increment, : for spinbox & scale + format
  # Look in text widget what we keep!
  if (is.tk2widget(x)) cl <- class(x)[1] else cl <- as.character(x)[1]
  res <- switch(cl,
    tk2button = c("image", "text", "textvariable", "command", "default"),
    tk2canvas = character(0),
    tk2checkbutton = c("image", "text", "textvariable", "variable",
      "command", "onvalue", "offvalue"),
    tk2combobox = c("postcommand", "textvariable", "values"),
    tk2ctext = c("values", "value", "selection", "maxundo", "undo",
      "spacing1", "spacing2", "spacing3", "tabs", "tabstyle"), # language
    tk2entry = c("invalidcommand", "textvariable", "validate",
      "validatecommand", "show"),
    tk2label = c("image", "text", "textvariable"),
    tk2labelframe = c("text"),
    tk2listbox = c("values", "value", "selection"),
    tk2mclistbox = c("values", "value", "selection"),
    tk2notebook = character(0),
    tk2notetab = c("image", "text"),
    tk2panedwindow = character(0),
    tk2progress = c("mode", "maximum", "value", "variable"),
    tk2radiobutton = c("image", "text", "textvariable",
      "command", "value", "variable"),
    tk2scale = c("command", "from", "to", "value", "variable"),
    tk2scrollbar = c("command"),
    tk2separator = character(0),
    #tk2sizegrip = character(0),
    tk2spinbox = c("validate", "validatecommand", "from", "to", "increment",
      "values", "format", "command"),
    tk2table = c("values", "value", "selection"),
    tk2text = c("values", "value", "selection", "maxundo", "undo",
      "spacing1", "spacing2", "spacing3", "tabs", "tabstyle"),
    tk2tree = c("values", "value", "selection"),
    stop("Unknown tk2widget, provide a tk2widget object or its class")
  )
  # Add label, tag & tip for all
  res <- c(res, "label", "tag", "tip")
  res
}

#' @export
#' @rdname tk2commands
tk2configList <- function(x) {
  # List config parameters for a given tk2widget
  # Note: most of the appearance is controlled by the theme, we keep here
  # only a subset of items that are most useful considering themed widgets:
  # height, width or length: the size of the widget
  # compound: how image and text are composed
  # justify and wrap: control of text flow
  # orient: for widgets that can be horizontal or vertical
  # selectmode: for widgets that allow for multiple selections
  # show: tree and/or headings for the treeview widget
  if (is.tk2widget(x)) cl <- class(x)[1] else cl <- as.character(x)[1]
  res <- switch(cl,
    tk2button = c("compound", "width"),
    tk2canvas = c("height", "width"),
    tk2checkbutton = c("compound", "width"),
    tk2combobox = c("justify", "height", "width"),
    tk2ctext = c("height", "width"),
    tk2entry = c("justify", "width"),
    tk2label = c("compound", "justify", "width", "wraplength"), # Use wrap!
    tk2labelframe = c("height", "width"),
    tk2listbox = c("height", "width", "selectmode"),
    tk2mclistbox = c("height", "width", "selectmode"),
    tk2notebook = c("height", "width"),
    tk2notetab = c("compound"),
    tk2panedwindow = c("orient", "height", "width"),
    tk2progress = c("length", "orient"),
    tk2radiobutton = c("compound", "width"),
    tk2scale = c("length", "orient"),
    tk2scrollbar = c("orient"),
    tk2separator = character(0),
    #tk2sizegrip = character(0),
    tk2spinbox = c("wrap"),
    tk2table = c("height", "width"),
    tk2text = c("height", "width"),
    tk2tree = c("height", "selectmode", "show"), # show tree and/or headings
    stop("Unknown tk2widget, provide a tk2widget object or its class")
  )
  # Add cursor and takefocus that are common to all
  # Should we really add these?
  #res <- c(res, "cursor", "takefocus")
  res
}

# Check if Tk or Ttk are available
#' @export
#' @rdname tk2commands
is.tk <- function()
  (tclvalue(.Tcl("catch { package present Tk }")) == "0")

#' @export
#' @rdname tk2commands
is.ttk <- function() {
  (is.tk() && as.numeric(tcl("set", "::tk_version")) >= 8.5)
}

---
title: "Tcl/Tk Additions"
author: "Philippe Grosjean (phgrosjean@sciviews.org)"
date: "`r Sys.Date()`"
output:
  rmarkdown::html_vignette:
    fig_caption: yes
vignette: >
  %\VignetteIndexEntry{Tcl/Tk Additions}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

> The {tcltk2} package provides many supplements to R through Tcl functions and Tk widgets.

\name{tk2reg}
\alias{tk2reg}
\alias{tk2reg.broadcast}
\alias{tk2reg.delete}
\alias{tk2reg.deletekey}
\alias{tk2reg.get}
\alias{tk2reg.keys}
\alias{tk2reg.set}
\alias{tk2reg.setkey}
\alias{tk2reg.type}
\alias{tk2reg.values}

\title{ Manipulate the registry under Windows }
\description{
  These functions access the Windows registry in a secure way (most errors
  are handled gracefully), and ensures correct conversion back and forth
  for atomic strings ('sz' and 'expand\_') and numbers ('dword' and
  'dword\_big\_endian'), and for vectors of strings ('multi\_sz').
}

\usage{
tk2reg.broadcast()
tk2reg.delete(keyname, valuename)
tk2reg.deletekey(keyname)
tk2reg.get(keyname, valuename)
tk2reg.keys(keyname)
tk2reg.set(keyname, valuename, data, type = c("sz", "expand_sz", "multi_sz",
    "dword", "dword_big_endian"))
tk2reg.setkey(keyname)
tk2reg.type(keyname, valuename)
tk2reg.values(keyname)
}

\arguments{
  \item{keyname}{ the name of the key. }
  \item{valuename}{ a value in this key. }
  \item{data}{ the data to place in the value. }
  \item{type}{ the type of value in the registry. By default, it is 'sz', that
    is, an atomic string. }
}

\value{
  Functions that should return registry value(s) or key(s) return them in a
  character string, or they return \code{NA} if the key/value is not found in
  the registry.

  \code{tk2reg.broadcast()}, \code{tk2reg.delete()}, \code{tk2reg.deletekey()}
  \code{tk2reg.set} and \code{tk2reg.setkey()} return \code{TRUE} in case of
  success and \code{FALSE} otherwise.

  \code{tk2reg.get()} should handle correctly the types 'sz', 'expand\_sz' and
  'multi\_sz' (note that 'expand\_sz' string is NOT expanded!), as well as
  'dword' and 'dword\_big\_endian' that are converted into numeric values. Other
  types are not converted and the Tcl expression is returned ('objTcl' class)
  untransformed.

  \code{tk2reg.set()} currently works with 'sz', 'expand\_sz', 'multi\_sz',
  'dword' and 'dword\_big\_endian' types. A couple of other types are accepted
  by the function... but they are not tested ('binary', 'link',
  'resource\_list'). 
}

\author{ Philippe Grosjean }

\note{ For Windows only. These functions issue an error when they are called
  under other platforms. Take care while manipulating the Windows registry!
  You can easily lock the system completely, if you delete important items,
  especially if you are logged as administrator on your computer. Make a backup
  of your registry first before experimenting with these function!!!
}

\seealso{ \code{\link{tk2dde}}, \code{\link{tk2ico}} }

\examples{
\dontrun{
## These cannot be run by examples() but should be OK when pasted
## into an interactive R session with the tcltk package loaded

### Examples of tk2reg - registry manipulation under Windows
## Rem: HKEY_LOCAL_MACHINE, HKEY_USERS, HKEY_CLASSES_ROOT, HKEY_CURRENT_USER,
##      HKEY_CURRENT_CONFIG, HKEY_PERFORMANCE_DATA, HKEY_DYN_DATA
Rkey <- "HKEY_CURRENT_USER\\\\Software\\\\R-core\\\\R"   # The R key
Rkey <- paste(Rkey, "\\\\", R.version$major, ".", R.version$minor, sep = "")
Rsubkey <- paste(Rkey, "subkey", sep = "\\\\")        # A subkey

## Get all subkeys for Software in the local machine
tk2reg.keys("HKEY_LOCAL_MACHINE\\\\Software")

## Get all names in the R key
tk2reg.values(Rkey)

## Get the path for the current R version
tk2reg.get(Rkey, "InstallPath")

## Create a subkey (explore the registry with regedit.exe to see it)
tk2reg.setkey(Rsubkey)
## Add a couple of keys in it
tk2reg.set(Rsubkey, "test", "a key added in the registry!", type = "sz")
tk2reg.set(Rsubkey, "test exp", "\%SystemRoot\%\\\\system32", type = "expand_sz")
tk2reg.set(Rsubkey, "test multi", LETTERS[1:5], type = "multi_sz")
tk2reg.set(Rsubkey, "test dword", 1024, type = "dword")
tk2reg.set(Rsubkey, "test big end", 1024, type = "dword_big_endian")

## Get the type of a value
tk2reg.type(Rsubkey, "test")
tk2reg.type(Rsubkey, "test exp")
tk2reg.type(Rsubkey, "test multi")
tk2reg.type(Rsubkey, "test dword")
tk2reg.type(Rsubkey, "test big end")

## Get a value in a key
tk2reg.get(Rsubkey, "test")
tk2reg.get(Rsubkey, "test exp")
tk2reg.get(Rsubkey, "test multi")
tk2reg.get(Rsubkey, "test dword")
tk2reg.get(Rsubkey, "test big end")

## Delete a name in a key (take care: dangerous!)
tk2reg.delete(Rsubkey, "test")
## Delete a whole key (take care: very dangerous!)
tk2reg.deletekey(Rsubkey)

## An alternate way to get the path
tk2reg.get(paste("HKEY_LOCAL_MACHINE", "SYSTEM", "CurrentControlSet",
    "Control", "Session Manager", "Environment", sep = "\\\\"), "path")

## Make sure that currently running apps are warned of your changes in the registry
tk2reg.broadcast()

## Delete temporary variables
rm(list = c("Rkey", "Rsubkey"))
}
}

\keyword{ utilities }

\concept{ Windows registry, keys, hives, permanent storage of configuration parameters }

\name{tk2reg}
\alias{tk2reg}
\alias{tk2reg.broadcast}
\alias{tk2reg.delete}
\alias{tk2reg.deletekey}
\alias{tk2reg.get}
\alias{tk2reg.keys}
\alias{tk2reg.set}
\alias{tk2reg.setkey}
\alias{tk2reg.type}
\alias{tk2reg.values}

\title{ Manipulate the registry under Windows }
\description{
  These functions are Windows-specific. They issue an error on other platforms.
}

\usage{
tk2reg.broadcast()
tk2reg.delete(keyname, valuename)
tk2reg.deletekey(keyname)
tk2reg.get(keyname, valuename)
tk2reg.keys(keyname)
tk2reg.set(keyname, valuename, data, type = c("sz", "expand_sz", "multi_sz",
    "dword", "dword_big_endian"))
tk2reg.setkey(keyname)
tk2reg.type(keyname, valuename)
tk2reg.values(keyname)
}

\arguments{
  \item{keyname}{ the name of the key. }
  \item{valuename}{ a value in this key. }
  \item{data}{ the data to place in the value. }
  \item{type}{ the type of value in the registry. By default, it is 'sz', that
    is, an atomic string. }
}

\keyword{ utilities }

\concept{ Windows registry, keys, hives, permanent storage of configuration parameters }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tk2dialogs.R
\name{tk2chooseFont}
\alias{tk2chooseFont}
\alias{tk2unicode_config}
\alias{tk2unicode_select}
\alias{tk2unicode_bind}
\title{Additional Tk dialog boxes}
\usage{
tk2chooseFont(...)

tk2unicode_config(parent)

tk2unicode_select(widget)

tk2unicode_bind(widget)
}
\arguments{
\item{...}{Further arguments passed to the dialog box.}

\item{parent}{The Tk toplevel dialog box that will be the parent of the
configuration dialog box.}

\item{widget}{A widget that can accept a unicode character. For
\code{tk2unicode_bind()} it must be a \code{tk2text} or a \code{tk2entry} widget.}
}
\value{
The selection made in the dialog box if \code{OK} is clicked, \code{""} otherwise for
\code{\link[=tk2chooseFont]{tk2chooseFont()}}.

The \code{\link[=tk2unicode_select]{tk2unicode_select()}} dialog pastes the selected unicode character in the
designed widget, but returns nothing. The \code{\link[=tk2unicode_config]{tk2unicode_config()}} changes the
configuration for the unicode composer, but returns nothing. If you decide to
do so, it saves the config on a file. This is done app-by-app, and the
default app name is \code{"R"}. You can change it by setting a different
value in the option \code{"tk2app"}, i.e., \code{options(tk2app = "myApp")}.
The \code{tk2unicode_bind()} is also invoked for its side-effect to install
required bindings to enable the unicode composer engine for the given widget
and it returns nothing.
}
\description{
Tk dialog boxes to select a font, unicode characters or a list of ordered
items.
}
\examples{
\dontrun{
library(tcltk2)
# These cannot be run by examples() but should be OK when pasted
# into an interactive R session with the tcltk package loaded

# Font selection
tk2chooseFont()
tk2chooseFont(font = "{courier} 9", title = "Choose a fixed font",
  fonttype = "fixed", style = 4, sizetype = "all")
tk2chooseFont(font = "Verdana 12 bold italic underline overstrike",
  fonttype = "prop", style = 2, sizetype = "point")

# Easy unicode character entry
tt <- tktoplevel()
txt <- tk2text(tt, width = 60, height = 20)
tkpack(txt)
e <- tk2entry(tt, width = 50)
tkpack(e)
# Get an unicode character for the text widget
tk2unicode_select(txt)
# and for the entry widget
tk2unicode_select(e)

# Bind the composer to both the text and the entry widgets
# and display the configuration box
# Once done, try the compose key + m + u, or compose + " + a
# or any othert sequence in both widgets
# or hit the compose key twice
tk2unicode_bind(txt)
tk2unicode_bind(e)
tk2unicode_config(tt)
}
}
\seealso{
\code{\link[=tk2text]{tk2text()}}, \code{\link[=tk2listbox]{tk2listbox()}}, \code{\link[=tk2list.insert]{tk2list.insert()}}
}
\author{
Philippe Grosjean
}
\keyword{utilities}

\name{tk2dde}
\alias{tk2dde}
\alias{tk2dde.exec}
\alias{tk2dde.poke}
\alias{tk2dde.request}
\alias{tk2dde.services}

\title{ Use DDE (Dynamic Data Exchange) under Windows }
\description{
  DDE is the first Microsoft's attempt to make an inter-application mechanism.
  It is now superseeded by (D)Com, but it is still available. Being simpler than
  Com, DDE is interesting for simple tasks. Applications like Word, Excel,
  Windows Explorer, Internet Explorer, etc... provide services one can access
  through DDE (see examples).
}
\usage{
tk2dde(topic = NULL)
tk2dde.services(service = "", topic = "")
tk2dde.exec(service, topic, command, async = FALSE)
tk2dde.poke(service, topic, item, data)
tk2dde.request(service, topic, item, binary = FALSE)
}

\arguments{
  \item{topic}{ the 'topic' to reach or expose. A DDE server is accessed as
    'service'|'topic'. In the case of \code{tk2DDE()}, a non null topic activates
    the DDE server, and a null topic deactivate it. }
  \item{service}{ the name of the service to reach. In \code{tk2DDEservices}, if
    both service and topic are empty, the list of all available DDE service is
    returned, otherwise, only available topics for a given service are listed. }
  \item{command}{ a string with the command to run in the external application
    (syntax depends on the server). }
  \item{async}{ is a command run asynchroneously (returns immediately, before
    the command is processed), or not? }
  \item{item}{ the concerned item (usually a variable name, a range in a
    worksheet, etc...). }
  \item{data}{ the new value for the item. }
  \item{binary}{ should the return bez treated as binary data or not? }
}

\author{ Philippe Grosjean }

\note{ This is only available under Windows. Trying to use these functions under
  other platforms raises an error. Under Windows, R is automatically configured
  as a DDE server with name TclEval|SciViewsR when this package is loaded }

\seealso{ \code{\link{tk2reg}} }

\examples{
\dontrun{
## These cannot be run by examples() but should be OK when pasted
## into an interactive R session with the tcltk package loaded

## Examples of DDE - Windows only

### Examples using wish ###
## Start a Wish84 console side-by-side with R.
## (to get wish, you need to install ActiveTcl from
## http://www.activestate.com/Products/ActiveTcl/)
## Once it is done, start 'Wish84' from the start menu)
## Register the Wish console as a DDE server, that is, type in it
## (\% is the Tcl prompt, don't type it!):
## \% package require dde
## \% dde servername wish

### In R:
tk2dde("R") # Return 0 if succeed
tk2dde.services()
## Evaluate some string in wish
tk2dde.exec("TclEval", "wish", "{puts {Hello World!}}")
## Give a value to a variable in wish
tk2dde.poke("TclEval", "wish", "myvar", "{This is a string!}")
## Note that you must surround strings with curly braces in Tcl!
tk2dde.poke("TclEval", "wish", "mynumvar", c(34.56, 78.9))

## In wish, check that vars exist and have correct value
## \% puts $myvar
## \% puts $mynumvar

## Get the value of one variable from wish into R
tk2dde.request("TclEval", "wish", "myvar")
tk2dde.request("TclEval", "wish", "mynumvar")
## Note that you don't know here if it is a string, a number, or so...
## You have to know and convert yourself!

## Now, the other way: execute a R function from wish
## You first need to register a R function for callback
## (For the moment, only functions without arguments are supported!)
doDDE <- function() cat("DDE execute!\n")	# A simple function
tclFun(doDDE)
## And in wish
## \% dde execute TclEval R doDDE

## Once you have defined a variable using tclVar, you can get or change it
## from the dde server. However, tclVar gives cryptic names like ::RTcl1.
## So we prefer to use tclVarName
myvar2 <- tclVarName("myvar2", "this is a test...")
tclvalue(myvar2) # This is the way we access to this variable in R

## In wish you get the value and change it:
## \% dde request TclEval R myvar2
## Again, dde poke does not work and must be replaced by an execute command
## This doesn't work (???)
## \% dde poke TclEval R myvar2 {yes! and it works...}
## ... but this is fine
## \% dde execute TclEval R {set myvar2 {yes! and it works...}}

## And in R...
tclvalue(myvar2)


### DDE at the command line with execdde.exe ###
## You can also change the value of a variable, or run a command in R from
## the command line using execdde.exe:
## - Download execdde.exe from http://www.sciviews.org/SciViews-R/execdde.zip
## - Unzip it and copy 'execdde.exe' somewhere in your path,
## - Start a DOS window
## - Enter the following commands ('>' is the prompt, don't type it):
## > execdde -s TclEval -t R -c doDDE > NUL
## > if errorlevel 1 echo An error occurs... branch accordingly in your batch!
## > execdde -s TclEval -t R -c "set myvar2 {ok from execdde!}" > NUL

## And in R:
tclvalue(myvar2)
## Note: thanks to separate event loops, it works also when R calculates...


### Other examples of DDE use:

### Manipulating Microsoft Excel ###
## Start Excel with a blank workbook, then...

## Change values in Excel from R:
tk2dde.poke("Excel", "Sheet1", "R1C1:R2C1", c("5.7", "6.34"))   # Some data
tk2dde.poke("Excel", "Sheet1", "R3C1", "= A1 + A2")        		# A formula

## Read values in Excel (note that results of formulas are returned)
Res <- tk2dde.request("Excel", "Sheet1", "R1C1:R3C1")
Res
as.numeric(Res)

## Execute a command in Excel
tk2dde.exec("Excel", "Sheet1", "{\[\Select(\"R1C1:R3C1\")]\[New(2,2)\]}")
## New(2,2) create a bar graph in a separate sheet
## Note: close this graph now, in order to run the rest properly!

## R as server and Excel as client
## Prepare R as server
tk2dde("R") # Now, my server name is TclEval|R !
## Create a variable that will be shared with Excel
XlVar <- tclVarName("XlVar", "this is a test!")
## Enter a formula somewhere in Excel to link to this variable
tk2dde.poke("Excel", "Sheet1", "R5C1", "=TclEval|R!'::XlVar'")

## Now, change the content in XlVar
tclvalue(XlVar) <- "Another text..."
## Rem: in my Excel, the content of the cell is not changed automatically...
## I must reenter the formula for the changes to be considered!?


### Manipulating Microsoft Word ###
## Start Word and then...
tk2dde.exec("Winword", "System", "{\[EndOfDocument\]}")
tk2dde.exec("Winword", "System", "{\[Insert \"Text from R.\"\]}", async = TRUE)

## Can also use:
##  AppMaximize 1
##  StatOfDocument
##  EditFind .Find="hello" .PatternMatch=1
##  FileExit 2
##  FileClose 2 # 2 = close without saving, 1 = save first, 0 = prompt

## If you want to start the program if not running, you must take care that
## you leave enough time to the program to initialize before you can use DDE
## commands with it! Here is an example (make sure that Word is not running):
is.WordStarted <- function()
    (length(tk2dde.services("Winword", "System")) > 0)
if (!is.WordStarted()) {
    ## Get the path to WinWord.exe
    Word <- try(tk2reg.get(paste("HKEY_LOCAL_MACHINE", "SOFTWARE", "Classes",
        "WordDocument", "protocol", "StdFileEditing", "server", sep = "\\"), ""))
    if (is.na(Word)) {
	cat("Microsoft Word is not found!\n")
    } else {  # Start Microsoft Word and wait for its complete loading
        ## Start Word
        system(Word, wait = FALSE)
        ## Wait a maximum of 15 sec...
        for (i in 1:15) {
            Sys.sleep(1)  # Wait 1 sec
            res <- is.WordStarted()
	    if (res) break
	}
	if (res) cat("Now Word is started and ready to get DDE commands!\n")
	tk2dde.exec("Winword", "System", "{\[Insert \"My text from R.\"\]}",
	    async = TRUE)
    }
}


### Control Matlab ###
## Start Matlab and then...
tk2dde.exec("Matlab", "Engine", "s = ones(5);")
## Then in Matlab, type s


### Manipulating Progman to create and delete shortcuts ###
tk2dde.exec("progman", "progman", "{\[CreateGroup(Bogus)\]}")
tk2dde.exec("progman", "progman", "{\[AddItem(notepad.exe,BogusPadLink)\]}")
tk2dde.exec("progman", "progman", "{\[ShowGroup(Bogus,0)\]}")
tk2dde.exec("progman", "progman", "{\[ShowGroup(Bogus,1)\]}")
## And delete it...
tk2dde.exec("progman", "progman", "{\[DeleteItem(BogusPadLink)\]}")
tk2dde.exec("progman", "progman", "{\[DeleteGroup(Bogus)\]}")
## Rem: the same can be done with tcom, if installed:
##.Tcl("package require tcom")
##.Tcl("set sh [::tcom::ref createobject {WScript.Shell}]")
##.Tcl("set lnk [$sh CreateShortcut {c:\\\\temp\\\\boguspad.lnk}]")
##.Tcl("$lnk TargetPath {\"notepad.exe\"}")
##.Tcl("$lnk WorkingDirectory {c:\\\\temp}")
##.Tcl("$lnk Arguments Tutorial.txt")
##.Tcl("$lnk Save")


### Controlling Internet Explorer ###
## Start Internet Explorer, then...
tk2dde.exec("iexplore", "WWW_OpenURL", "{http://www.r-project.org/}")
## It works, but return an error in R???
## This is similar to the browseURL() in package util (just for the demo!)
HelpIndex <- file.path(R.home(), "doc", "html", "index.html")
tk2dde.exec("iexplore", "WWW_ShowFile", paste("{", HelpIndex, "}", sep = ""))
## Info on the current opened window
tk2dde.request("iexplore", "WWW_GetWindowInfo", "1")


### Controlling Windows explorer ###
tk2dde.exec("Folders", "AppProperties",
    "{\[ViewFolder(\"C:\\\",\"C:\\\",5)\]}")
## ... or you can also try 'ExplodeFolder'
## Search in folder
tk2dde.exec("Folders", "AppProperties", "{\[FindFolder(\"\",\"C:\\\")\]}")
}
}

\keyword{ utilities }
\keyword{ IO }

\concept{ Interprocess communication, DDE }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/setLanguage.R
\name{setLanguage}
\alias{setLanguage}
\alias{getLanguage}
\alias{tclmclocale}
\alias{tclmcset}
\alias{tclmc}
\title{Change or get the language used in R and Tcl/Tk, strings translation in Tcl}
\usage{
setLanguage(lang)

getLanguage()

tclmclocale(lang)

tclmcset(lang, msg, translation)

tclmc(fmt, ..., domain = NULL)
}
\arguments{
\item{lang}{An identification for the targeted language, for instance, \"en\"
for English, \"en_US\" for american English, \"fr\" for French, \"de\" for
German, \"it\" for Italian, etc. Facultative argument for \code{\link[=tclmclocale]{tclmclocale()}}.}

\item{msg}{A single character string with the message to translate.}

\item{translation}{The corresponding version in \code{lang}. Substitutions markers
like \\%s for strings, or \\%d for numbers are allowed (same syntax as
\code{\link[base:sprintf]{base::gettextf()}}). These translations are added in the Tcl catalog in the
main domain, i.e., you don't need to give a domain name with \code{\link[=tclmc]{tclmc()}} to
retrieve the translation.}

\item{fmt}{A single character vector of format string.}

\item{...}{Values to be passed into \code{fmt} for the substitution.}

\item{domain}{The 'domain", i;e., Tcl namespace where the translation is
defined. Use \code{NULL} (the default) or \code{""} for the main domain where
translations using \code{\link[=tclmcset]{tclmcset()}}    are stored.}
}
\value{
\code{\link[=setLanguage]{setLanguage()}} returns \code{TRUE} if language was successfully changed in
Tcl/Tk, \code{FALSE} otherwise. \code{\link[=getLanguage]{getLanguage()}} returns a string with current
language in use for R, or an empty string if it cannot determinate which is
the language currently used, and a \code{tcl.language} attribute with the
different catalogs that are used in priority order (ending with \code{""} for no
translation, i.e., Tcl translations do not return an error, but the initial
string if the item is not found in the catalog).
\code{\link[=tclmclocale]{tclmclocale()}} allows to change and get language for Tcl only, without
changing anything for R.

The two functions \code{\link[=tclmcset]{tclmcset()}} and \code{\link[=tclmc]{tclmc()}} allow to record and retrieve the
translation of strings in the main R domain. Moreover, \code{\link[=tclmc]{tclmc()}} also allows
to retrieve translations of Tcl strings in other Tcl namespaces (a.k.a.,
domains), see the examples.
}
\description{
The function changes dynamically the language used by both R (messages only)
and Tcl/Tk, retrieves its current value, and manage string translation in
Tcl.
}
\note{
You need the msgcat Tcl package to use this (but it is provided with
all recent distributions of Tcl/Tk by default).
}
\examples{
# What is the language used by Tcl?
tclmclocale()

# Define a simple translation in French and German
tclmcset("de", "Yes", "Ja")
tclmcset("fr", "Yes", "Oui")

# Determine which language is currently in use in R
(oldlang <- getLanguage())
if (oldlang != "") {
  # Switch to English; test a command that issues a warning and a Tcl string
  setLanguage("en_US")
  1:3 + 1:2
  tclmc("Yes")

  # Switch to German and test
  setLanguage("de")
  1:3 + 1:2
  tclmc("Yes")

  # Switch to Belgian French and test
  setLanguage("fr_BE")
  1:3 + 1:2
  tclmc("Yes")

  # A more complex trnaslation message with a substitution
  tclmcset("fr", "Directory contains \%d files",
    "Le repertoire contient \%d fichiers")
  tclmc("Directory contains \%d files", 9)
  # or from a R/Tcl variable...
  nfiles <- tclVar(12)
  tclmc("Directory contains \%d files", tclvalue(nfiles))

  # Retrieve a translation defined in the "tk" domain
  tclmc("Replace existing file?", domain = "tk")

  # Tcl dialog boxes are translated according to the current language
  \dontrun{
    tkgetOpenFile()
  }

  # Restore previous language
  setLanguage(oldlang)
}
}
\author{
Philippe Grosjean
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tclTask.R
\name{tclAfter}
\alias{tclAfter}
\alias{tclAfterCancel}
\alias{tclAfterInfo}
\alias{print.tclTask}
\alias{tclTaskSchedule}
\alias{tclTaskRun}
\alias{tclTaskGet}
\alias{tclTaskChange}
\alias{tclTaskDelete}
\title{Schedule and manage delayed tasks}
\usage{
tclAfter(wait, fun)

tclAfterCancel(task)

tclAfterInfo(task = NULL)

\method{print}{tclTask}(x, ...)

tclTaskSchedule(wait, expr, id = "task#", redo = FALSE)

tclTaskRun(id)

tclTaskGet(id = NULL, all = FALSE)

tclTaskChange(id, expr, wait, redo)

tclTaskDelete(id)
}
\arguments{
\item{wait}{Time in ms to delay the task (take care: approximate value,
depends on when event loops are triggered). Using a value lower or equal to
zero, the task is scheduled on the next event loop.}

\item{fun}{Name of the R function to run (you may not supply arguments to
this function, otherwise it is not scheduled properly; take care of scoping,
since a copy of the function will be run from within Tcl).}

\item{task}{A Tcl task timer, or its name in Tcl (in the form of
'after#xxx').}

\item{x}{A 'tclTask' object.}

\item{...}{Further argument to the \code{print()} method.}

\item{expr}{An expression to run after 'wait'.}

\item{id}{The R identifier of the task to schedule, if this id contains \verb{#},
then, it is replaced by next available number, but you cannot schedule more
than a thousand tasks with the same name (the system will give up well
before, anyway). If \code{NULL} in \code{\link[=tclTaskGet]{tclTaskGet()}}, retrieve the list of all
existing tasks.}

\item{redo}{Should the task be rescheduled n times, indefinitely
(\code{redo = TRUE}) or not (\code{redo = FALSE}, default, or a value <= 0).}

\item{all}{If \code{id = NULL}, \code{all = TRUE} indicate to list all tasks, including
hidden ones (with id starting with a dot).}
}
\value{
The \code{tclAfterXxx()} functions return a 'tclObj' with the result of the
corresponding Tcl function. \code{\link[=tclAfter]{tclAfter()}} returns the created Tcl timer in
this object. If 'task' does not exists, \code{\link[=tclAfterInfo]{tclAfterInfo()}} returns \code{NULL}.

\code{\link[=tclTaskGet]{tclTaskGet()}} returns a 'tclTask' object, a list of such objects, or \code{NULL}
if not found.

The four remaining \code{tclTaskXxx()} functions return invisibly \code{TRUE} if the
process is done successfully, \code{FALSE} otherwise.
\code{\link[=tclTaskRun]{tclTaskRun()}} forces running a task now, even if it is scheduled later.
}
\description{
Tcl allows fo scheduling execution of code on the next event loop or after a
given time (\code{after} Tcl command). \code{tclTaskXxx()} functions use it to schedule
execution of R code with much control from within R (central management of
scheduled tasks, possibility to define redoable tasks, use of S3 objects to
keep track of tasks information. The \code{tclAfterXxx()} functions are low-level
access to the Tcl \code{after} command.
}
\examples{
\dontrun{
# These cannot be run by examples() but should be OK when pasted
# into an interactive R session with the tcltk package loaded

# Run just once, after 1 sec
test <- function () cat("==== Hello from Tcl! ====\n")
tclTaskSchedule(1000, test())
Sys.sleep(2)

# Run ten times a task with a specified id
test2 <- function () cat("==== Hello again from Tcl! ====\n")
tclTaskSchedule(1000, test2(), id = "test2", redo = 10)
Sys.sleep(1)

# Run a function with arguments (will be evaluated in global environment)
test3 <- function (txt) cat(txt, "\n")
msg <- "==== First message ===="
tclTaskSchedule(1000, test3(msg), id = "test3", redo = TRUE)
Sys.sleep(2)
msg <- "==== Second message ===="
Sys.sleep(2)

# Get info on pending tasks
tclTaskGet() # List all (non hidden) tasks
tclTaskGet("test2")
# List all active Tcl timers
tclAfterInfo()

# Change a task (run 'test3' only once more, after 60 sec)
tclTaskChange("test3", wait = 60000, redo = 1)
Sys.sleep(1)
# ... but don't wait so long and force running 'test3' right now
tclTaskRun("test3")

Sys.sleep(3)
# finally, delete all pending tasks
tclTaskDelete(NULL)
}
}
\seealso{
\code{\link[=tclFun]{tclFun()}}, \code{\link[base:taskCallback]{base::addTaskCallback()}}, \code{\link[base:Sys.sleep]{base::Sys.sleep()}}
}
\author{
Philippe Grosjean
}
\keyword{utilities}

\name{tk2dialogs}
\alias{tk2chooseFont}
\alias{tk2swaplist}

\title{ Additional Tk dialog boxes  }
\description{
  Tk dialog boxes to select a font or a list of ordered items.
}
\usage{
tk2chooseFont(\dots)
tk2swaplist(items, selection, title = "Select items", \dots)
}

\arguments{
  \item{items}{ a vector of numbers, logicals, characters, factor or ordered. }
  \item{selection}{ preselected items, in the right order. }
  \item{title}{ title of the dialog box. }
  \item{\dots}{ further arguments passed to the dialog box. }
}

\value{
  The selection made in the dialog box if \code{OK} is clicked, \code{""}
  otherwise for \code{tk2chooseFont()}, or an zero-length vector for
  \code{tk2swaplist()}.
}

\author{ Philippe Grosjean }
\note{ If you use tile 0.7.2 or above, these dialog boxes will automatically
  use it. Otherwise, the dialog boxes will use plain Tk widgets (not yet for
  \code{tk2swaplist()}). }

\seealso{ \code{\link{tk2widgets}}, \code{\link{tk2commands}} }

\examples{
\dontrun{
## These cannot be run by examples() but should be OK when pasted
## into an interactive R session with the tcltk package loaded

tk2chooseFont()
tk2chooseFont(font = "{courier} 9", title = "Choose a fixed font",
    fonttype = "fixed", style = 4, sizetype = "all")
tk2chooseFont(font = "Verdana 12 bold italic underline overstrike",
    fonttype = "prop", style = 2, sizetype = "point")

tk2swaplist(1:10, 1:5) # integer
tk2swaplist(as.numeric(1:10), 1:5) # double
tk2swaplist(paste("V", 1:10), paste("V", 1:5)) # character
tk2swaplist(as.factor(1:10), 1:5) # factor
tk2swaplist(as.ordered(1:10), 1:5) # ordered
}
}

\keyword{ utilities }

\concept{ More Tcl/Tk dialog boxes, font and color selection }

# mclistbox.pod
# Copyright (c) 1999, Bryan Oakley
# All Rights Reserved.
#
# this uses a modified version of pod2html. Specifically, it uses the
# non-standard R<> directive, which inserts a line break.

=pod

=head2 NAME

mclistbox::mclistbox - Create and manipulate a multi-column listbox

=head2 SYNOPSIS

B<package require mclistbox ?1.02>

B<mclistbox::mclistbox> I<pathName> I<?options?>

=head2 EXPORTS

mclistbox

=head2 STANDARD OPTIONS

B<-background>
B<-borderwidth>
B<-cursor>
B<-exportselection>
B<-font>
B<-foreground>
B<-height>
B<-highlightbackground>
B<-highlightcolor>
B<-highlightthickness>
B<-relief>
B<-selectbackground>
B<-selectborderwidth>
B<-selectforeground>
B<-setgrid>
B<-takefocus>
B<-width>
B<-xscrollcommand>
B<-yscrollcommand>

See the I<options> manual entry for detailed descriptions of the above options.


=head2 WIDGET-SPECIFIC OPTIONS

Command-Line Name: B<-columnborderwidth> R<>
Database Name: B<columnBorderWidth>R<>
Database Class: B<BorderWidth>


Specifies a non-negative value indicating the width of the 3-D border
to draw around the outside of the column (if such a border is being
drawn; the B<columnrelief> option typically determines this).  The value may
also be used when drawing 3-D effects in the interior of the
widget. The value may have any of the forms acceptable to
Tk_GetPixels.

Command-Line Name: B<-columnrelief> R<>
Database Name: B<columnRelief> R<>
Database Class: B<Relief>

Specifies the 3-D effect desired for the column.  Acceptable values
are raised, B<sunken>, B<flat>, B<ridge>, B<solid>, and B<groove>. The
value indicates how the interior of the column should appear relative to its
exterior;  for example, raised means the interior of the column should
appear to protrude from the screen, relative to the exterior of the
column. 

Command-Line Name: B<-fillcolumn> R<>
Database Name: B<fillColumn>R<>
Database Class: B<FillColumn>

Specifies which column should grow or shrink such that all columns
exactly fill the listbox widget. 

Command-Line Name: B<-labelanchor> R<>
Database Name: B<labelAnchor>R<>
Database Class: B<Anchor>

Specifies how the information in the column label is
to be displayed. Must be one of the values B<n>, B<ne>, B<e>, B<se>,
B<s>, B<sw>, B<w>, B<nw>, or B<center>. For example, nw means display
the information such that its top-left corner is at the top-left
corner of the column label. 

Command-Line Name: B<-labelbackground> or B<-labelbg>R<>
Database Name: B<labelBackground> R<>
Database Class: B<Background>

Specifies the normal background color to use when displaying the
label. 

Command-Line Name: B<-labelborderwidth or -labelbd> R<>
Database Name: B<labelBorderWidth> R<>
Database Class: B<BorderWidth>

Specifies a non-negative value indicating the width of the 3-D border
to draw around the outside of the column label (if such a border is
being drawn; the relief option typically determines this).  The value
may also be used when drawing 3-D effects in the interior of the
column label. The value may have any of the forms acceptable to
Tk_GetPixels.

Command-Line Name: B<-labelfont> R<>
Database Name: B<labelFont> R<>
Database Class: B<Font>

Specifies the font to use when drawing text inside the column label.

Command-Line Name: B<-labelforeground> R<>
Database Name: B<labelForeground> R<>
Database Class: B<Foreground>

Specifies the normal foreground color to use when displaying the
column label. 

Command-Line Name: B<-labelheight> R<>
Database Name: B<labelHeight>R<>
Database Class: B<Height>

Specifies a desired height for the label. If an image or bitmap is
being displayed in the label then the value is in screen units
(i.e. any of the forms acceptable to Tk_GetPixels); for text it is in
lines of text. If this option isn't specified, the label's desired
height is computed from the size of the image or bitmap or text being
displayed in it. 

Command-Line Name: B<-labelrelief> R<>
Database Name: B<labelRelief>R<>
Database Class: B<Relief>

Specifies the 3-D effect desired for the column label.  Acceptable
values are B<raised>, B<sunken>, B<flat>, B<ridge>, B<solid>, and
B<groove>. The value indicates how the interior of the column label
should appear relative to its exterior; for example, raised means the
interior of the column label should appear to protrude from the
screen, relative to the exterior of the column label.


Command-Line Name: B<-labels> R<>
Database Name: B<labels>R<>
Database Class: B<Labels>

A boolean value which determines whether column labels are shown or
not.

Command-Line Name: B<-selectcommand> I<command> R<>
Database Name: B<selectCommand>R<>
Database Class: B<Command>

Specifies a Tcl command to be run whenever the selection in the mclistbox
changes. The command will be called with the result of the
curselection command.

Command-Line Name: B<-resizablecolumns> R<>
Database Name: B<resizableColumns> R<>
Database Class: B<ResizableColumns> 

Specifies whether the columns can be resized interactively. If set to
true, mouse bindings will be defined to allow the columns to be
resized. See B<DEFAULT BINDINGS> for more information.

Command-Line Name: B<-selectmode> R<>
Database Name: B<selectMode>R<>
Database Class: B<SelectMode>

Specifies one of several styles for manipulating the selection. The
value of the option may be arbitrary, but the default bindings expect
it to be either B<single>, B<browse>, B<multiple>, or B<extended>; the
default value is browse.

=head2 DESCRIPTION

The B<mclistbox> command creates a new window (given by the
I<pathName> argument) and makes it into a mclistbox widget. Additional
options, described above, may be specified on the command line to
configure aspects of the mclistbox such as its colors, font, text, and
relief.  The mclistbox command returns its pathName argument.  At the
time this command is invoked, there must not exist a window named
pathName, but pathName's parent must exist.  

A mclistbox is a widget that displays a list of strings, one per
line. When first created, a new mclistbox has no elements and no
columns. Columns and elementd may
be added or deleted using widget commands described below.  In
addition, one or more elements may be selected as described
below. When an element in one column is selected, the elements in all
columns are selected. 

It is not necessary for all the elements to be displayed in the
mclistbox window at once; commands described below may be used to change
the view in the window.  Mclistboxes allow scrolling in both directions
using the standard xScrollCommand and yScrollCommand options. They
also support scanning, as described below.

=head2 WIDGET COMMAND

The widget command accepts many of the same arguments as the standard
listbox command. The following commands behave identically to the
standard listbox commands:

=over 4

B<activate>
B<bbox>
B<cget>
B<configure>
B<curselection>
B<delete>
B<index>
B<nearest>
B<scan>
B<see>
B<selection>
B<size>
B<xview>
B<yview>

=back

In addition, there are a few commands unique to the mclistbox:

=over 4

=item I<pathName> B<column> I<option> I<args>

=over 4

=item I<pathName> B<column> B<add> I<name> I<options>

Adds a new column with the specified I<name>. I<options> may be any of
the valid column configuration options.  

=item I<pathName> B<column> B<cget> I<name> I<option>

Returns the value for the named I<option>

=item I<pathName> B<column> B<configure> I<name> I<options>

Configures or queries attributes of the I<named> column.

Columns may be configured with the standard listbox options
(-background, -font, etc). In addition, the following options are
available:

=over 4

=item B<-background>

Specifies the background color for this column. If the value is an
empty string the column will be configured with the default background
color.  

=item B<-bitmap>

Specifies a bitmap to display in the column label. If the bitmap option
is specified then it overrides the -label option. the bitmap option
may be reset to an empty string to re-enable a text display. 

=item B<-font>

Specifies the font for this column. If the value is an empty string
the column will be configured with the default font. Note that if
fonts with different line heights are used for separate columns, the
data in each column will not line up properly.

=item B<-foreground> 

Specifies the foreground color for this column. If the value is an
empty string the column will be configured with the default foreground
color.  


=item B<-image>

Specifies an image to display in the column label. If the image option
is specified then it overrides other options that specify a bitmap or
textual value to display in the column label; the image option may be reset
to an empty string to re-enable a bitmap or text display. 

=item B<-label>

Specifies the desired text to be used in the column label.

=item B<-position>

Specifies the position of the column. 0 (zero) is the first column, 1
(one) is the second and so on.  

=item B<-resizable>

Specifies whether the column is resizable. This only takes effect if 
the global option B<-resizablecolumns> is true.

=item B<-visible>

Specifies whether the column is visible or not. 

=item B<-width>

Specifies the desired width of the column in characters. If the font
doesn't have a uniform width then the width of the character "0" is
used in translating from character units to screen units. If zero or
less, then the desired width for the window is made just large enough
to hold all the elements in the listbox.

=back

=item I<pathName> B<column> B<delete> I<name>

Deletes the column identified by I<name>

=item I<pathName> B<column> B<names>

Returns a list containing all of the column I<names>. The order of the
list is the same as the visual order of the columns, and is the same
as the order the data is returned by the get command.

=item I<pathName> B<column> B<nearext> I<x>

Given a x-coordinate within the mclistbox window, this command returns
the name of the column nearest to that x-coordinate.   

=back

=item I<pathName> B<get> I<first> I<?last?> 

If last is omitted, returns a list containing the element from each
column of the listbox element indicated by I<first>, or an empty list
if first refers to a non-existent element. If I<last> is specified,
the command returns a list of lists. Each of the sublists will contain
an element from each column of the mclistbox element.  Both first and
last may have any of the standard forms for indices.

The data will be returned in the order that the columns appear
visually, and is the same order as reported by the B<column names>
command. 

=item I<pathName> B<insert> I<index> I<list> I<?list ...?>

Inserts zero or more new elements in the mclistbox just before the element
given by index.  If index is specified as end then the new elements
are added to the end of the list.  Returns an empty string. Each
I<list> argument becomes one row in the mclistbox. If there are not
enough elements to fill all of the columns, blank values will be used
to fill in the missing columns.

=item I<pathName> B<label> B<bind> I<name> I<sequence> I<command>

This command associates I<command> with a column label such that
whenever the event sequence given by I<sequence> occurs over the label
for the column named I<name> the I<command> will be invoked. This
widget command is similar to the bind command except that it operates
on column labels rather than the entire widget. See the bind manual
entry for complete details on the syntax of I<sequence> and the
substitutions performed on I<command> before invoking it. 

=back

=head2 EXAMPLE

The following example creates a listbox with three columns:

    mclistbox .listbox -width 55

    .listbox column add name -label "First Name"    -width 20
    .listbox column add size -label "Last Name"     -width 20
    .listbox column add mod  -label "Organization"  -width 15

To add items to the list, do the following:

    .listbox insert end [list "Bryan" "Oakley" "ChannelPoint"]
    .listbox insert end [list "John"  "Ousterhout" "Scriptics"]

... or more concisely:

    .listbox insert end \
            [list "Bryan" "Oakley" "ChannelPoint"] \
            [list "John"  "Ousterhout" "Scriptics"]

To cause a proc to be called when the user clicks on the column label
of the column named "date", use "label bind" as in the following
example: 

    .listbox label bind date <ButtonPress-1> "sortByDate %W"

    proc sortByDate {w} {
        # code to sort the contents of the mclistbox named $w
    }


=head2 DEFAULT BINDINGS

Bindings in the mclistbox widget are the same as for a standard
listbox widget. In addition, if B<-resizableColumns> is set to true,
the following additional bindings are set:

B<ButtonPress-1> starts a column resize if the cursor is over a column
header and within a few pixels of the left or right edge of the
label. B<B1-Motion> causes the columns on either side of the cursor to
be resized based on the amount the cursor has moved. B<ButtonRelease-1> 
ends the column resize. 

If a binding is set on a label via the B<label bind> command and
B<-resizablecolumns> is set, the binding will not fire the event
happens in the few pixels inside the left and right edges of the label
where a resize can occur. This is to prevent both the label bindings
and resize events to happen with a single button press.

=head2 BUGS

If you use different fonts for different columns, the data will not
line up properly if those fonts don't have the same line height. There
is currently no workaround other than to use fonts with the same
line height. 

Mixing labels and images in column headers can produce unevenly sized
headers, which in turn will cause the data in the columns not to line
up properly.

Event bindings that fire while the mouse cursor is over a column
return wrong values for %W, %x and %y. The returned values represent
the internal widgets clicked over rather than relative to the
mclistbox widget (there's no way to work around this in tk as of 8.1b2). 
To convert a value returned for the %W keyword, use
the function ::mclistbox::convert. Sorry, no real documentation here;
see the example (example.tcl) that comes with this distribution

The command "<pathName> xview index" doesn't work. 

I'm sure there are others...

\name{tk2ico}
\alias{tk2ico}
\alias{tk2ico.create}
\alias{tk2ico.destroy}
\alias{tk2ico.hicon}
\alias{tk2ico.info}
\alias{tk2ico.list}
\alias{tk2ico.sizes}
\alias{tk2ico.load}
\alias{tk2ico.pos<-}
\alias{tk2ico.setFromFile}
\alias{tk2ico.set}
\alias{tk2ico.taskbar.add}
\alias{tk2ico.taskbar.delete}
\alias{tk2ico.taskbar.modify}
\alias{tk2ico.text}
\alias{tk2ico.text<-}

\title{ Manipulate icons under Windows }
\description{
  These function are only useful for Windows, but they can be used without error
  on other platform for making platform-independent code that has an additional
  behaviour under Windows. On the other platforms, these function just return
  NULL silently.
}

\usage{
tk2ico.create(iconfile, res = 0, size = 16)
tk2ico.destroy(icon)
tk2ico.list(file = "shell32.dll")
tk2ico.sizes(file = "shell32.dll", res = "application")
tk2ico.load(file = "shell32.dll", res = "application", size = 16)
tk2ico.set(win, icon, pos = NULL, type = c("all", "small", "big"))
tk2ico.setFromFile(win, iconfile)

## Deprecated functions since drop of winico.dll support
tk2ico.hicon(icon)
tk2ico.info(icon, convert = TRUE)
tk2ico.text(icon)
tk2ico.text(icon) <- value
tk2ico.pos(icon) <- value
tk2ico.taskbar.add(icon, pos = 0, text = tk2ico.text(icon),
    leftmenu = NULL, rightmenu = NULL)
tk2ico.taskbar.delete(icon)
tk2ico.taskbar.modify(icon, pos = NULL, text = NULL)
}


\arguments{
  \item{iconfile}{ a file with a .ico, or .exe extension, containing one or
    more Windows icons }
  \item{file}{ a file having icon resources (.exe, or .dll) }
  \item{res}{ the name of the resource from where the icon should be extracted }
  \item{size}{ the size of the icon to use. For windows icons, 16 should be
    fine usually }
  \item{win}{ a Tk window, or an integer representing the handle (HWND) of a
    foreign window whose icon will be changed (take care, the function returns
	\code{TRUE} even if the handle is wrong! }  
  \item{icon}{ an icon object. }

  \item{convert}{ do we convert the result into a data.frame? }
  \item{pos}{ a position (starting from 0) pointing to an icon in a multi-icon
    object. Note that \code{pos} is not used in \code{tk2ico.set()} if
	\code{type = "all"} (in this case, best icons matching both \"small\" and
	\"large\" sizes are searched in the \code{icon} resource. }
  \item{type}{ do we change only the small, the large, or both icons? }
  \item{value}{ a string with the new text for the icon in
    \code{tk2ico.text()} or a numerical value indicating the new default
	position in the icon resource for \code{tk2ico.pos()}. }
  \item{text}{ change a text for an icon. }
  \item{leftmenu}{ a \"tkwin\" object to display when the user left-clicks on
    the taskbar icon (usually, a Tk popup menu), or \code{NULL} for nothing. }
  \item{rightmenu}{ idem as \'lefmenu\' but for a right-click on the taskbar
    icon. }
}

\value{
  These function do nothing and return \code{NULL} on other platforms than
  Windows.
}

\author{ Philippe Grosjean }

\keyword{ utilities }

\concept{ icons, windows }

\name{tk2dde}
\alias{tk2dde}
\alias{tk2dde.exec}
\alias{tk2dde.poke}
\alias{tk2dde.request}
\alias{tk2dde.services}
\alias{.tk2dde.require}

\title{ Use DDE (Dynamic Data Exchange) under Windows }
\description{
  These functions are Windows-specific. They issue an error under a different
  platform.
}

\usage{
tk2dde(topic = NULL)
tk2dde.services(service = "", topic = "")
tk2dde.exec(service, topic, command, async = FALSE)
tk2dde.poke(service, topic, item, data)
tk2dde.request(service, topic, item, binary = FALSE)
}

\arguments{
  \item{topic}{ the 'topic' to reach or expose. A DDE server is accessed as
    'service'|'topic'. In the case of \code{tk2DDE()}, a non null topic activates
    the DDE server, and a null topic deactivate it. }
  \item{service}{ the name of the service to reach. In \code{tk2DDEservices}, if
    both service and topic are empty, the list of all available DDE service is
    returned, otherwise, only available topics for a given service are listed. }
  \item{command}{ a string with the command to run in the external application
    (syntax depends on the server). }
  \item{async}{ is a command run asynchroneously (returns immediately, before
    the command is processed), or not? }
  \item{item}{ the concerned item (usually a variable name, a range in a
    worksheet, etc...). }
  \item{data}{ the new value for the item. }
  \item{binary}{ should the return bez treated as binary data or not? }
}

\keyword{ utilities }
\keyword{ IO }

\concept{ Interprocess communication, DDE }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tk2dde.R
\name{tk2dde}
\alias{tk2dde}
\alias{tk2dde.exec}
\alias{tk2dde.poke}
\alias{tk2dde.request}
\alias{tk2dde.services}
\title{Use DDE (Dynamic Data Exchange) under Windows}
\usage{
tk2dde(topic = NULL)

tk2dde.exec(service, topic, command, async = FALSE)

tk2dde.poke(service, topic, item, data)

tk2dde.request(service, topic, item, binary = FALSE)

tk2dde.services(service = "", topic = "")
}
\arguments{
\item{topic}{The 'topic' to reach or expose. A DDE server is accessed as
service'|'topic'. In the case of \code{\link[=tk2dde]{tk2dde()}}, a non null topic activates
the DDE server, and a null topic deactivate it.}

\item{service}{The name of the service to reach. In \code{tk2dde.services}, if
both service and topic are empty, the list of all available DDE service is
returned, otherwise, only available topics for a given service are listed.}

\item{command}{A string with the command to run in the external application
(syntax depends on the server).}

\item{async}{Is a command run asynchroneously (returns immediately, before
the command is processed), or not?}

\item{item}{The concerned item (usually a variable name, a range in a
worksheet, etc...).}

\item{data}{The new value for the item.}

\item{binary}{Should the return be treated as binary data or not?}
}
\description{
DDE is the first Microsoft's attempt to make an inter-application mechanism.
It is now superseeded by (D)Com, but it is still available (although declared
as unsupported). Being simpler than Com, DDE is interesting for simple tasks.
Applications like Word or Excel provide services one can access through DDE
(see examples). This code if left for backward compatibility, and also, just
in case you will find some use of it. But for new projects in general, you
should not use this any more.
}
\note{
This is only available under Windows. Trying to use these functions under
other platforms raises an error. Under Windows, R is automatically configured
as a DDE server with name 'TclEval|SciViewsR' when this package is loaded.
}
\examples{
\dontrun{
# These cannot be run by examples() but should be OK when pasted
# into an interactive R session with the tcltk package loaded

# Examples of DDE - Windows only

### Examples using wish ###
# Start a Wish84 console side-by-side with R.
# (to get wish, you need to install ActiveTcl from
# http://www.activestate.com/Products/ActiveTcl/)
# Once it is done, start 'Wish84' from the start menu)
# Register the Wish console as a DDE server, that is, type in it
# (\% is the Tcl prompt, do not type it!):
# \% package require dde
# \% dde servername wish

### In R:
tk2dde("R") # Return 0 if succeed
tk2dde.services()
# Evaluate some string in wish
tk2dde.exec("TclEval", "wish", "{puts {Hello World!}}")
# Give a value to a variable in wish
tk2dde.poke("TclEval", "wish", "myvar", "{This is a string!}")
# Note that you must surround strings with curly braces in Tcl!
tk2dde.poke("TclEval", "wish", "mynumvar", c(34.56, 78.9))

# In wish, check that vars exist and have correct value
# \% puts $myvar
# \% puts $mynumvar

# Get the value of one variable from wish into R
tk2dde.request("TclEval", "wish", "myvar")
tk2dde.request("TclEval", "wish", "mynumvar")
# Note that you do not know here if it is a string, a number, or so...
# You have to know and convert yourself!

# Now, the other way: execute a R function from wish
# You first need to register a R function for callback
# (For the moment, only functions without arguments are supported!)
doDDE <- function() cat("DDE execute!")	# A simple function
tclFun(doDDE)
# And in wish
# \% dde execute TclEval R doDDE

# Once you have defined a variable using tclVar, you can get or change it
# from the dde server. However, tclVar gives cryptic names like ::RTcl1.
# So we prefer to use tclVarName()
myvar2 <- tclVarName("myvar2", "this is a test...")
tclvalue(myvar2) # This is the way we access to this variable in R

# In wish you get the value and change it:
# \% dde request TclEval R myvar2
# Again, dde poke does not work and must be replaced by an execute command
# This does not work (???)
# \% dde poke TclEval R myvar2 {yes! and it works...}
# ... but this is fine
# \% dde execute TclEval R {set myvar2 {yes! and it works...}}

# And in R...
tclvalue(myvar2)

### DDE at the command line with execdde.exe ###
# You can also change the value of a variable, or run a command in R from
# the command line using execdde.exe:
# - Download execdde.exe from http://www.sciviews.org/SciViews-R/execdde.zip
# - Unzip it and copy 'execdde.exe' somewhere in your path,
# - Start a DOS window
# - Enter the following commands ('>' is the prompt, do not type it):
# > execdde -s TclEval -t R -c doDDE > NUL
# > if errorlevel 1 echo An error occurs... branch accordingly in your batch!
# > execdde -s TclEval -t R -c "set myvar2 'ok from execdde'" > NUL

# And in R:
tclvalue(myvar2)
# Note: thanks to separate event loops, it works also when R calculates...

### Manipulating Microsoft Excel ###
# Start Excel with a blank workbook, then...

# Change values in Excel from R:
tk2dde.poke("Excel", "Sheet1", "R1C1:R2C1", c("5.7", "6.34"))   # Some data
tk2dde.poke("Excel", "Sheet1", "R3C1", "= A1 + A2")             # A formula

# Read values in Excel (note that results of formulas are returned)
Res <- tk2dde.request("Excel", "Sheet1", "R1C1:R3C1")
Res
as.numeric(Res)

}
}
\seealso{
\code{\link[=tk2reg.get]{tk2reg.get()}}
}
\author{
Philippe Grosjean
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tk2ico.R
\name{tk2ico.create}
\alias{tk2ico.create}
\alias{tk2ico.destroy}
\alias{tk2ico.list}
\alias{tk2ico.sizes}
\alias{tk2ico.load}
\alias{tk2ico.setFromFile}
\alias{tk2ico.set}
\title{Manipulate icons under Windows}
\usage{
tk2ico.create(iconfile, res = 0, size = 16)

tk2ico.destroy(icon)

tk2ico.list(file = "shell32.dll")

tk2ico.sizes(file = "shell32.dll", res = "application")

tk2ico.load(file = "shell32.dll", res = "application", size = 16)

tk2ico.setFromFile(win, iconfile)

tk2ico.set(win, icon)
}
\arguments{
\item{iconfile}{A file with a .ico, or .exe extension, containing one or more
Windows icons}

\item{res}{The name of the resource from where the icon should be extracted.}

\item{size}{Te size of the icon to use. For windows icons, 16 should be fine
usually.}

\item{icon}{A icon object.}

\item{file}{A file having icon resources (.exe, or .dll).}

\item{win}{A Tk window, or an integer representing the handle (HWND) of a
foreign window whose icon will be changed (take care, the function returns
\code{TRUE} even if the handle is wrong!}
}
\value{
An icon object, which is a reference to an image resource in Tcl. Its classes
are \code{c("tclObj", "tclIcon")}. Do not forget to destroy it using
\code{\link[=tk2ico.destroy]{tk2ico.destroy()}} when you do not need it any more!
If \code{\link[=tk2ico.load]{tk2ico.load()}} fails, it returns \code{NULL} instead of a Tcl object.
}
\description{
Create, load and work with Windows icons. Change icons for Windows. These
functions are only useful for Windows, but they silently return \code{NULL} on
other platforms for writing compatible code (Windows icons instructions can
be simply ignored).
}
\note{
This is Windows-specific. It is implemented using the ico Tcl package.
}
\examples{
\dontrun{
# These cannot be run by examples() but should be OK when pasted
# into an interactive R session with the tcltk package loaded

### Examples of tk2ico - icon manipulation under Windows
tt2 <- tktoplevel()
# Load a system icon (there are: "application", "asterisk", "error",
# "exclamation", "hand", "question", "information", "warning", and "winlogo".
Warn <- tk2ico.load(res = "warning")
# Change the icon of my window tt2
tk2ico.set(tt2, Warn)
# Do not forget to destroy icon to free resource when not needed any more
tk2ico.destroy(Warn)
rm(Warn)

### Otherwise, the list of icons in a file are:
tk2ico.list()
# and for a given icon, the various sizes are:
tk2ico.sizes(res = 4)

### One can set icon of a window from an .ico or .exe file directly
tk2ico.setFromFile(tt, default = file.path(R.home("bin"), "Rgui.exe"))

tk2ico.setFromFile(tt2, system.file("gui", "SciViews.ico", package = "tcltk2"))

### When done, dispose of the window and clean the workspace
tkdestroy(tt2)
rm(tt2)
}
}
\seealso{
\code{\link[=tk2dde.exec]{tk2dde.exec()}}, \code{\link[=tk2reg.get]{tk2reg.get()}}
}
\author{
Philippe Grosjean
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tk2fonts.R
\name{tk2font.get}
\alias{tk2font.get}
\alias{tk2font.set}
\alias{tk2font.setstyle}
\title{Manipulate Tk fonts}
\usage{
tk2font.get(font, what = c("family", "size", "bold", "italic"))

tk2font.set(font, settings)

tk2font.setstyle(text = TRUE, system = FALSE, default.styles = FALSE)
}
\arguments{
\item{font}{The name of one or several cached Tk font.}

\item{what}{A list of font characteristics to get: 'family', 'size', 'bold',
italic', 'underline' and/or 'overstrike'. By default, everything except
underline' and 'overstrike'.}

\item{settings}{Settings of fonts. There are two possible forms: (1) a vector
of character strings of same length as font with Tk fonts description like
-family Times -size 12 -weight bold', for instance, or (2) a list of
font characteristics (list with components 'family', 'size', 'bold', 'italic',
'underline' and 'overstrike').}

\item{text}{Do we synchronise text Tk fonts (text, titles, and fixed-font
text) with current settings in \code{.Fonts} inside the \code{SciViews:TempEnv}
environment?}

\item{system}{Do we synchronise system Tk fonts (widgets, window caption,
menus, tooltips, ...) with current system configuration? This is highly
platform dependent. Currently, system settings are gathered only under
Windows, thanks to the \code{winSystemFonts()} function.}

\item{default.styles}{Do we add \code{.fontsStyleXXX} in the \code{SciViews:TempEnv}
environment, where \code{XXX} is one of the four default styles: 'Classic',
'Alternate', 'Presentation' or 'Fancy'.}
}
\value{
\code{\link[=tk2font.get]{tk2font.get()}} retrieves a list with font characteristics (same format
as the \verb{settings =} argument) for the first Tk font found in its \verb{font =}
argument, or \code{""} if the font is not found. \code{\link[=tk2font.set]{tk2font.set()}} changes current
font settings or, possibly, create the Tk font.
\code{\link[=tk2font.setstyle]{tk2font.setstyle()}} changes the current Tk fonts settings according to
actual system and/or text configuration fonts.
}
\description{
Get or set fonts used by Tk widgets from within R.
}
\examples{
\dontrun{
# These cannot be run by examples() but should be OK when pasted
# into an interactive R session with the tcltk package loaded
# Refresh both text and system Tk fonts
tk2font.setstyle(system = TRUE, default.styles = TRUE)
# Get characteristics of the default font
tk2font.get("TkDefaultFont")
}
}
\seealso{
\code{\link[=tk2chooseFont]{tk2chooseFont()}}
}
\author{
Philippe Grosjean
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tk2edit.R
\name{tk2edit}
\alias{tk2edit}
\title{Edit a matrix or data frame in spreadsheet-like editor}
\usage{
tk2edit(
  x,
  title = "Matrix Editor",
  header = NULL,
  maxHeight = 600,
  maxWidth = 800,
  fontsize = 9,
  ...
)
}
\arguments{
\item{x}{A matrix or data frame to edit.}

\item{title}{The title of the editor window.}

\item{header}{Do we display a header?}

\item{maxHeight}{The maximum height of the editor window.}

\item{maxWidth}{The maximum width of the editor window.}

\item{fontsize}{The size of the font to use in the editor window.}

\item{...}{Further arguments to pass to the function.}
}
\value{
The function is used for its side-effet, that is, to modify a matrix or data
frame in a spreadsheet-like editor.
}
\description{
A tkTable widget is used to display and edit a matrix or data frame. One
can edit entries, add or delete rows and columns, ...
}
\note{
You need the tkTable widget to use this function.
}
\examples{
\dontrun{
# These cannot be run by examples() but should be OK when pasted
# into an interactive R session with the tcltk package loaded
data(iris)
tk2edit(iris)
}
}
\seealso{
\code{\link[=tclSetValue]{tclSetValue()}}
}
\author{
Jeffrey J. Hallman
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tk2swaplist.R
\name{tk2swaplist}
\alias{tk2swaplist}
\title{A list selector that allows to select and arrange items freely}
\usage{
tk2swaplist(items, selection, title = "Select items", ...)
}
\arguments{
\item{items}{A vector with all items.}

\item{selection}{A vector with preselected items (must be a subset of \code{items}).}

\item{title}{The title of the dialog box, by default, "Select items".}

\item{...}{Further parameters passed to swaplist, see its tcl man page:
https://core.tcl-lang.org/tklib/doc/trunk/embedded/www/tklib/files/modules/swaplist/swaplist.html.}
}
\value{
A vector with the selected items in the chosen order.
}
\description{
The swaplist is perfect to select and arrange items in a given order from a
fixed initial set of possible items.
}
\examples{
\dontrun{
library(tcltk2)
# tk2swaplist() makes its use super-easy
tk2swaplist(1:9, selection = c(1, 3, 5))

# Use of the swaplist on your own
tclRequire("swaplist")
tt <- tktoplevel()
opts <- tclVar()
sl <- tcl("swaplist::swaplist", tt, opts, 1:9, c(1, 3, 5))
cat("You choose:", tclvalue(opts), "\n")
rm(opts, sl, tt)
}
}
\seealso{
\code{\link[=tk2listbox]{tk2listbox()}}, \code{\link[=tk2tablelist]{tk2tablelist()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tclVarFun.R
\name{makeTclNames}
\alias{makeTclNames}
\alias{tclFun}
\alias{tclGetValue}
\alias{tclSetValue}
\alias{tclVarExists}
\alias{tclVarFind}
\alias{tclVarName}
\title{Manipulate R variables and functions from tcl and back}
\usage{
makeTclNames(names, unique = FALSE)

tclFun(f, name = deparse(substitute(f)))

tclGetValue(name)

tclSetValue(name, value)

tclVarExists(name)

tclVarFind(pattern)

tclVarName(name, init = "", keep.existing = TRUE)
}
\arguments{
\item{names}{Transform names so that they are valid for variables in Tcl.}

\item{unique}{Should these names be unique in the vector?}

\item{f}{An R function. currently, do no support functions with arguments.}

\item{name}{The name of a variable.}

\item{value}{The value to place in a variable.}

\item{pattern}{A pattern to search for.}

\item{init}{Initial value to use when creating the variable.}

\item{keep.existing}{If the tcl variable already exist, should we keep its
content?}
}
\value{
Most of these functions return a 'tclVar' object.
}
\description{
These functions are intended to provide a better "duality" between the name
of variables in both R and tcl, including for function calls. It is possible
to define a variable with the same name in R and tcl (the content is
identical, but copied and coerced in the two respective environments). It is
also possible to get the value of a tcl variable from R, and to call a R
function from within tcl. These features are provided in the tcltk package,
but Tcl variable usually have different internal names as R equivalents.
}
\details{
These functions are similar to \code{\link[tcltk:TclInterface]{tcltk::tclVar()}} from package tcltk, except
for the following change: here, it is possible to propose a name for the
created tcl variable, or to set or retrieve the content of a tcl variable
that is not mirrored in R.
}
\examples{
\dontrun{
# These cannot be run by examples() but should be OK when pasted
# into an interactive R session with the tcltk package loaded

# Tcl functions and variables manipulation
tclVarExists("tcl_version")
tclVarExists("probably_non_existant")
tclVarFind("tcl*")

# Using tclVarName() and tclGetValue()...
# intented for better match between R and Tcl variables
Test <- tclVarName("Test", "this is a test!")
# Now 'Test' exist both in R and in Tcl... In R, you need to use
tclvalue(Test) # to retrieve its content
# If a variable already exists in Tcl, its content is preserved using
# keep.existing = TRUE

# Create a variable in Tcl and assign "just a test..." to it
tclSetValue("A_Variable", "just to test...")
# Create the dual variable with same name
A_Variable <- tclVarName("A_Variable", "something else?")
tclvalue(A_Variable) # Content of the variable is not changed!

# If you want to retrieve the content of a Tcl variable,
# but do not want to create a reference to it in R, use:

tclSetValue("Another_Variable", 1:5)
tclGetValue("Another_Variable") # Get its content in R (no conversion!)
tclSetValue("Another_Variable", paste("Am I", c("happy", "sad"), "?"))
tclGetValue("Another_Variable") # Get its content in R (no conversion!)
}
}
\seealso{
\code{\link[=tk2edit]{tk2edit()}}, \code{\link[tcltk:TclInterface]{tcltk::tclVar()}}
}
\author{
Philippe Grosjean
}
\keyword{utilities}

\name{tk2ico}
\alias{tk2ico}
\alias{tk2ico.create}
\alias{tk2ico.destroy}
\alias{tk2ico.hicon}
\alias{tk2ico.info}
\alias{tk2ico.list}
\alias{tk2ico.sizes}
\alias{tk2ico.load}
\alias{tk2ico.pos<-}
\alias{tk2ico.setFromFile}
\alias{tk2ico.set}
\alias{tk2ico.taskbar.add}
\alias{tk2ico.taskbar.delete}
\alias{tk2ico.taskbar.modify}
\alias{tk2ico.text}
\alias{tk2ico.text<-}

\title{ Manipulate icons under Windows }
\description{
  Create, load and work with Windows icons. Change icons fo Windows, use icons in
  the taskbar under Windows 9X/2000/XP, ... These function are only useful for
  Windows, but they silently return \code{NULL} on other platforms for writing
  compatible code (Windows icons instructions can be simply ignored).
}

\usage{
tk2ico.create(iconfile, res = 0, size = 16)
tk2ico.destroy(icon)
tk2ico.list(file = "shell32.dll")
tk2ico.sizes(file = "shell32.dll", res = "application")
tk2ico.load(file = "shell32.dll", res = "application", size = 16)
tk2ico.set(win, icon, pos = NULL, type = c("all", "small", "big"))
tk2ico.setFromFile(win, iconfile)

## Deprecated functions since drop of winico.dll support
tk2ico.hicon(icon)
tk2ico.info(icon, convert = TRUE)
tk2ico.text(icon)
tk2ico.text(icon) <- value
tk2ico.pos(icon) <- value
tk2ico.taskbar.add(icon, pos = 0, text = tk2ico.text(icon),
    leftmenu = NULL, rightmenu = NULL)
tk2ico.taskbar.delete(icon)
tk2ico.taskbar.modify(icon, pos = NULL, text = NULL)
}

\arguments{
  \item{iconfile}{ a file with a .ico, or .exe extension, containing one or
    more Windows icons }
  \item{file}{ a file having icon resources (.exe, or .dll) }
  \item{res}{ the name of the resource from where the icon should be extracted }
  \item{size}{ the size of the icon to use. For windows icons, 16 should be
    fine usually }
  \item{win}{ a Tk window, or an integer representing the handle (HWND) of a
    foreign window whose icon will be changed (take care, the function returns
	\code{TRUE} even if the handle is wrong! }  
  \item{icon}{ an icon object. }

  \item{convert}{ do we convert the result into a data.frame? }
  \item{pos}{ a position (starting from 0) pointing to an icon in a multi-icon
    object. Note that \code{pos} is not used in \code{tk2ico.set()} if
	\code{type = "all"} (in this case, best icons matching both \"small\" and
	\"large\" sizes are searched in the \code{icon} resource. }
  \item{type}{ do we change only the small, the large, or both icons? }
  \item{value}{ a string with the new text for the icon in
    \code{tk2ico.text()} or a numerical value indicating the new default
	position in the icon resource for \code{tk2ico.pos()}. }
  \item{text}{ change a text for an icon. }
  \item{leftmenu}{ a \"tkwin\" object to display when the user left-clicks on
    the taskbar icon (usually, a Tk popup menu), or \code{NULL} for nothing. }
  \item{rightmenu}{ idem as \'lefmenu\' but for a right-click on the taskbar
    icon. }
}

\value{
  An icon object, which is a reference to an image resource in Tcl. Its classes
  are \code{c("tclObj", "tclIcon")}. Do not forget to destroy it using
  \code{tk2ico.destroy()} when you do not need it any more!
  
  If \code{tk2ico.load()} fails, it returns \code{NULL} instead of a Tcl object.
}

\author{ Philippe Grosjean }

\note{ This is Windows-specific. It was implemented using the winico Tcl library
  in tcltk2 <= 1.2-0. Now it switches to the ico Tcl package, which is supported
  on all platforms. Hence, in future versions of this package, it will be also
  available for Linux/Unix and Max OS X. Under other platforms these functions
  currently just return \code{NULL} without error or warning. So, code that must
  run on different platforms could use these functions all the
  time, as soon as they correctly deal with possible \code{NULL} return.
  The functions \code{tk2ico.hicon()}, \code{tk2ico.info()}, \code{tk2ico.text()},
  \code{tk2ico.pos()}, \code{tk2ico.taskbar.add()}, \code{tk2ico.taskbar.delete()}
  and \code{tk2ico.taskbar.modify()} are deprecated (and by the way, they issue
  an error, because the corresponding Tcl code is not available in the 'ico' Tcl
  package) and will be removed in futures versions of the tcltk2 package. }

\seealso{ \code{\link{tk2dde}}, \code{\link{tk2reg}} }

\examples{
\dontrun{
## These cannot be run by examples() but should be OK when pasted
## into an interactive R session with the tcltk package loaded

### Examples of tk2ico - icon manipulation under Windows
tt2 <- tktoplevel()
## Load a system icon (there are: "application", "asterisk", "error",
## "exclamation", "hand", "question", "information", "warning", and "winlogo".
Warn <- tk2ico.load(res = "warning")
## Change the icon of my window tt2
tk2ico.set(tt2, Warn)
## Do not forget to destroy icon to free resource when not needed any more
tk2ico.destroy(Warn)
rm(Warn)

### Otherwise, the list of icons in a file are:
tk2ico.list()
## and for a given icon, the various sizes are:
tk2ico.sizes(res = 4)

### One can set icon of a window from an .ico or .exe file directly
tk2ico.setFromFile(tt, default = file.path(R.home("bin"), "Rgui.exe"))
 
tk2ico.setFromFile(tt2, system.file("gui", "SciViews.ico", package = "tcltk2"))

### When done, dispose of the window and clean the workspace
tkdestroy(tt2)
rm(tt2)
}
}

\keyword{ utilities }

\concept{ icons, windows }

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tk2widgets.R
\name{tk2button}
\alias{tk2button}
\alias{tk2canvas}
\alias{tk2checkbutton}
\alias{tk2combobox}
\alias{tk2entry}
\alias{tk2frame}
\alias{tk2label}
\alias{tk2labelframe}
\alias{tk2listbox}
\alias{tk2mclistbox}
\alias{tk2menu}
\alias{tk2menubutton}
\alias{tk2message}
\alias{tk2notebook}
\alias{tk2panedwindow}
\alias{tk2progress}
\alias{tk2radiobutton}
\alias{tk2scale}
\alias{tk2scrollbar}
\alias{tk2separator}
\alias{tk2spinbox}
\alias{tk2text}
\alias{tk2ctext}
\alias{tk2tree}
\alias{tk2table}
\alias{tk2tablelist}
\title{A series of versatile using either themable ttk widgets}
\usage{
tk2button(parent, tip = "", ...)

tk2canvas(parent, tip = "", ...)

tk2checkbutton(parent, tip = "", ...)

tk2combobox(parent, tip = "", ...)

tk2entry(parent, tip = "", ...)

tk2frame(parent, ...)

tk2label(parent, tip, label, tag, cfglist, wrap = FALSE, ...)

tk2labelframe(parent, ...)

tk2listbox(
  parent,
  values,
  value,
  selection,
  selectmode = c("extended", "single", "browse", "multiple"),
  height = 5,
  tip = "",
  scroll = "both",
  autoscroll = "x",
  enabled = TRUE,
  ...
)

tk2mclistbox(parent, tip = "", ...)

tk2menu(parent, activebackground, activeforeground, ...)

tk2menubutton(parent, tip = "", ...)

tk2message(
  parent,
  text = "",
  justify = c("left", "center", "right"),
  width = -1,
  aspect = 150,
  tip = "",
  ...
)

tk2notebook(parent, tabs, ...)

tk2panedwindow(parent, orientation = c("horizontal", "vertical"), ...)

tk2progress(parent, orientation = c("horizontal", "vertical"), tip = "", ...)

tk2radiobutton(parent, tip = "", ...)

tk2scale(parent, orientation = c("horizontal", "vertical"), tip = "", ...)

tk2scrollbar(parent, orientation = c("horizontal", "vertical"), ...)

tk2separator(parent, orientation = c("horizontal", "vertical"), ...)

tk2spinbox(parent, tip = "", ...)

tk2text(parent, tip = "", ...)

tk2ctext(parent, tip = "", ...)

tk2tree(parent, tip = "", ...)

tk2table(parent, ...)

tk2tablelist(parent, ...)
}
\arguments{
\item{parent}{The parent window.}

\item{tip}{A tooltip to display for this widget (optional).}

\item{...}{Further arguments passed to the widget.}

\item{label}{A single character string used to label that widget (optional).}

\item{tag}{Ay object that you would like to associate with this widget
(optional).}

\item{cfglist}{A named list with configuration parameters and values to
apply.}

\item{wrap}{Do we wrap long lines in the widget?}

\item{values}{A character vector with values to use to populate the widget.}

\item{value}{A character vector with current value for the widget, or
currently selected values, if multiple selection is allowed. Takes precedence
on \code{selection}.}

\item{selection}{A numeric (indices) vector with current selection.}

\item{selectmode}{The selection mode for this widget. \code{extended} is the
usual choice for multiselection \code{tk2listbox()}.}

\item{height}{The height of the widget.}

\item{scroll}{Do we add scrollbars? Possible values are \code{"x"}, \code{"y"},
\code{"both"} or \code{"none"}; can be abridged.}

\item{autoscroll}{Do we automatically hide scrollbars if not needed? Possible
values are the same as for the \code{scroll} argument.}

\item{enabled}{Is the widget enabled or disabled?}

\item{activebackground}{Color to use for active background of menu items (if
not provided, a reasonable default value is used).}

\item{activeforeground}{Color to use for active foreground of menu items (if
not provided, a reasonable default value is used).}

\item{text}{The text to display in the widget.}

\item{justify}{How text is justified?}

\item{width}{The desired width. Use a negative value to use \code{aspect} instead.}

\item{aspect}{Sets the aspect ratio of the widget (100 = square, 200 = twice
as large, 50 = twice as tall). Only used if \code{width} is negative.}

\item{tabs}{The tabs to create in the notebook widget.}

\item{orientation}{Either \code{"horizontal"} or \code{"vertical"}.}
}
\value{
The reference to the created widget.
}
\description{
A series of widgets you can use in your Tk windows/dialog boxes.
}
\note{
You need Tk 8.5 or above to use these widgets.
}
\examples{
\dontrun{
# These cannot be run by examples() but should be OK when pasted
# into an interactive R session with the tcltk package loaded

# A tk2notebook example
tt2 <- tktoplevel()
nb <- tk2notebook(tt2, tabs = c("Test", "Button"))
tkpack(nb, fill = "both", expand = 1)
tb1 <- tk2notetab(nb, "Test")
lab <- tk2label(tb1, text = "Nothing here.")
tkpack(lab)
tb2 <- tk2notetab(nb, "Button")
but <- tk2button(tb2, text = "Click me", command = function() tkdestroy(tt2))
tkgrid(but)
tk2notetab.select(nb, "Button")
tk2notetab.text(nb) # Text of the currently selected tab

# A simple tk2panedwindow example
tt2 <- tktoplevel()
pw <- tk2panedwindow(tt2, orient = "vertical")
lpw.1 <- tk2text(pw)
lpw.2 <- tk2text(pw)
tkadd(pw, lpw.1)#, minsize = 100)
tkadd(pw, lpw.2)#, minsize = 70)
but <- tk2button(tt2, text = "OK", width = 10,
  command = function() tkdestroy(tt2))
tkpack(pw, fill = "both", expand = "yes")
tkpack(but)
# Resize the window and move the panel separator with the mouse

# A tk2combobox example
tt2 <- tktoplevel()
cb <- tk2combobox(tt2)
tkgrid(cb)
# Fill the combobox list
fruits <- c("Apple", "Orange", "Banana")
tk2list.set(cb, fruits)
tk2list.insert(cb, "end", "Scoubidou", "Pear")
tk2list.delete(cb, 3)   # 0-based index!
tk2list.size(cb)
tk2list.get(cb)   # All items
# Link current selection to a variable
Fruit <- tclVar("Pear")
tkconfigure(cb, textvariable = Fruit)
# Create a button to get the content of the combobox
but <- tk2button(tt2, text = "OK", width = 10,
  command = function() {tkdestroy(tt2); cat(tclvalue(Fruit), "\n")})
tkgrid(but)

# An example of a tk2spinbox widget
tt2 <- tktoplevel()
tspin <- tk2spinbox(tt2, from = 2, to = 20, increment = 2)
tkgrid(tspin)
# This widget is not added yet into tcltk2!
#tdial <- tk2dial(tt2, from = 0, to = 20, resolution = 0.5, width = 70,
#	tickinterval = 2)
#tkgrid(tdial)
tbut <- tk2button(tt2, text = "OK", width = 10,
  command = function() tkdestroy(tt2))
tkgrid(tbut)

# A tk2mclistbox example
tt2 <- tktoplevel()
mlb <- tk2mclistbox(tt2, width = 55, resizablecolumns = TRUE)
# Define the columns
tk2column(mlb, "add", "name", label = "First name", width = 20)
tk2column(mlb, "add", "lastname", label = "Last name", width = 20)
tk2column(mlb, "add", "org", label = "Organisation", width = 15)
tkgrid(mlb)
# Fill the multicolumn list (we can use a vector, or a matrix of character strings)
item1 <- c("Bryan", "Oackley", "ChannelPoint")
items <- matrix(c("John", "Ousterhout", "Scriptics", "Steve", "Miller", "TclTk inc."),
  ncol = 3, byrow = TRUE)
tk2insert.multi(mlb, "end", item1)
tk2insert.multi(mlb, "end", items)
# TODO: bind events
# Ex: .listbox label bind date <ButtonPress-1> "sortByDate \%W"
# See the example.tcl in .\libs\mclistbox1.02 for a more complex example
# Create a button to close the dialog box
but <- tk2button(tt2, text = "OK", width = 10,
  command = function() tkdestroy(tt2))
tkgrid(but)

# A simple tk2table example (Tktable is required here!)
myRarray <- c("Animal", "\"sphinx moth\"", "oyster", "Type", "insect", "mollusk")
dim(myRarray) <- c(3, 2)
for (i in (0:2))
  for (j in (0:1))
    .Tcl(paste("set tclarray(", i, ",", j, ") ", myRarray[i+1, j+1], sep = ""))
tt2 <- tktoplevel()
table1 <- tk2table(tt2, variable = "tclarray", rows = "3", cols = "2",
  titlerows = "1", selectmode = "extended", colwidth = "25", background = "white")
tkpack(table1)
# A tablelist example
tt <- tktoplevel()
tlist <- tk2tablelist(tt, columntitles = c("First column", "Second column"),
  stretch = "all", expand = 1)
tkpack(tlist, fill = "both")
tkinsert(tlist, "end", c("first row", "another value"))
tkinsert(tlist, "end", c("another row", "bla bla"))
tbut <- tk2button(tt, text = "Done", command = function () tkdestroy(tt))
tkpack(tbut)
}
}
\seealso{
\code{\link[=is.ttk]{is.ttk()}}
}
\author{
Philippe Grosjean
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tk2commands.R
\name{tk2column}
\alias{tk2column}
\alias{tk2list.set}
\alias{tk2list.insert}
\alias{tk2list.delete}
\alias{tk2list.get}
\alias{tk2list.size}
\alias{tk2state.set}
\alias{tk2insert.multi}
\alias{tk2notetraverse}
\alias{tk2notetab}
\alias{tk2notetab.select}
\alias{tk2notetab.text}
\alias{tk2theme.elements}
\alias{tk2theme.list}
\alias{tk2theme}
\alias{tk2style}
\alias{tk2dataList}
\alias{tk2configList}
\alias{is.tk}
\alias{is.ttk}
\title{Tk commands associated with the tk2XXX widgets}
\usage{
tk2column(
  widget,
  action = c("add", "configure", "delete", "names", "cget", "nearest"),
  ...
)

tk2list.set(widget, items)

tk2list.insert(widget, index = "end", ...)

tk2list.delete(widget, first, last = first)

tk2list.get(widget, first = 0, last = "end")

tk2list.size(widget)

tk2state.set(widget, state = c("normal", "disabled", "readonly"))

tk2insert.multi(widget, where = "end", items)

tk2notetraverse(nb)

tk2notetab(nb, tab)

tk2notetab.select(nb, tab)

tk2notetab.text(nb)

tk2theme.elements()

tk2theme.list()

tk2theme(theme = NULL)

tk2style(
  class,
  style,
  state = c("default", "active", "disabled", "focus", "!focus", "pressed", "selected",
    "background", "readonly", "alternate", "invalid", "hover", "all"),
  default = NULL
)

tk2dataList(x)

tk2configList(x)

is.tk()

is.ttk()
}
\arguments{
\item{widget}{The widget to which these actions apply.}

\item{action}{Which kind of action?}

\item{...}{Further arguments to the action.}

\item{items}{The items to add (either a vector for a single line, or a matrix
for more items).}

\item{index}{The 0-based index where to insert items in the list.}

\item{first}{The 0-based first index to consider in the list.}

\item{last}{The 0-based last index to consider in the list, or \code{"end"} for
using the last element of the list.}

\item{state}{The new state of the widget, or the state to inquiry.}

\item{where}{Where are these item added in the list (by default, at the end).}

\item{nb}{A tk2notebook widget ('tclObj' object).}

\item{tab}{The name (text) of a tab in a notebook.}

\item{theme}{A theme to use (character string).}

\item{class}{The class of the tk2widget (either the Tk class, like \code{TButton},
or the name of the function that creates it, like \code{\link[=tk2button]{tk2button()}}.}

\item{style}{A character string with the name of the style to retrieve.}

\item{default}{The default value to return in case this style is not found.}

\item{x}{Either a tk2widget object, or a character string with its class
name.}
}
\value{
Nothing, these functions are used for their side-effect of changing the state
of Tk widgets
}
\description{
These commands supplement those available in the tcltk package to ease
manipulation of tk2XXX widgets.
}
\details{
\code{\link[=tk2column]{tk2column()}} manipulate columns of a tk2mclistbox widget,
\code{\link[=tk2insert.multi]{tk2insert.multi()}} is used to insert multiple field entries in a
tk2mclistbox widget,
\code{\link[=is.tk]{is.tk()}} determines if the tk package is loaded (on some platforms it is
possible to load the tcltk package without tk, for instance, in batch mode).
\code{\link[=is.ttk]{is.ttk()}} determines if 'ttk' widgets (styled widgets) used by the
\code{tk2XXX()} functions are available (you need Tk >= 8.5).
}
\note{
In comparison with traditional Tk widgets, ttk proposes an advances mechanism
for styling the widgets with "themes". By default, it adapts to the current
platform (for instance, under Windows, all widgets take the appearance of
Windows themed widgets (even with custom themes applied!). Usual Tk widgets
are ALWAYS displayed in old-looking fashion under Windows. If you want, you
can switch dynamically to a different theme among those available (list them
using \code{\link[=tk2theme.list]{tk2theme.list()}}, and switch to another one with \code{tk2theme(newtheme)}.
This is most useful to see how your GUI elements and dialog boxes look like
on foreign systems. If you prefer, let's say, a Unix look of the R GUI
elements under Windows, these functions are also useful. If you are more
adventurous, you can even design your own themes (see the tile documentation
on the Tcl wiki).
}
\examples{
\dontrun{
# These cannot be run by examples() but should be OK when pasted
# into an interactive R session with the tcltk package loaded

tt <- tktoplevel()
# A label with a image and some text
file <- system.file("gui", "SciViews.gif", package = "tcltk2")

# Make this a tk2image function...
Image <- tclVar()
tkimage.create("photo", Image, file = file)

tlabel <- tk2label(tt, image = Image,
  text = "A label with an image")
tkpack(tlabel)
config(tlabel, compound = "left")

tlabel2 <- tk2label(tt, text = "A disabled label")
tkpack(tlabel2)
disabled(tlabel2) <- TRUE

fruits <- c("Apple", "Orange", "Banana")
tcombo <- tk2combobox(tt, values = fruits)
tkpack(tcombo)
tkinsert(tcombo, 0, "Apple")

# Buttons
tbut <- tk2button(tt, text = "Enabled")
tbut2 <- tk2button(tt, text = "Disabled")
tkpack(tbut, tbut2)
tkconfigure(tbut2, state = "disabled")

tcheck <- tk2checkbutton(tt, text = "Some checkbox")
tcheck2 <- tk2checkbutton(tt, text = "Disabled checkbox")
tkconfigure(tcheck2, state = "disabled")
tcheck3 <- tk2checkbutton(tt, text = "Disabled and selected")
tkpack(tcheck, tcheck2, tcheck3)
cbValue <- tclVar("1")
tkconfigure(tcheck3, variable = cbValue)
tkconfigure(tcheck3, state = "disabled")

tradio <- tk2radiobutton(tt, text = "Some radiobutton")
tradio2 <- tk2radiobutton(tt, text = "Disabled and checked")
tkpack(tradio, tradio2)
tkconfigure(tradio2, state = "checked")
tkconfigure(tradio2, state = "disabled")

# Menu allowing to change ttk theme
topMenu <- tkmenu(tt)           # Create a menu
tkconfigure(tt, menu = topMenu) # Add it to the 'tt' window
themes <- tk2theme.list()
themeMenu <- tkmenu(topMenu, tearoff = FALSE)
if ("alt" \%in\% themes) tkadd(themeMenu, "command", label = "alt",
  command = function() tk2theme("alt"))
if ("aqua" \%in\% themes) tkadd(themeMenu, "command", label = "aqua",
  command = function() tk2theme("aqua"))
if ("clam" \%in\% themes) tkadd(themeMenu, "command", label = "clam",
  command = function() tk2theme("clam"))
tkadd(themeMenu, "command", label = "clearlooks",
  command = function() tk2theme("clearlooks"))
if ("classic" \%in\% themes) tkadd(themeMenu, "command", label = "classic",
  command = function() tk2theme("classic"))
if ("default" \%in\% themes) tkadd(themeMenu, "command", label = "default",
  command = function() tk2theme("default"))
tkadd(themeMenu, "command", label = "keramik",
  command = function() tk2theme("keramik"))
tkadd(themeMenu, "command", label = "plastik",
  command = function() tk2theme("plastik"))
tkadd(themeMenu, "command", label = "radiance (fonts change too)!",
  command = function() tk2theme("radiance"))
if ("vista" \%in\% themes) tkadd(themeMenu, "command", label = "vista",
  command = function() tk2theme("vista"))
if ("winnative" \%in\% themes) tkadd(themeMenu, "command", label = "winnative",
  command = function() tk2theme("winnative"))
if ("xpnative" \%in\% themes) tkadd(themeMenu, "command", label = "xpnative",
  command = function() tk2theme("xpnative"))
tkadd(themeMenu, "separator")
tkadd(themeMenu, "command", label = "Quit", command = function() tkdestroy(tt))
tkadd(topMenu, "cascade", label = "Theme", menu = themeMenu)
tkfocus(tt)
}
}
\seealso{
\code{\link[=tk2button]{tk2button()}}, \code{\link[=tk2tip]{tk2tip()}}
}
\author{
Philippe Grosjean
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tk2methods.R
\name{is.tk2widget}
\alias{is.tk2widget}
\alias{print.tk2widget}
\alias{tk2cfglist}
\alias{print.tk2cfglist}
\alias{state}
\alias{state.tk2widget}
\alias{label}
\alias{label.tk2widget}
\alias{label<-}
\alias{label<-.tk2widget}
\alias{tag}
\alias{tag.tk2widget}
\alias{tag<-}
\alias{tag<-.tk2widget}
\alias{disabled}
\alias{disabled.tk2widget}
\alias{disabled<-}
\alias{disabled<-.tk2widget}
\alias{values}
\alias{values.tk2widget}
\alias{values.tk2listbox}
\alias{values<-}
\alias{values<-.tk2widget}
\alias{values<-.tk2listbox}
\alias{value}
\alias{value.tk2widget}
\alias{value.tk2listbox}
\alias{value<-}
\alias{value<-.tk2widget}
\alias{value<-.tk2listbox}
\alias{selection}
\alias{selection.tk2widget}
\alias{selection.tk2listbox}
\alias{selection<-}
\alias{selection<-.tk2widget}
\alias{selection<-.tk2listbox}
\alias{visibleItem}
\alias{visibleItem.tk2widget}
\alias{visibleItem.tk2listbox}
\alias{size}
\alias{size.tk2widget}
\alias{size.tk2listbox}
\alias{config}
\alias{config.tk2widget}
\alias{config.tk2label}
\alias{config<-}
\alias{config<-.tk2widget}
\alias{config<-.tk2label}
\title{A series of methods applicable to tk2widget or tk2cfglist objects}
\usage{
is.tk2widget(x)

\method{print}{tk2widget}(x, ...)

tk2cfglist(...)

\method{print}{tk2cfglist}(x, ...)

state(x, ...)

\method{state}{tk2widget}(x, ...)

label(x, ...)

\method{label}{tk2widget}(x, ...)

label(x) <- value

\method{label}{tk2widget}(x) <- value

tag(x, ...)

\method{tag}{tk2widget}(x, ...)

tag(x) <- value

\method{tag}{tk2widget}(x) <- value

disabled(x, ...)

\method{disabled}{tk2widget}(x, ...)

disabled(x) <- value

\method{disabled}{tk2widget}(x) <- value

values(x, ...)

\method{values}{tk2widget}(x, ...)

\method{values}{tk2listbox}(x, ...)

values(x) <- value

\method{values}{tk2widget}(x) <- value

\method{values}{tk2listbox}(x) <- value

value(x, ...)

\method{value}{tk2widget}(x, ...)

\method{value}{tk2listbox}(x, ...)

value(x) <- value

\method{value}{tk2widget}(x) <- value

\method{value}{tk2listbox}(x) <- value

selection(x, ...)

\method{selection}{tk2widget}(x, ...)

\method{selection}{tk2listbox}(x, ...)

selection(x) <- value

\method{selection}{tk2widget}(x) <- value

\method{selection}{tk2listbox}(x) <- value

visibleItem(x, index, ...)

\method{visibleItem}{tk2widget}(x, index, ...)

\method{visibleItem}{tk2listbox}(x, index, ...)

size(x, ...)

\method{size}{tk2widget}(x, ...)

\method{size}{tk2listbox}(x, ...)

config(x, ...)

\method{config}{tk2widget}(x, cfglist, ...)

\method{config}{tk2label}(x, cfglist, ...)

config(x) <- value

\method{config}{tk2widget}(x) <- value

\method{config}{tk2label}(x) <- value
}
\arguments{
\item{x}{A tk2widget object.}

\item{...}{A series of named arguments corresponding to parameters and values
to use for the configuration for \code{tk2cfglist()}, or reserved arguments for
future use for the other function (not used yet).}

\item{value}{A value to assign to the object's method.}

\item{index}{The zero-based index of the item to make visible.}

\item{cfglist}{a list containing one or more named items, with the name
being a Tcl/Tk property and items being the new value for the property.}
}
\value{
Depends on the function. The \code{is.xxx()} function return \code{TRUE} or \code{FALSE} if
the object is of the right class or not. The assignations form return the
assigned value. The direct form return the item.
}
\description{
Tk2widgets can be used as tcltk widgets, but they propose also an
object-oriented interaction through these different methods.
}
\seealso{
\code{\link[=tk2button]{tk2button()}}, \code{\link[=tk2tip]{tk2tip()}}
}
\author{
Philippe Grosjean
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tk2tip.R
\name{tk2tip}
\alias{tk2tip}
\alias{tk2killtip}
\alias{tip}
\alias{tip.tk2widget}
\alias{tip<-}
\alias{tip<-.tk2widget}
\title{Display and manage tooltips in Tk widgets}
\usage{
tk2tip(widget, message)

tk2killtip()

tip(x, ...)

\method{tip}{tk2widget}(x, ...)

tip(x) <- value

\method{tip}{tk2widget}(x) <- value
}
\arguments{
\item{widget}{The widget to which a tooltip is attached.}

\item{message}{The message of the tooltip (\code{""} to remove the tooltip).}

\item{x}{A tk2widget object.}

\item{...}{Further arguments to the method (unused, but reserved for
future use).}

\item{value}{The message of the tooltip, or \code{""} to remove the tip.}
}
\value{
The current tip or \code{NULL} depending on the function.
}
\description{
\code{\link[=tk2tip]{tk2tip()}} provides a simple mechanism to display tooltips on Tk widgets when
the mouse cursor hoovers on top of them.
}
\note{
This implementation is done in pure Tcl code.
}
\examples{
\dontrun{
# These cannot be run by examples() but should be OK when pasted
# into an interactive R session with the tcltk package loaded

# Using plain Tcl/Tk label and button (tk2XXX equivalent have built-in
# tooltip features)
tt <- tktoplevel()
lb <- tklabel(tt, text = "Move mouse over me, or over the button to see tooltip")
tkgrid(lb)
tk2tip(lb, "A tooltip for the label \ndisplayed on two lines")
but <- tkbutton(tt, text = "Exit", width = 10,
  command = function() tkdestroy(tt))
tkgrid(but)
tk2tip(but, "Exit from this dialog box")

# To test tk2killtip(), move mouse on top of a widget
# so that the tip is visible, and force killing it manually using:
tk2killtip()
# Move again to the widget: the tip is displayed again.

# With tk2widgets, the tip() method can also be used:
lb2 <- tk2label(tt, text = "Move also over me to see the tooltip")
tkgrid(lb2)
tip(lb2) # No tip yet
tip(lb2) <- "Now the tooltip is there!"
# Move the mouse over that last label

tip(lb2) # Yes, this is my tooltip
tip(lb2) <- NULL # To eliminate the tooltip for this widget
}
}
\seealso{
\code{\link[=tk2button]{tk2button()}}, \code{\link[=label]{label()}}
}
\author{
Philippe Grosjean
}
\keyword{utilities}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tcltk2-package.R
\docType{package}
\name{tcltk2-package}
\alias{tcltk2}
\alias{tcltk2-package}
\title{tcltk2: Tcl/Tk Additions}
\description{
A series of additional Tcl commands and Tk widgets to supplement the tcltk package.
}
\details{
The \{tcltk2\} package provides additions to the \{tcltk\} base package.

Many Tk widgets are added. Also, features of the Tcl language that can be
useful also in R are made accessible through R functions.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/SciViews/tcltk2/}
  \item Report bugs at \url{https://github.com/SciViews/tcltk2/issues/}
}

}
\author{
\strong{Maintainer}: Philippe Grosjean \email{phgrosjean@sciviews.org} (\href{https://orcid.org/0000-0002-2694-9471}{ORCID})

}
\keyword{internal}
